/*  latex-expand.c

    Source file for latex-expand

    Copyright (C) 1997-2002  Richard Gooch

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Richard Gooch may be reached by email at  karma-request@atnf.csiro.au
    The postal address is:
      Richard Gooch, c/o ATNF, P. O. Box 76, Epping, N.S.W., 2121, Australia.
*/

/*
    This programme will expand a LaTeX file.


    Written by      Richard Gooch   23-OCT-1997

    Updated by      Richard Gooch   26-FEB-1999: Took account of filenames with
  '.' characters.

    Updated by      Richard Gooch   16-JUL-1999: Support variable expansion in
  filenames.

    Last updated by Richard Gooch   21-DEC-2002: Do not process comment lines.


*/
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <unistd.h>
#include <errno.h>
#include <sys/types.h>
#include <math.h>
#include <karma.h>
#include <karma_ch.h>
#include <karma_st.h>


/*  Private functions  */
STATIC_FUNCTION (void expand_file, (CONST char *source) );


/*  Public functions follow  */

int main (int argc, char **argv)
{
    static char usage_string[] = "Usage:\tlatex-expand sourcefile";

    if (argc != 2)
    {
	fprintf (stderr, "%s\n", usage_string);
	exit (RV_BAD_PARAM);
    }
    expand_file (argv[1]);
    return (RV_OK);
}   /*  End Function main  */


static void expand_file (CONST char *source)
/*  [SUMMARY] Expand a LaTeX file.
    <source> The source file.
    [RETURNS] Nothing.
*/
{
    Channel channel;
    char *ptr, *start;
    char filename[STRING_LENGTH], buffer[STRING_LENGTH];

    if ( !st_expr_expand (filename, STRING_LENGTH, source, NULL, NULL, NULL) )
    {
	fprintf (stderr, "Error expanding expression: \"%s\"\n", filename);
	exit (RV_UNDEF_ERROR);
    }
    if ( ( ptr = strstr (filename, ".tex") ) == NULL )
	strcat (filename, ".tex");
    if ( ( channel = ch_open_file (filename, "r") ) == NULL )
    {
	fprintf (stderr, "Error opening file: \"%s\"\t%s\n",
		 filename, ERRSTRING);
	exit (RV_CANNOT_OPEN);
    }
    printf ("%%Opened file: \"%s\"\n", filename);
    while ( ch_getl (channel, buffer, STRING_LENGTH) )
    {
	unsigned int brace_count;

	for (start = buffer; TRUE; )
	{
	    if ( (*start == '\0') || (*start == '%') )
	    {
		puts (start);
		break;
	    }
	    if ( ( ptr = strstr (start, "\\input{") ) == NULL )
	    {
		puts (start);
		break;
	    }
	    *ptr = '\0';
	    puts (start);
	    start = ptr;
	    /*  Skip until closing brace/terminator found  */
	    for (ptr += 7, brace_count = 1;
		 (*ptr != '\0') && (brace_count > 0);
		 ++ptr)
	    {
		switch (*ptr)
		{
		  case '{':
		    ++brace_count;
		    break;
		  case '}':
		    --brace_count;
		    break;
		}
	    }
	    if (brace_count > 0)
	    {
		*start = '\\';
		puts (start);
		break;
	    }
	    ptr[-1] = '\0';
	    expand_file (start + 7);
	    start = ptr;
	}
    }
    ch_close (channel);
    printf ("%%Closed file: \"%s\"\n", filename);
}   /*  End Function expand_file  */
