/*  This programme will sort all Karma packages in level order.


    Written by      Richard Gooch   15-APR-1993

    Updated by      Richard Gooch   30-NOV-1994

    Updated by      Richard Gooch   7-DEC-1994: Stripped declaration of  errno
  and added #include <errno.h>

    Updated by      Richard Gooch   28-MAR-1997: Pass in doc directory.

    Last updated by Richard Gooch   15-AUG-1999: Made egcs happy.


*/
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <dirent.h>
#include <errno.h>
#include <math.h>
#include <karma.h>
#include <karma_ch.h>
#include <karma_st.h>
#include <karma_r.h>
#include <karma_m.h>
#include <karma_a.h>

int main (int argc, char *argv[])
{
    unsigned int num_levels = 0;
    unsigned int level_count;
    char *ptr1, *ptr2;
    unsigned int *num_packages;
    char ***package_names;
    char ***ptr_package_names;
    char *dirname;
    DIR *dirp;
    struct dirent *dp;
    static char usage_string[] = "Usage:\tsort_packages doc_directory\n";
    static char function_name[] = "main";

    if (argc != 2)
    {
	fprintf (stderr, usage_string);
	exit (RV_TOO_MANY_PARAM);
    }
    dirname = argv[1];
    /*  Start to go through source directory    */
    errno = 0;
    if ( ( dirp = opendir (dirname) ) == NULL )
    {
	fprintf (stderr, "Error reading directory: %s\t%s\n",
		 dirname, ERRSTRING);
	exit (RV_SYS_ERROR);
    }
    /*  First pass: count number of levels  */
    while ( ( dp = readdir (dirp) ) != NULL )
    {
	/*  Still another entry */
        if ( (strcmp (".", dp->d_name) == 0) ||
	     (strcmp ("..", dp->d_name) == 0) )
        {
	    continue;
	}
	/*  File is not a special directory */
	/*  Find start of package name (after possible '_')  */
	if ( ( ptr1 = strchr (dp->d_name, '_') ) == NULL )
	{
	    ptr1 = dp->d_name;
	}
	else
	{
	    ++ptr1;
	}
	if ( ( ptr2 = strchr (ptr1, '.') ) == NULL ) continue;
	if (strcmp (ptr2, ".doc") != 0) continue;
	if ( num_levels < (int) (ptr2 - ptr1) ) num_levels = ptr2 - ptr1;
    }
    if ( ( num_packages = (unsigned int *)
	   m_alloc (sizeof *num_packages * num_levels) )
	 == NULL )
    {
	m_abort (function_name, "level byte counters");
    }
    m_clear ( (char *) num_packages, sizeof *num_packages * num_levels );
    if ( ( package_names = (char ***)
	   m_alloc (sizeof *package_names * num_levels) )
	 == NULL )
    {
	m_abort (function_name, "level pointers");
    }
    m_clear ( (char *) package_names, sizeof *package_names * num_levels );
    if ( ( ptr_package_names = (char ***)
	   m_alloc (sizeof *ptr_package_names * num_levels) )
	 == NULL )
    {
	m_abort (function_name, "ptr_level pointers");
    }
    rewinddir (dirp);
    /*  Second pass: number of packages in each level  */
    while ( ( dp = readdir (dirp) ) != NULL )
    {
	/*  Still another entry */
        if ( (strcmp (".", dp->d_name) == 0) ||
	     (strcmp ("..", dp->d_name) == 0) )
        {
	    continue;
	}
	/*  File is not a special directory */
	/*  Find start of package name (after possible '_')  */
	if ( ( ptr1 = strchr (dp->d_name, '_') ) == NULL )
	{
	    ptr1 = dp->d_name;
	}
	else
	{
	    ++ptr1;
	}
	if ( ( ptr2 = strchr (ptr1, '.') ) == NULL ) continue;
	if (strcmp (ptr2, ".doc") != 0) continue;
	++num_packages[ptr2 - ptr1 - 1];
    }
    /*  Allocate storage for package name pointer arrays  */
    for (level_count = 0; level_count < num_levels; ++level_count)
    {
	if ( ( package_names[level_count] = (char **)
	m_alloc ( sizeof **package_names *
		  (num_packages[level_count] + 1) ) )
	     == NULL )
	{
	    m_abort (function_name, "package name pointer arrays");
	}
	m_clear ( (char *) package_names[level_count], sizeof **package_names *
		  (num_packages[level_count] + 1) );
    }
    m_copy ( (char *) ptr_package_names, (char *) package_names,
	     sizeof *ptr_package_names * num_levels );
    rewinddir (dirp);
    /*  Third pass: copy package names  */
    while ( ( dp = readdir (dirp) ) != NULL )
    {
	/*  Still another entry  */
        if ( (strcmp (".", dp->d_name) == 0) ||
	     (strcmp ("..", dp->d_name) == 0) )
        {
	    continue;
	}
	/*  File is not a special directory */
	/*  Find start of package name (after possible '_')  */
	if ( ( ptr1 = strchr ( dp->d_name, '_' ) ) == NULL )
	{
	    ptr1 = dp->d_name;
	}
	else
	{
	    ++ptr1;
	}
	if ( ( ptr2 = strchr (ptr1, '.') ) == NULL ) continue;
	if (strcmp (ptr2, ".doc") != 0) continue;
	if ( ( *ptr_package_names[ptr2 - ptr1 - 1] =
	       m_alloc (ptr2 - dp->d_name + 1) )
	     == NULL )
	{
	    m_abort (function_name, "package name");
	}
	strncpy (*ptr_package_names[ptr2 - ptr1 - 1], dp->d_name,
		 ptr2 - dp->d_name);
	ptr_package_names[ptr2 - ptr1 - 1][0][ptr2 - dp->d_name] = '\0';
	++ptr_package_names[ptr2 - ptr1 - 1];
    }
    /*  Close source directory (and deallocate memory used) */
    closedir (dirp);
    /*  Print them all  */
    m_copy ( (char *) ptr_package_names, (char *) package_names,
	     sizeof *ptr_package_names * num_levels );
    for (level_count = 0; level_count < num_levels; ++level_count)
    {
/*
	st_qsort (package_names[level_count], 0, level_count);
*/
	while (*ptr_package_names[level_count] != NULL)
	{
	    fprintf (stdout, "%s\n", *ptr_package_names[level_count]);
	    ++ptr_package_names[level_count];
	}
    }
    return (0);
}   /*  End Function main  */
