/*  main_xt.c

    Main file for  koords  (X11 co-ordinate generator tool for Karma).

    Copyright (C) 1996-2003  Richard Gooch

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Richard Gooch may be reached by email at  karma-request@atnf.csiro.au
    The postal address is:
      Richard Gooch, c/o ATNF, P. O. Box 76, Epping, N.S.W., 2121, Australia.
*/

/*
    This Karma module will enable interactive generation of an astronomical
    co-ordinate system on a target image, using a reference image.
    This module runs on an X11 server.


    Written by      Richard Gooch   14-OCT-1996: Copied from kshell/main_xt.c

    Updated by      Richard Gooch   17-OCT-1996: Updated pick sequencing.

    Updated by      Richard Gooch   28-OCT-1996: Added hostname and port number
  to title.

    Updated by      Richard Gooch   3-NOV-1996: Made use of new
  <viewimg_track _compute> routine.

    Updated by      Richard Gooch   1-DEC-1996: Made use of
  <xtmisc_init_app_initialise>.

    Updated by      Richard Gooch   27-DEC-1996: Made use of <ds_event_*>
  system. Switched to <dmx> package. Took account of change to
  <xtmisc_init_app_initialise>.

    Updated by      Richard Gooch   30-DEC-1996: Took account of another change
  to <xtmisc_init_app_initialise>.

    Updated by      Richard Gooch   31-DEC-1996: Took account of change to
  to Dataclip widget.

    Updated by      Richard Gooch   12-JAN-1997: Made use of new
  XkwNtrapDirectoryDatasets resource for FilewinWidgetClass.

    Updated by      Richard Gooch   2-FEB-1997: No longer register pointer to
  KwcsAstro object with canvas.

    Updated by      Richard Gooch   12-FEB-1997: Only use GLISH "image" and
  "cube" events.

    Updated by      Richard Gooch   9-MAR-1997: Took account of change to
  <xtmisc_init_app_initialise>.

    Updated by      Richard Gooch   26-MAR-1997: Track display now also updates
  when dragging zoom box.

    Updated by      Richard Gooch   19-APR-1997: Now use internal file loader
  in Filewinpopup widget.

    Updated by      Richard Gooch   16-MAY-1997: Turn off integer X and Y zoom
  by default.

    Updated by      Richard Gooch   18-MAY-1997: Support TrueColour canvas.

    Updated by      Richard Gooch   10-JUN-1997: Remove existing co-ordinate
  system from target if required.

    Updated by      Richard Gooch   22-JUN-1997: Took account of change to
  <xtmisc_init_app_initialise> to double varargs list.

    Updated by      Richard Gooch   14-AUG-1997: Use cursorName resource
  rather than manually fiddling with cursor.

    Updated by      Richard Gooch   31-OCT-1997: Took account of change to
  <XkwDataclipNewArray>.

    Updated by      Richard Gooch   7-DEC-1997: Added font resource for new
  intensity menu.

    Updated by      Richard Gooch   14-FEB-1998: Added showCrosshair resource
  for MagnifierPopup.

    Updated by      Richard Gooch   15-FEB-1998: Added direct user control of
  capture circle.

    Updated by      Richard Gooch   1-MAR-1998: Made use of new filewinWidget
  resource.

    Updated by      Richard Gooch   31-MAY-1998: Support Galactic co-ordinates.

    Updated by      Richard Gooch   12-AUG-1998: Update cursor position on all
  canvas events.

    Updated by      Richard Gooch   2-SEP-1998: Removed overlay code.

    Updated by      Richard Gooch   5-SEP-1998: Used new comms. setup.

    Updated by      Richard Gooch   17-SEP-1998: Use SimpleMenu class name in
  resources.

    Updated by      Richard Gooch   2-FEB-1999: Fixed bug in <undo_cbk> where
  number of target points was not set to number of reference points. Make
  apply and target export buttons sensitive/insensitive as appropriate.

    Updated by      Richard Gooch   30-AUG-1999: Added dirname font resource.

    Updated by      Richard Gooch   31-AUG-1999: Took account of interface
  change to function <viewimg_track _compute>. Added 'l' key binding in
  function <track_ref_event>.

    Updated by      Richard Gooch   9-SEP-1999: Removed track display code and
  made use of XkwNupdateTrackLabels resource for ImageDisplay widget.

    Updated by      Richard Gooch   11-DEC-1999: Removed unnecessary externs.

    Updated by      Richard Gooch   22-JUN-2000: Removed unnecessary #include's

    Updated by      Richard Gooch   12-JUL-2000: Added #include <karma_kwin.h>.

    Updated by      Richard Gooch   11-MAY-2001: Changed dirname font resource.

    Updated by      Richard Gooch   3-OCT-2001: Added translations for Filewin.

    Updated by      Richard Gooch   21-DEC-2001: Included k_version.h directly.

    Updated by      Richard Gooch   22-MAY-2002: Set initial visible canvas to
  first available of {PseudoColour, TrueColour}.

    Updated by      Richard Gooch   22-DEC-2002: Added #include <stdlib.h>

    Updated by      Richard Gooch   14-JUN-2003: #include cleanups.

    Last updated by Richard Gooch   3-OCT-2003: Added #include <Xkw/Dataclip.h>


*/
#include <stdlib.h>
#include <stdio.h>
#include <X11/Xatom.h>
#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/Xaw/Command.h>
#include <X11/Xaw/Form.h>
#include <X11/Xaw/Dialog.h>
#include <karma.h>
#include <k_version.h>
#define X11
#include <k_event_codes.h>
#include <karma_viewimg.h>
#include <karma_xtmisc.h>
#include <karma_iarray.h>
#include <karma_kwin.h>
#include <karma_wcs.h>
#include <karma_ds.h>
#include <karma_im.h>
#include <karma_hi.h>
#include <karma_xc.h>
#include <karma_ic.h>
#include <karma_a.h>
#include <karma_c.h>
#include <Xkw/ImageDisplay.h>
#include <Xkw/Filewin.h>
#include <Xkw/Dataclip.h>
#ifdef OS_SunOS
#  include <floatingpoint.h>
#endif
#include "koords.h"


#define VERSION "1.6.13"


/*  Private functions  */
STATIC_FUNCTION (void process_and_setup,
		 (multi_array *multi_desc, CONST char *filename,
		  iarray *array, KwcsAstro *ap, flag reference,
		  ViewableImage *image, ViewableImage *magnified_image,
		  ViewableImage *panner_image,
		  Widget shell, Widget image_display) );
STATIC_FUNCTION (flag track_ref_event,
		 (ViewableImage vimage, double x, double y,
		  void *value, unsigned int event_code,
		  void *e_info, void **f_info,
		  double x_lin, double y_lin, unsigned int value_type,
		  double x_im, double y_im, int x_pix, int y_pix) );
STATIC_FUNCTION (flag track_tar_event,
		 (ViewableImage vimage, double x, double y,
		  void *value, unsigned int event_code,
		  void *e_info, void **f_info,
		  double x_lin, double y_lin, unsigned int value_type) );
STATIC_FUNCTION (void undo_cbk, (Widget w, XtPointer client_data,
				 XtPointer call_data) );
STATIC_FUNCTION (flag reference_data_func,
		 (void *object, multi_array *multi_desc,
		  CONST char *domain, CONST char *name) );
STATIC_FUNCTION (flag target_data_func,
		 (void *object, multi_array *multi_desc,
		  CONST char *domain, CONST char *name) );


/*  Private data  */
String fallback_resources[] =
{
    "Koords.geometry:                                 +0+0",
    "Koords*multiCanvas.Canvas*cursorName:            target",
    "Koords*pseudoColourCanvas*background:            black",
    "Koords*directColourCanvas*background:            black",
    "Koords*trueColourCanvas*background:              black",
    "Koords*pseudoColourStereoCanvas*background:      black",
    "Koords*directColourStereoCanvas*background:      black",
    "Koords*trueColourStereoCanvas*background:        black",
    "Koords*pseudoColourCanvas*foreground:            white",
    "Koords*Command*background:                       grey70",
    "Koords*Repeater*background:                      grey70",
    "Koords*Ktoggle*background:                       grey80",
    "Koords*closeButton*background:                   grey90",
    "Koords*ChoiceMenu.background:                    turquoise",
    "Koords*ExclusiveMenu.background:                 turquoise",
    "Koords*Value*background:                         #d0a0a0",
    "Koords*ImageDisplay*quit*background:             orange",
    "Koords*ScrollableList.highlightPixel:            Sea Green",
    "Koords*SimpleSlider.foreground:                  Sea Green",
    "Koords*SimpleProgressMeter.background:           turquoise3",
    "Koords*SimpleProgressMeter.foreground:           grey60",
    "Koords*tarShell.geometry:                        -0+0",
    "Koords*ImageDisplay*trackLabel0*font:            8x13bold",
    "Koords*ImageDisplay*trackLabel1*font:            8x13bold",
    "Koords*ImageDisplay*trackLabel2*font:            8x13bold",
    "Koords*ImageDisplay*trackLabel3*font:            8x13bold",
    "Koords*ExclusiveMenu.font:                       8x13bold",
    "Koords*SimpleMenu*font:                          10x20",
    "Koords*ImageDisplay.cmapSize:                    100",
    "Koords*ImageDisplay.Command.font:                7x13bold",
    "Koords*ImageDisplay.applicationBox.Command.font: 7x13bold",
    "Koords*ImageDisplay.applicationBox.Ktoggle.font: 7x13bold",
    "Koords*ImageDisplay.ChoiceMenu.font:             7x13bold",
    "Koords*ImageDisplay.exportMenu.menuButton.font:  7x13bold",
    "Koords*ImageDisplay.overlayMenu.menuButton.font: 7x13bold",
    "Koords*ImageDisplay*zoomMenu*Unzoom*foreground:  red",
    "Koords*Dataclip.AutoValueScale:                  False",
    "Koords*SimpleSlider.borderWidth:                 0",
    "Koords*IntensityPolicy*autoIntensityScale:       False",
    "Koords*ZoomPolicy*integerXZoom:                  False",
    "Koords*ZoomPolicy*integerYZoom:                  False",
    "Koords*MagnifierPopup.showCrosshair:             True",
    "Koords*Filewin.dirname.textSink.font:            6x13",
    "Koords*font:                                     9x15bold",
    "Koords*borderColor:                              black",
    "Koords*background:                               aquamarine",
    "Koords*foreground:                               black",
    NULL
};
static XrmOptionDescRec Options[] =
{
    {"-private_cmap", ".topForm.multiCanvas.pseudoColourCanvas.forceNewCmap",
     XrmoptionNoArg, (XPointer) "True"},
    {"-num_colours", ".topForm.cmapSize", XrmoptionSepArg, (XPointer) NULL},
    {"-fullscreen", ".topForm.fullscreen", XrmoptionNoArg, (XPointer) "True"},
    {"-verbose", "*verbose", XrmoptionNoArg, (XPointer) "True"},
};
static Widget main_shell = NULL;
static Widget ref_image_display = NULL;
static KWorldCanvas ref_canvas = NULL;
static Widget tar_shell = NULL;
static Widget tar_image_display = NULL;
static KWorldCanvas tar_canvas = NULL;
static Widget apply_btn = NULL;


/*  Public functions follow  */

int main (int argc, char **argv)
{
    KWorldCanvas pc_canvas, true_canvas, wc;
    KCallbackFunc cf;
    XtAppContext app_context;
    Widget filewin, filepopup, w, box, magnifier, tar_export_menu;
    Widget ref_undo_btn, tar_undo_btn, compute_btn;
    extern dataset_t reference, target;
    extern char title_name[STRING_LENGTH];
    static char function_name[] = "main";

#ifdef SIGFPE_ABORT
	ieee_handler ("set", "division", SIGFPE_ABORT);
#endif
    /*  Initialise module  */
    im_register_module_name ("koords");
    im_register_module_version_date (VERSION);
    im_register_lib_version (KARMA_VERSION);
    /*  Start up Xt  */
    main_shell = xtmisc_init_app_initialise (&app_context, "Koords",
					     Options, XtNumber (Options),
					     &argc, argv, fallback_resources,
					     XTMISC_INIT_ATT_MIN_CCELLS, 100,
					     XTMISC_INIT_ATT_COMMS_SETUP, TRUE,
					     XTMISC_INIT_ATT_BECOME_SERVER,
					     TRUE,
					     XTMISC_INIT_ATT_TITLE, title_name,
					     XTMISC_INIT_ATT_END,
					     NULL);
    /*  Setup the reference window  */
    ref_image_display = XtVaCreateManagedWidget ("topForm",
						 imageDisplayWidgetClass,
						 main_shell,
						 XtNborderWidth, 0,
						 XkwNnumTrackLabels, 3,
						 XkwNupdateTrackLabels, TRUE,
						 NULL);
    /*  Continue setting up the reference window  */
    xtmisc_set_icon (main_shell, ic_write_kimage_icon);
    if ( ( w = XtNameToWidget (ref_image_display, "filesButton") ) != NULL )
    {
	/*  Force main "Files" button to be re-labelled as "Reference Images"
	 */
	XtVaSetValues (w,
		       XtNlabel, "Ref. Image",
		       NULL);
    }
    filepopup = XtNameToWidget (ref_image_display, "filewinPopup");
    XtVaGetValues (filepopup, XkwNfilewinWidget, &filewin, NULL);
    XtVaSetValues (filepopup,
		   XkwNfilenameTester, XkwFilewinStandardFileTester_nD,
		   XkwNdomainName, "REFERENCE",
		   NULL);
    XtVaSetValues (filewin,
		   XkwNtrapDirectoryDatasets, True,
		   NULL);
    ds_event_register_func (reference_data_func, NULL);
    /*  Get the application box  */
    box = XtNameToWidget (ref_image_display, "applicationBox");
    w = XtVaCreateManagedWidget ("button", commandWidgetClass, box,
				 XtNlabel, "Undo Pair",
				 NULL);
    ref_undo_btn = w;
    XtAddCallback (w, XtNcallback, undo_cbk, NULL);
    XtRealizeWidget (main_shell);
    XtVaGetValues (ref_image_display,
		   XkwNpseudoColourCanvas, &pc_canvas,
		   XkwNtrueColourCanvas, &true_canvas,
		   XkwNmagnifier, &magnifier,
		   NULL);
    ref_canvas = (pc_canvas == NULL) ? true_canvas : pc_canvas;
    if (ref_canvas == NULL)
    {
	fprintf (stderr, "PseudoColour or TrueColour canvas not found!\n");
	exit (RV_UNDEF_ERROR);
    }
    cf = viewimg_register_position_event_func (ref_canvas,
					       ( flag (*) () ) track_ref_event,
					       (void *) ref_image_display);
    c_move_callback (cf, FALSE);
    viewimg_register_position_event_func (ref_canvas,
					  ( flag (*) () ) reference_event_func,
					  NULL);
    canvas_register_refresh_func (ref_canvas, main_refresh_func, &reference);
    XtVaSetValues (ref_image_display, XkwNvisibleCanvas, ref_canvas, NULL);
    /*  Set up the reference magnifier window  */
    XtVaGetValues (magnifier,
		   XkwNpseudoColourCanvas, &pc_canvas,
		   XkwNtrueColourCanvas, &true_canvas,
		   NULL);
    wc = (pc_canvas == NULL) ? true_canvas : pc_canvas;
    reference.mag_canvas = canvas_get_pixcanvas (wc);
    viewimg_register_position_event_func (wc,
					  ( flag (*) () ) mag_event_func,
					  &reference);
    canvas_register_refresh_func (wc, mag_refresh_func, &reference);
    /*  Setup the target window  */
    tar_shell = XtVaCreatePopupShell ("tarShell", topLevelShellWidgetClass,
				      main_shell,
				      XtNtitle, "Slice Window",
				      NULL);
    tar_image_display = XtVaCreateManagedWidget ("topForm",
						 imageDisplayWidgetClass,
						 tar_shell,
						 XtNborderWidth, 0,
                                                 XkwNnumTrackLabels, 2,
						 XkwNshowQuitButton, FALSE,
						 NULL);
    /*  Continue setting up the target window  */
    if ( ( w = XtNameToWidget (tar_image_display, "filesButton") ) != NULL )
    {
	/*  Force target "Files" button to be re-labelled as "Target Images" */
	XtVaSetValues (w,
		       XtNlabel, "Target Image",
		       NULL);
    }
    filepopup = XtNameToWidget (tar_image_display, "filewinPopup");
    XtVaGetValues (filepopup, XkwNfilewinWidget, &filewin, NULL);
    XtVaSetValues (filepopup,
		   XkwNfilenameTester, XkwFilewinStandardFileTester_nD,
		   XkwNdomainName, "TARGET",
		   NULL);
    XtVaSetValues (filewin,
		   XkwNtrapDirectoryDatasets, True,
		   NULL);
    ds_event_register_func (target_data_func, NULL);
    /*  Get application box of target window */
    box = XtNameToWidget (tar_image_display, "applicationBox");
    w = XtVaCreateManagedWidget ("button", commandWidgetClass, box,
				 XtNlabel, "Undo Pair",
				 NULL);
    tar_undo_btn = w;
    XtAddCallback (w, XtNcallback, undo_cbk, NULL);
    w = XtVaCreateManagedWidget ("button", commandWidgetClass, box,
				 XtNlabel, "Compute Coord. System",
				 NULL);
    compute_btn = w;
    XtAddCallback (w, XtNcallback, compute_and_store, NULL);
    w = XtVaCreateManagedWidget ("button", commandWidgetClass, box,
				 XtNlabel, "Apply Coord. System",
				 XtNsensitive, False,
				 NULL);
    apply_btn = w;
    XtAddCallback (w, XtNcallback, apply_coordinates, NULL);
    XtAddCallback (ref_undo_btn, XtNcallback, xtmisc_insensitive_cbk,
		   apply_btn);
    XtAddCallback (tar_undo_btn, XtNcallback, xtmisc_insensitive_cbk,
		   apply_btn);
    XtAddCallback (compute_btn, XtNcallback, xtmisc_sensitive_cbk, apply_btn);
    /*  Extra work for the target export menu button  */
    if ( ( tar_export_menu = XtNameToWidget (tar_image_display, "exportMenu") )
	 == NULL )
    {
	fprintf (stderr, "Cannot find exportMenu widget for target window\n");
	a_prog_bug (function_name);
    }
    XtSetSensitive (tar_export_menu, False);
    XtAddCallback (ref_undo_btn, XtNcallback, xtmisc_insensitive_cbk,
		   tar_export_menu);
    XtAddCallback (tar_undo_btn, XtNcallback, xtmisc_insensitive_cbk,
		   tar_export_menu);
    XtAddCallback (compute_btn, XtNcallback, xtmisc_insensitive_cbk,
		   tar_export_menu);
    XtAddCallback (apply_btn, XtNcallback, xtmisc_sensitive_cbk,
		   tar_export_menu);
    /*  Realise and set up canvas  */
    XtRealizeWidget (tar_shell);
    XtVaGetValues (tar_image_display,
		   XkwNpseudoColourCanvas, &pc_canvas,
		   XkwNtrueColourCanvas, &true_canvas,
		   XkwNmagnifier, &magnifier,
		   NULL);
    tar_canvas = (pc_canvas == NULL) ? true_canvas : pc_canvas;
    if (tar_canvas == NULL)
    {
	fprintf (stderr, "PseudoColour canvas not found!\n");
	exit (RV_UNDEF_ERROR);
    }
    cf = viewimg_register_position_event_func (tar_canvas,
					       ( flag (*) () ) track_tar_event,
					       (void *) tar_image_display);
    c_move_callback (cf, FALSE);
    viewimg_register_position_event_func (tar_canvas,
					  ( flag (*) () ) target_event_func,
					  (void *) ref_canvas);
    canvas_register_refresh_func (tar_canvas, main_refresh_func, &target);
    XtVaSetValues (tar_image_display, XkwNvisibleCanvas, tar_canvas, NULL);
    /*  Set up the target magnifier window  */
    XtVaGetValues (magnifier,
		   XkwNpseudoColourCanvas, &pc_canvas,
		   XkwNtrueColourCanvas, &true_canvas,
		   NULL);
    wc = (pc_canvas == NULL) ? true_canvas : pc_canvas;
    target.mag_canvas = canvas_get_pixcanvas (wc);
    viewimg_register_position_event_func (wc,
					  ( flag (*) () ) mag_event_func,
					  &target);
    canvas_register_refresh_func (wc, mag_refresh_func, &target);
    XtPopup (tar_shell, XtGrabNone);
    XtAppMainLoop (app_context);
    return (RV_OK);
}   /*  End Function main   */

void ring_bell ()
/*  [SUMMARY] Ring the bell.
    [RETURNS] Nothing.
*/
{
    XBell (XtDisplay (ref_image_display), 100);
}   /*  End Function ring_bell  */


/*  Private routines follow  */

static void process_and_setup (multi_array *multi_desc, CONST char *filename,
			       iarray *array, KwcsAstro *ap,
			       flag reference, ViewableImage *image,
			       ViewableImage *magnified_image,
			       ViewableImage *panner_image,
			       Widget shell, Widget image_display)
{
    KWorldCanvas canvas, mag_canvas, pan_canvas;
    double min, max;
    Widget izoomwinpopup, magnifier, panner;
    char *ptr;
    char stripped_filename[STRING_LENGTH];
    char title[STRING_LENGTH];
    extern char title_name[STRING_LENGTH];

    strcpy (stripped_filename, filename);
    if ( ( ptr = strrchr (stripped_filename, '.') ) != NULL )
    {
	if (strcmp (ptr, ".kf") == 0) *ptr = '\0';
    }
    izoomwinpopup = XtNameToWidget (image_display, "izoomwinpopup");
    XtVaGetValues (image_display,
		   XkwNpseudoColourCanvas, &canvas,
		   XkwNmagnifier, &magnifier,
		   XkwNpanner, &panner,
		   NULL);
    XtVaGetValues (magnifier, XkwNpseudoColourCanvas, &mag_canvas, NULL);
    XtVaGetValues (panner, XkwNpseudoColourCanvas, &pan_canvas, NULL);
    if (canvas == NULL)
    {
	XtVaGetValues (image_display,
		       XkwNtrueColourCanvas, &canvas,
		       NULL);
	XtVaGetValues (magnifier, XkwNtrueColourCanvas, &mag_canvas, NULL);
	XtVaGetValues (panner, XkwNtrueColourCanvas, &pan_canvas, NULL);
    }
    if ( !process_image (multi_desc, array, ap, reference,
			 canvas, mag_canvas, pan_canvas,
			 image, magnified_image, panner_image,
			 &min, &max) ) return;
    sprintf (title, "%s  file: %s\n", title_name, filename);
    XtVaSetValues (shell,
		   XtNtitle, title,
		   NULL);
    XtVaSetValues (image_display,
		   XkwNimageName, stripped_filename,
		   NULL);
    XkwDataclipNewArray (izoomwinpopup, *array, min, max, TRUE);
    XtVaSetValues (image_display,
		   XkwNvisibleCanvas, canvas,
		   NULL);
}   /*  End Function process_and_setup  */

static flag track_ref_event (ViewableImage vimage, double x, double y,
			     void *value, unsigned int event_code,
			     void *e_info, void **f_info,
			     double x_lin, double y_lin,
			     unsigned int value_type,
			     double x_im, double y_im, int x_pix, int y_pix)
/*  [PURPOSE] This routine is a position event consumer for a world canvas
    which has a number of ViewableImage objects associated with it.
    <viewimg> The active viewable image.
    <x> The horizontal world co-ordinate of the event.
    <y> The vertical world co-ordinate of the event.
    These values will have been transformed by the registered transform
    function (see  canvas_register_transform_func  ) for the associated
    world canvas.
    <value> A pointer to the data value in the viewable image corresponding
    to the event co-ordinates.
    <event_code> The arbitrary event code.
    <e_info> The arbitrary event information.
    <f_info> The arbitrary function information pointer.
    <x_lin> The linear horizontal world co-ordinate (the co-ordinate prior
    to the transform function being called).
    <y_lin> The linear vertical world co-ordinate (the co-ordinate prior
    to the transform function being called).
    <value_type> The type of the data value. This may be K_DCOMPLEX or
    K_UB_RGB.
    <x_im> The horizontal image pixel co-ordinate. 0.0 is the pixel centre.
    <y_im> The vertical image pixel co-ordinate. 0.0 is the pixel centre.
    <x_pix> The horizontal screen pixel co-ordinate.
    <y_pix> The vertical screen pixel co-ordinate.
    [RETURNS] TRUE if the event was consumed, else FALSE indicating that
    the event is still to be processed.
*/
{
    extern dataset_t reference;
    /*static char function_name[] = "track_ref_event";*/

    reference.x_lin = x_lin;
    reference.y_lin = y_lin;
    return (FALSE);
}   /*  End Function track_ref_event  */

static flag track_tar_event (ViewableImage vimage, double x, double y,
			     void *value, unsigned int event_code,
			     void *e_info, void **f_info,
			     double x_lin, double y_lin,
			     unsigned int value_type)
/*  [PURPOSE] This routine is a position event consumer for a world canvas
    which has a number of ViewableImage objects associated with it.
    <viewimg> The active viewable image.
    <x> The horizontal world co-ordinate of the event.
    <y> The vertical world co-ordinate of the event.
    These values will have been transformed by the registered transform
    function (see  canvas_register_transform_func  ) for the associated
    world canvas.
    <value> A pointer to the data value in the viewable image corresponding
    to the event co-ordinates.
    <event_code> The arbitrary event code.
    <e_info> The arbitrary event information.
    <f_info> The arbitrary function information pointer.
    <x_lin> The linear horizontal world co-ordinate (the co-ordinate prior
    to the transform function being called).
    <y_lin> The linear vertical world co-ordinate (the co-ordinate prior
    to the transform function being called).
    <value_type> The type of the data value. This may be K_DCOMPLEX or
    K_UB_RGB.
    [RETURNS] TRUE if the event was consumed, else FALSE indicating that
    the event is still to be processed.
*/
{
    KWorldCanvas canvas, magnifier_canvas;
    flag galactic;
    unsigned int hdim, vdim, rdim;
    unsigned int num_restr, count;
    unsigned long pointer_x_index, pointer_y_index, r_index;
    Widget image_display = (Widget) *f_info;
    Widget first_track_label, second_track_label, magnifier;
    unsigned char *rgb_ptr = (unsigned char *) value;
    char *xlabel, *ylabel;
    char **restr_names;
    double *restr_values;
    array_desc *arr_desc;
    char txt[STRING_LENGTH];
    char value_string[STRING_LENGTH], index_string[STRING_LENGTH];
    char world_string[STRING_LENGTH], extra_string[STRING_LENGTH];
    extern dataset_t reference, target;
    static char function_name[] = "track_tar_event";

    /*  Create value string  */
    if (target.array == NULL) return (TRUE);
    switch (value_type)
    {
      case K_DCOMPLEX:
	/*  Compute value  */
	iarray_format_value (target.array, value_string,
			     *(double *) value, TOOBIG, TOOBIG);
	break;
      case K_UB_RGB:
	sprintf (value_string, "RGB: %u %u %u",
		 rgb_ptr[0], rgb_ptr[1], rgb_ptr[2]);
	break;
      default:
	fprintf (stderr, "Illegal type: %u\n", value_type);
	a_prog_bug (function_name);
	break;
    }
    canvas = viewimg_get_worldcanvas (vimage);
    canvas_get_specification (canvas, &xlabel, &ylabel, &num_restr,
			      &restr_names, &restr_values);
    viewimg_get_attributes (vimage,
			    VIEWIMG_VATT_ARRAY_DESC, &arr_desc,
			    VIEWIMG_VATT_HDIM, &hdim,
			    VIEWIMG_VATT_VDIM, &vdim,
			    VIEWIMG_VATT_END);
    /*  Convert linear world co-ordinates to array indices and display  */
    pointer_x_index = ds_get_coord_num (arr_desc->dimensions[hdim], x_lin,
					SEARCH_BIAS_CLOSEST);
    pointer_y_index = ds_get_coord_num (arr_desc->dimensions[vdim], y_lin,
					SEARCH_BIAS_CLOSEST);
    sprintf (index_string, "x: %lu  y: %lu  ",
	     pointer_x_index, pointer_y_index);
    /*  Add any restriction information  */
    for (count = 0; count < num_restr; ++count)
    {
	if ( ( rdim = ds_f_dim_in_array (arr_desc, restr_names[count]) )
	     >= arr_desc->num_dimensions ) continue;
	r_index = ds_get_coord_num (arr_desc->dimensions[rdim],
				    restr_values[count], SEARCH_BIAS_CLOSEST);
	sprintf (txt, "z%u: %lu  ", rdim, r_index);
	strcat (index_string, txt);
    }
    strcat (index_string, value_string);
    first_track_label = XtNameToWidget (image_display, "trackLabel0");
    second_track_label = XtNameToWidget (image_display, "trackLabel1");
    XtVaSetValues (first_track_label, XtNlabel, index_string, NULL);
    /*  Now display the world co-ordinate information  */
    if (target.ap == NULL)
    {
	sprintf (world_string, "%5e %s  %5e %s  ", x, xlabel, y, ylabel);
	/*  Add any restriction information  */
	for (count = 0; count < num_restr; ++count)
	{
	    sprintf (txt, "%5e %s  ", restr_values[count], restr_names[count]);
	    strcat (world_string, txt);
	}
    }
    else
    {
	galactic = wcs_astro_test_if_glon ( reference.ap,
					    iarray_dim_name (reference.array,
							     1) );
	wcs_astro_format_all (target.ap, world_string,
			      galactic ? "GLON-ARC" : "RA---ARC", x_lin,
			      galactic ? "GLAT-ARC" : "DEC--ARC", y_lin,
			      NULL, 0.0,
			      num_restr, (CONST char **) restr_names,
			      restr_values, extra_string);
    }
    XtVaSetValues (second_track_label, XtNlabel, world_string, NULL);
    XtVaGetValues (image_display, XkwNmagnifier, &magnifier, NULL);
    XtVaGetValues (magnifier, XkwNvisibleCanvas, &magnifier_canvas, NULL);
    target.x_lin = x_lin;
    target.y_lin = y_lin;
    return (FALSE);
}   /*  End Function track_tar_event  */

static void undo_cbk (Widget w, XtPointer client_data, XtPointer call_data)
{
    extern flag last_click_was_reference;
    extern dataset_t reference, target;

    last_click_was_reference = FALSE;
    if (reference.num_points < 1)
    {
	XBell (XtDisplay (w), 100);
	return;
    }
    --reference.num_points;
    target.num_points = reference.num_points;
    kwin_resize (canvas_get_pixcanvas (ref_canvas), FALSE, 0, 0, 0, 0);
    kwin_resize (canvas_get_pixcanvas (tar_canvas), FALSE, 0, 0, 0, 0);
    if (target.ap != NULL) wcs_astro_destroy (target.ap);
    target.ap = NULL;
}   /*  End Function undo_cbk  */

static flag reference_data_func (void *object, multi_array *multi_desc,
				 CONST char *domain, CONST char *name)
/*  [SUMMARY] Process a new reference image event.
    <object> The object information pointer.
    <multi_desc> The multi_array descriptor.
    <domain> The domain from where the event was generated.
    <name> The name of the event in its domain.
    [RETURNS] TRUE if further callbacks should not be called, else FALSE.
*/
{
    char filename[STRING_LENGTH];
    extern dataset_t reference;
    static ViewableImage ref_vimage = NULL;
    static ViewableImage ref_magnified_vimage = NULL;
    static ViewableImage ref_panner_vimage = NULL;
    /*static char function_name[] = "reference_data_func";*/

    if (strcmp (domain, "TARGET") == 0) return (FALSE);
    if ( (strcmp (domain, "GLISH") == 0) && (strcmp (name, "reference") != 0) )
	return (FALSE);
    if (strcmp (domain, "REFERENCE") == 0) filename[0] = '\0';
    else
    {
	strcpy (filename, domain);
	strcat (filename, "::");
    }
    strcat (filename, name);
    process_and_setup (multi_desc, filename, &reference.array, &reference.ap,
		       TRUE,
		       &ref_vimage, &ref_magnified_vimage, &ref_panner_vimage,
		       main_shell, ref_image_display);
    return (TRUE);
}   /*  End Function reference_data_func  */

static flag target_data_func (void *object, multi_array *multi_desc,
			      CONST char *domain, CONST char *name)
/*  [SUMMARY] Process a new reference image event.
    <object> The object information pointer.
    <multi_desc> The multi_array descriptor.
    <domain> The domain from where the event was generated.
    <name> The name of the event in its domain.
    [RETURNS] TRUE if further callbacks should not be called, else FALSE.
*/
{
    char filename[STRING_LENGTH];
    extern KwcsAstro target_original_ap;
    extern dataset_t target;
    static ViewableImage tar_vimage = NULL;
    static ViewableImage tar_magnified_vimage = NULL;
    static ViewableImage tar_panner_vimage = NULL;
    /*static char function_name[] = "target_data_func";*/

    if (strcmp (domain, "REFERENCE") == 0) return (FALSE);
    if ( (strcmp (domain, "GLISH") == 0) && (strcmp (name, "target") != 0) )
	return (FALSE);
    if (strcmp (domain, "TARGET") == 0) filename[0] = '\0';
    else
    {
	strcpy (filename, domain);
	strcat (filename, "::");
    }
    strcat (filename, name);
    process_and_setup (multi_desc, filename, &target.array,&target_original_ap,
		       FALSE,
		       &tar_vimage, &tar_magnified_vimage, &tar_panner_vimage,
		       tar_shell, tar_image_display);
    if (target_original_ap != NULL)
    {
	fprintf (stderr,
		 "WARNING: Target image already has co-ordinate system. New ");
	fprintf (stderr, "co-ordinate system\nwill overwrite old one.\n");
    }
    return (TRUE);
}   /*  End Function target_data_func  */
