/*  main_xt.c

    Main file for  kpvslice  (X11 Position-Velocity slice tool for Karma).

    Copyright (C) 1996-2004  Richard Gooch

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Richard Gooch may be reached by email at  karma-request@atnf.csiro.au
    The postal address is:
      Richard Gooch, c/o ATNF, P. O. Box 76, Epping, N.S.W., 2121, Australia.
*/

/*
    This Karma module will enable interactive selection of a Position-Velocity
    slice through a data cube.
    This module runs on an X11 server.


    Written by      Richard Gooch   12-JUN-1996: Copied from kview module.

    Updated by      Richard Gooch   16-JUN-1996: Added check for *__ SCALE and
  *__ OFFSET data.

    Updated by      Richard Gooch   19-JUN-1996: Made use of
  <iarray_format_value> routine.

    Updated by      Richard Gooch   22-JUN-1996: Made use of
  <XkwFilewinStandardFileTester_nD> and <XkwFilewinStandardFileTester_3D>
  routines.

    Updated by      Richard Gooch   27-JUN-1996: Switched to
  <canvas_use_astro_transform>

    Updated by      Richard Gooch   1-AUG-1996: Created separate directory
  select callbacks for cubes and images, otherwise Miriad cubes would be sent
  to the image loader.

    Updated by      Richard Gooch   10-AUG-1996: Added trace and main refresh
  functions.

    Updated by      Richard Gooch   11-AUG-1996: Computed 1st moment map.

    Updated by      Richard Gooch   13-AUG-1996: Added image mode menu.

    Updated by      Richard Gooch   14-AUG-1996: Added moment control popup.

    Updated by      Richard Gooch   15-AUG-1996: Added version number to title.

    Updated by      Richard Gooch   22-AUG-1996: Made code more robust to
  images/cubes/moment maps having 0 data range.

    Updated by      Richard Gooch   31-AUG-1996: Only compute moments in
  response to "Apply Parameters" button. Show cube min and max in moment
  control window.

    Updated by      Richard Gooch   3-SEP-1996: Restored main canvas magnifier
  which did not display under some conditions.

    Updated by      Richard Gooch   6-SEP-1996: Added median algorithm.

    Updated by      Richard Gooch   11-SEP-1996: Added "Show Trace" toggle.

    Updated by      Richard Gooch   14-SEP-1996: Made use of MomentGenerator
  widget.

    Updated by      Richard Gooch   16-SEP-1996: Made use of TracePopup
  widget.

    Updated by      Richard Gooch   19-SEP-1996: Changed to special function
  call interface for setting new array for MomentGenerator and TracePopup.

    Updated by      Richard Gooch   1-OCT-1996: Fixed resource names for new
  ExportMenu widget. Added "0th channel" image mode.

    Updated by      Richard Gooch   28-OCT-1996: Added hostname and port number
  to title.

    Updated by      Richard Gooch   3-NOV-1996: Made use of new
  <viewimg_track _compute> routine.

    Updated by      Richard Gooch   10-NOV-1996: Made use of new attributes for
  ViewableImage objects to override the canvas intensity range.

    Updated by      Richard Gooch   11-NOV-1996: Only refresh magnifier canvas
  if visible.

    Updated by      Richard Gooch   14-NOV-1996: Trap when displaying
  ViewableImage for moment maps not yet computed.

    Updated by      Richard Gooch   1-DEC-1996: Made use of
  <xtmisc_init_app_initialise>.

    Updated by      Richard Gooch   27-DEC-1996: Made use of <ds_event_*>
  system. Switched to <dmx> package. Took account of change to
  <xtmisc_init_app_initialise>.

    Updated by      Richard Gooch   30-DEC-1996: Took account of another change
  to <xtmisc_init_app_initialise>.

    Updated by      Richard Gooch   31-DEC-1996: Took account of change to
  to Dataclip widget.

    Updated by      Richard Gooch   12-JAN-1997: Made use of new
  XkwNtrapDirectoryDatasets resource for FilewinWidgetClass.

    Updated by      Richard Gooch   27-JAN-1997: Took account of change to
  interface for <XkwTracePopupShowTrace> function.

    Updated by      Richard Gooch   28-JAN-1997: Took account of change back to
  previous interface for <XkwTracePopupShowTrace> function.

    Updated by      Richard Gooch   2-FEB-1997: No longer register pointer to
  KwcsAstro object with canvas.

    Updated by      Richard Gooch   12-FEB-1997: Only use GLISH "image" and
  "cube" events.

    Updated by      Richard Gooch   15-FEB-1997: No longer use global KwcsAstro
  objects: get from data, canvas or viewable image.

    Updated by      Richard Gooch   9-MAR-1997: Took account of change to
  <xtmisc_init_app_initialise>.

    Updated by      Richard Gooch   26-MAR-1997: Track display now also updates
  when dragging zoom box.

    Updated by      Richard Gooch   19-APR-1997: Now use internal file loader
  in Filewinpopup widget.

    Updated by      Richard Gooch   7-JUN-1997: Added image profile display.

    Updated by      Richard Gooch   18-JUN-1997: Removed IEEE floating point
  error handler setup.

    Updated by      Richard Gooch   19-JUN-1997: Set X units for image profile
  to be angular offset.

    Updated by      Richard Gooch   22-JUN-1997: Took account of change to
  <xtmisc_init_app_initialise> to double varargs list.

    Updated by      Richard Gooch   25-JUN-1997: Turn off integer X and Y zoom
  by default.

    Updated by      Richard Gooch   19-AUG-1997: Set resources in create call
  for second fileselector.

    Updated by      Richard Gooch   31-OCT-1997: Took account of change to
  <XkwDataclipNewArray>.

    Updated by      Richard Gooch   7-DEC-1997: Added font resource for new
  intensity menu and removed code to set pan centre for magnifier as it is now
  done automatically by the MagnifierPopup widget.

    Updated by      Richard Gooch   25-FEB-1998: Fixed bug in function
  <update_trace_profile> which failed if no image was loaded (i.e. viewing
  computed moment maps).

    Updated by      Richard Gooch   1-MAR-1998: Made use of new filewinWidget
  resource.

    Updated by      Richard Gooch   2-SEP-1998: Removed overlay code.

    Updated by      Richard Gooch   5-SEP-1998: Used new comms. setup.

    Updated by      Richard Gooch   17-SEP-1998: Use SimpleMenu class name in
  resources.

    Updated by      Richard Gooch   9-NOV-1998: Added support for TrueColour
  canvas only displays.

    Updated by      Richard Gooch   3-APR-1999: No longer set module name.
  Added "-trace_internal_colour" and "-trace_external_colour" options.

    Updated by      Richard Gooch   2-AUG-1999: Supported unloading of data.

    Updated by      Richard Gooch   12-AUG-1999: Added "-image" and "-cube"
  command-line switches.

    Updated by      Richard Gooch   30-AUG-1999: Added dirname font resource.

    Updated by      Richard Gooch   31-AUG-1999: Took account of interface
  change to function <viewimg_track _compute>. Added 'l' key binding in
  function <track_canvas_event>.

    Updated by      Richard Gooch   9-SEP-1999: Removed track display code and
  made use of XkwNupdateTrackLabels resource for ImageDisplay widget.

    Updated by      Richard Gooch   11-DEC-1999: Removed unnecessary externs.

    Updated by      Richard Gooch   24-DEC-1999: Select visible canvases at
  initialisation.

    Updated by      Richard Gooch   22-JUN-2000: Update Z profile on all events
  rather than limiting to pointer move events. Invalidate Z profile when event
  lies outside cube boundaries.

    Updated by      Richard Gooch   26-JUN-2000: Do not update Z profile on
  pointer enter events.

    Updated by      Richard Gooch   12-JUL-2000: Added #include <karma_kwin.h>.
  Removed unnecessary #includes.

    Updated by      Richard Gooch   20-MAR-2001: Moved code to slice_xt.c and
  supported multiple cubes and slice windows.

    Updated by      Richard Gooch   22-MAR-2001: No longer fix slice position.

    Updated by      Richard Gooch   11-MAY-2001: Changed dirname font resource.

    Updated by      Richard Gooch   3-OCT-2001: Added translations for Filewin.

    Updated by      Richard Gooch   6-DEC-2001: Took account of interface
  change to <wcs_astro_transform*>.

    Updated by      Richard Gooch   21-DEC-2001: Included k_version.h directly.

    Updated by      Richard Gooch   14-AUG-2002: Added #include <stdlib.h>.

    Updated by      Richard Gooch   15-AUG-2002: Supported non-sky axes in
  <update_trace_profile>.

    Updated by      Richard Gooch   14-JUN-2003: #include cleanups.

    Updated by      Richard Gooch   4-OCT-2003: Added #include <Xkw/Dataclip.h>

    Updated by      Richard Gooch   4-MAR-2004: Support loading of non-float
  image on command line. Support loading of cubes on command line.

    Last updated by Richard Gooch   15-MAY-2004: Some Font and colour changes.


*/
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <X11/Xatom.h>
#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/Xaw/Command.h>
#include <X11/Xaw/Form.h>
#include <X11/Xaw/Dialog.h>
#include <karma.h>
#define X11
#include <k_event_codes.h>
#include <karma_viewimg.h>
#include <karma_foreign.h>
#include <karma_xtmisc.h>
#include <karma_dsxfr.h>
#include <karma_ds.h>
#include <karma_im.h>
#include <karma_ic.h>
#include <karma_a.h>
#include <karma_m.h>
#include <karma_c.h>
#include <k_version.h>
#include <Xkw/ImageDisplay.h>
#include <Xkw/Dialogpopup.h>
#include <Xkw/Filewin.h>
#include <Xkw/Ktoggle.h>
#include <Xkw/ExclusiveMenu.h>
#include <Xkw/MomentGenerator.h>
#include <Xkw/TracePopup.h>
#include <Xkw/Dataclip.h>
#define NEED_IMAGE_MODE_ALTERNATIVES
#include "kpvslice.h"


#define VERSION "2.0.7"


/*  External functions  */
EXTERN_FUNCTION (CONST char *create_slice_win,
		 (Widget main_shell, KWorldCanvas main_canvas,
		  flag popup_filebrowser) );


/*  Private functions  */
STATIC_FUNCTION (void image_mode_cbk,
		 (Widget w, XtPointer client_data, XtPointer call_data) );
STATIC_FUNCTION (void show_image, (ViewableImage vimage) );
STATIC_FUNCTION (void show_mom0, () );
STATIC_FUNCTION (void show_mom1, () );
STATIC_FUNCTION (void moment_cbk,
		 (Widget w, XtPointer client_data, XtPointer call_data) );
STATIC_FUNCTION (void cubes_button_cbk,
		 (Widget w, XtPointer client_data, XtPointer call_data) );
STATIC_FUNCTION (void destroy_func, (iarray array, void *info) );
STATIC_FUNCTION (void show_channel0, () );
STATIC_FUNCTION (flag image_event_func,
		 (void *object, multi_array *multi_desc,
		  CONST char *domain, CONST char *name) );


/*  Private data  */
String fallback_resources[] =
{
    "Kpvslice.geometry:                                 +0+0",
    "Kpvslice*pseudoColourCanvas*background:            black",
    "Kpvslice*directColourCanvas*background:            black",
    "Kpvslice*trueColourCanvas*background:              black",
    "Kpvslice*pseudoColourStereoCanvas*background:      black",
    "Kpvslice*directColourStereoCanvas*background:      black",
    "Kpvslice*trueColourStereoCanvas*background:        black",
    "Kpvslice*pseudoColourCanvas*foreground:            white",
    "Kpvslice*Command*background:                       grey70",
    "Kpvslice*Repeater*background:                      grey70",
    "Kpvslice*Ktoggle*background:                       grey80",
    "Kpvslice*closeButton*background:                   grey90",
    "Kpvslice*ChoiceMenu.background:                    turquoise",
    "Kpvslice*ExclusiveMenu.background:                 turquoise",
    "Kpvslice*Value*background:                         #d0a0a0",
    "Kpvslice*ImageDisplay*quit*background:             orange",
    "Kpvslice*ScrollableList.highlightPixel:            aquamarine3",
    "Kpvslice*SimpleProgressMeter.background:           turquoise3",
    "Kpvslice*SimpleProgressMeter.foreground:           grey60",
    "Kpvslice*SimpleSlider.foreground:                  Sea Green",
    "Kpvslice*trackLabel0*font:                         8x13bold",
    "Kpvslice*trackLabel1*font:                         8x13bold",
    "Kpvslice*trackLabel2*font:                         8x13bold",
    "Kpvslice*trackLabel3*font:                         8x13bold",
    "Kpvslice*ExclusiveMenu.font:                       8x13bold",
    "Kpvslice*SimpleMenu*font:                          10x20",
    "Kpvslice*ImageDisplay.cmapSize:                    100",
    "Kpvslice*ImageDisplay.Command.font:                7x13bold",
    "Kpvslice*ImageDisplay.applicationBox.Command.font: 7x13bold",
    "Kpvslice*ImageDisplay.applicationBox.Ktoggle.font: 7x13bold",
    "Kpvslice*ImageDisplay.ChoiceMenu.font:             7x13bold",
    "Kpvslice*ImageDisplay.exportMenu.menuButton.font:  7x13bold",
    "Kpvslice*ImageDisplay.overlayMenu.menuButton.font: 7x13bold",
    "Kpvslice*ImageDisplay*zoomMenu*Unzoom*foreground:  red",
    "Kpvslice*Dataclip.AutoValueScale:                  False",
    "Kpvslice*SimpleSlider.borderWidth:                 0",
    "Kpvslice*IntensityPolicy*autoIntensityScale:       False",
    "Kpvslice*ZoomPolicy*integerXZoom:                  False",
    "Kpvslice*ZoomPolicy*integerYZoom:                  False",
    "Kpvslice*Filewin.dirname.textSink.font:            6x13",
    "Kpvslice*font:                                     9x15bold",
    "Kpvslice*borderColor:                              black",
    "Kpvslice*background:                               aquamarine2",
    "Kpvslice*foreground:                               black",
    NULL
};
static XrmOptionDescRec Options[] =
{
    {"-private_cmap", ".topForm.multiCanvas.pseudoColourCanvas.forceNewCmap",
     XrmoptionNoArg, (XPointer) "True"},
    {"-num_colours", ".topForm.cmapSize", XrmoptionSepArg, (XPointer) NULL},
    {"-fullscreen", ".topForm.fullscreen", XrmoptionNoArg, (XPointer) "True"},
    {"-verbose", "*verbose", XrmoptionNoArg, (XPointer) "True"},
};
static Widget main_shell = NULL;
static Widget main_image_display = NULL;
static ViewableImage magnified_image = NULL;
static Widget moment_winpopup = NULL;
static Widget itrace_winpopup = NULL;
static ViewableImage channel0_vimage = NULL;
static double image_min = TOOBIG;
static double image_max = -TOOBIG;


/*  Public functions follow  */

int main (int argc, char **argv)
{
    KWorldCanvas pc_canvas, tc_canvas;
    flag auto_update = TRUE;
    int argcount;
    XtAppContext app_context;
    Widget filewin, filepopup, w, box;
    Widget menu_btn, itrace_btn;
    extern KWorldCanvas main_canvas;
    extern char title_name[STRING_LENGTH];

    /*  Initialise module  */
    im_register_module_version_date (VERSION);
    im_register_lib_version (KARMA_VERSION);
    /*  Start up Xt  */
    main_shell = xtmisc_init_app_initialise (&app_context, "Kpvslice",
					     Options, XtNumber (Options),
					     &argc, argv, fallback_resources,
					     XTMISC_INIT_ATT_MIN_CCELLS, 200,
					     XTMISC_INIT_ATT_COMMS_SETUP, TRUE,
					     XTMISC_INIT_ATT_BECOME_SERVER,
					     TRUE,
					     XTMISC_INIT_ATT_TITLE, title_name,
					     XTMISC_INIT_ATT_END,
					     NULL);
    dsxfr_register_connection_limits (1, -1);
    xtmisc_set_icon (main_shell, ic_write_kimage_icon);
    main_image_display = XtVaCreateManagedWidget ("topForm",
						  imageDisplayWidgetClass,
						  main_shell,
						  XtNborderWidth, 0,
						  XkwNnumTrackLabels, 4,
						  XkwNupdateTrackLabels, TRUE,
						  NULL);
    if ( ( w = XtNameToWidget (main_image_display, "filesButton") ) != NULL )
    {
	/*  Force main "Files" button to be re-labelled as "Image"  */
	XtVaSetValues (w,
		       XtNlabel, "Image",
		       NULL);
    }
    filepopup = XtNameToWidget (main_image_display, "filewinPopup");
    XtVaGetValues (filepopup, XkwNfilewinWidget, &filewin, NULL);
    XtVaSetValues (filepopup,
		   XkwNfilenameTester, XkwFilewinStandardFileTester_nD,
		   XkwNpreferFloat, TRUE,
		   XkwNdomainName, "IMAGE",
		   NULL);
    XtVaSetValues (filewin,
		   XkwNtrapDirectoryDatasets, True,
		   NULL);
    ds_event_register_func (image_event_func, main_shell);
    /*  Get the application box  */
    box = XtNameToWidget (main_image_display, "applicationBox");
    w = XtVaCreateManagedWidget ("filesButton", commandWidgetClass,box,
				 XtNlabel, "New Cube",
				 NULL);
    XtAddCallback (w, XtNcallback, cubes_button_cbk, main_shell);
    /*  Create the image mode menu  */
    w = XtVaCreateManagedWidget ("menuButton", exclusiveMenuWidgetClass, box,
				 XtNmenuName, "imageModeMenu",
				 XkwNchoiceName, "Image Mode",
				 XkwNnumItems, NUM_IMAGE_MODE_ALTERNATIVES,
				 XkwNitemStrings, image_mode_alternatives,
				 NULL);
    menu_btn = w;
    /*  Create the auto update toggle  */
    w = XtVaCreateManagedWidget ("autoUpdateToggle", ktoggleWidgetClass, box,
				 XtNlabel, "Auto Update",
				 XtNstate, auto_update,
				 XkwNcrosses, FALSE,
				 NULL);
    XtAddCallback (w, XtNcallback, xtmisc_toggle_cbk,(XtPointer) &auto_update);
    /*  Create the intensity trace window popup button  */
    w = XtVaCreateManagedWidget ("traceButton", commandWidgetClass, box,
				 XtNlabel, "I-Profile",
				 NULL);
    itrace_btn = w;
    XtRealizeWidget (main_shell);
    XtVaGetValues (main_image_display,
		   XkwNpseudoColourCanvas, &pc_canvas,
		   XkwNtrueColourCanvas, &tc_canvas,
		   NULL);
    main_canvas = (pc_canvas == NULL) ? tc_canvas : pc_canvas;
    if (main_canvas == NULL)
    {
	fprintf (stderr, "PseudoColour or TrueColour canvas not found!\n");
	exit (RV_UNDEF_ERROR);
    }
    XtVaSetValues (main_image_display, XkwNvisibleCanvas, main_canvas, NULL);
    XtAddCallback (menu_btn, XkwNselectCallback, image_mode_cbk, NULL);
    /*  This draws the slice & calls extraction  */
    canvas_register_position_event_func (main_canvas,
					 ( flag (*) () )endpoint_position_func,
					 &auto_update);
    canvas_register_refresh_func (main_canvas, main_refresh_func, NULL);
    moment_winpopup = XtVaCreatePopupShell ("momentwinpopup",
					    momentGeneratorWidgetClass,
					    main_shell,
					    XtNtitle,
					    "Moment Generator Control Window",
					    NULL);
    XtAddCallback (moment_winpopup, XkwNmomentCallback, moment_cbk, NULL);
    XtRealizeWidget (moment_winpopup);
    itrace_winpopup = XtVaCreatePopupShell ("tracewinpopup",
					    tracePopupWidgetClass, main_shell,
					    XtNtitle, "I-Profile Window",
					    NULL);
    XtAddCallback (itrace_btn, XtNcallback, xtmisc_popup_cbk, itrace_winpopup);
    XtRealizeWidget (itrace_winpopup);
    /*  Option processing after widgets created and realised  */
    for (argcount = 1; argcount < argc; ++argcount)
    {
	extern char *trace_internal_colour, *trace_external_colour;

	if (strcmp (argv[argcount], "-trace_internal_colour") == 0)
	    trace_internal_colour = argv[++argcount];
	else if (strcmp (argv[argcount], "-trace_external_colour") == 0)
	    trace_external_colour = argv[++argcount];
	else if ( (strcmp (argv[argcount], "-image") == 0) ||
		  (strcmp (argv[argcount], "-cube") == 0) )
	{
	    multi_array *multi_desc;
	    CONST char *domain;
	    char fname[STRING_LENGTH];

	    if (argcount + 1 >= argc)
	    {
		fprintf (stderr, "File missing for dataset\n");
		continue;
	    }
	    if (strcmp (argv[argcount], "-image") == 0) domain = "IMAGE";
	    else domain = create_slice_win (main_shell, main_canvas, FALSE);
	    ++argcount;
	    if (strncmp (argv[argcount], "./", 2) == 0)
	    {
		strcpy (fname, argv[argcount] + 2);
	    }
	    else strcpy (fname, argv[argcount]);
	    if ( ( multi_desc =
		   foreign_guess_and_read (fname, K_CH_MAP_LOCAL,
					   FALSE, NULL,
					   FA_GUESS_READ_FITS_TO_FLOAT, TRUE,
					   FA_GUESS_READ_END) )
		 == NULL )
	    {
		fprintf (stderr, "Error reading file: \"%s\"\n", fname);
		continue;
	    }
	    ds_event_dispatch (multi_desc, domain, fname);
	    ds_dealloc_multi (multi_desc);
	    m_clear (fname, STRING_LENGTH);  /*  Force me to take care  */
	}
	else fprintf (stderr, "Option: \"%s\" ignored\n", argv[argcount]);
    }
    XtAppMainLoop (app_context);
    return (RV_OK);
}   /*  End Function main   */

void print_slice_info (CONST char *window_string)
/*  [SUMMARY] Print information about the slice.
    <window_string> The string to display.
    [RETURNS] Nothing.
*/
{
    Widget label;

    label = XtNameToWidget (main_image_display, "trackLabel3");
    XtVaSetValues (label,
		   XtNlabel, window_string,
		   NULL);
}   /*  End Function print_slice_info  */

void update_trace_profile (double x0, double y0, double x1, double y1)
/*  [SUMMARY] Update the intensity profile widget with the image trace.
    <x0> The starting horizontal co-ordinate in world co-ordinates.
    <y0> The starting vertical co-ordinate in world co-ordinates.
    <x1> The end horizontal co-ordinate in world co-ordinates.
    <y1> The end vertical co-ordinate in world co-ordinates.
    [RETURNS] Nothing.
*/
{
    iarray array = NULL;          /*  Initialised to keep compiler happy  */
    KwcsAstro ap;
    ViewableImage vimage = NULL;  /*  Initialised to keep compiler happy  */
    flag is_sky = FALSE;
    unsigned int count, length;
    double x, y, xstep, ystep, xmax, ymax, wstep, imin, imax;
    dim_desc *xdim, *ydim;
    extern iarray image_arr, mom0_arr, mom1_arr;
    extern ViewableImage image_vimage, mom0_vimage, mom1_vimage;
    extern unsigned int image_mode;
    static iarray profile_arr = NULL;
    static char function_name[] = "update_trace_profile";

    switch (image_mode)
    {
      case IMAGE_MODE_LOADED:
      case IMAGE_MODE_CHANNEL0:
	array = image_arr;
	vimage = image_vimage;
	break;
      case IMAGE_MODE_MOM0:
	array = mom0_arr;
	vimage = mom0_vimage;
	break;
      case IMAGE_MODE_MOM1:
	array = mom1_arr;
	vimage = mom1_vimage;
	break;
      default:
	fprintf (stderr, "Illegal image mode: %u\n", image_mode);
	a_prog_bug (function_name);
	break;
    }
    if (array == NULL)
    {
	fprintf (stderr, "%s: no image data available\n", function_name);
	return;
    }
    if (vimage == NULL)
    {
	fprintf (stderr, "%s: no ViewableImage available\n", function_name);
	return;
    }
    viewimg_get_attributes (vimage,
			    VIEWIMG_VATT_VALUE_MIN, &imin,
			    VIEWIMG_VATT_VALUE_MAX, &imax,
			    VIEWIMG_VATT_END);
    ap = iarray_get_associated_wcs_astro (array);
    xdim = iarray_get_dim_desc (array, 1);
    ydim = iarray_get_dim_desc (array, 0);
    if ( ( wcs_astro_test_if_lon (ap, xdim->name) &&
	   wcs_astro_test_if_lat (ap, ydim->name) ) ||
	 ( wcs_astro_test_if_lat (ap, xdim->name) &&
	   wcs_astro_test_if_lon (ap, ydim->name) ) ) is_sky = TRUE;
    if (ap)
    {
	/*  Convert to linear co-ordinates in image frame  */
	wcs_astro_transform3 (ap, FALSE, 1,
			      xdim->name, &x0, TRUE,
			      ydim->name, &y0, TRUE,
			      NULL, NULL, FALSE,
			      0, NULL, NULL);
	wcs_astro_transform3 (ap, FALSE, 1,
			      xdim->name, &x1, TRUE,
			      ydim->name, &y1, TRUE,
			      NULL, NULL, FALSE,
			      0, NULL, NULL);
    }
    /*  Convert to co-ordinate indices  */
    x0 = ds_get_coord_num (xdim, x0, SEARCH_BIAS_CLOSEST);
    y0 = ds_get_coord_num (ydim, y0, SEARCH_BIAS_CLOSEST);
    x1 = ds_get_coord_num (xdim, x1, SEARCH_BIAS_CLOSEST);
    y1 = ds_get_coord_num (ydim, y1, SEARCH_BIAS_CLOSEST);
    if ( fabs (x1 - x0) > fabs (y1 - y0) )
    {
	length = fabs (x1 - x0) + 1.0;
	if (length < 2) return;
	xstep = (x1 > x0) ? 1.0 : -1.0;
	if (y1 != y0) ystep = (y1 - y0) / (double) (length - 1);
	else ystep = 0.0;
    }
    else
    {
	length = fabs (y1 - y0) + 1.0;
	if (length < 2) return;
	ystep = (y1 > y0) ? 1.0 : -1.0;
	if (x1 != x0) xstep = (x1 - x0) / (double) (length - 1);
	else xstep = 0.0;
    }
    if ( (profile_arr == NULL) ||
	 (iarray_dim_length (profile_arr, 0) != length) )
    {
	iarray_dealloc2 (profile_arr);
	if ( ( profile_arr = iarray_create_1D (length, K_FLOAT) ) == NULL )
	    m_abort (function_name, "profile array");
	iarray_set_value_name (profile_arr, iarray_value_name (array),
			       TRUE);
	iarray_set_dim_name (profile_arr, 0, is_sky ? "ANGLE" : "OFFSET",
			     TRUE);
	XkwTracePopupNewArray (itrace_winpopup, profile_arr,
			       imin, imax);
    }
    /*  Extract data from image along line  */
    xmax = iarray_dim_length (array, 1) - 1;
    ymax = iarray_dim_length (array, 0) - 1;
    for (count = 0, x = x0, y = y0; count < length;
	 ++count, x += xstep, y += ystep)
    {
	if (x < 0.0) x = 0.0;
	else if (x > xmax) x = xmax;
	if (y < 0.0) y = 0.0;
	else if (y > ymax) y = ymax;
	F1 (profile_arr, count) = F2 (array, (int) y, (int) x);
    }
    /*  Now fiddle the horizontal axis. First compute two adjacent points along
	the locus  */
    x0 = ds_get_coordinate (xdim, x0);
    y0 = ds_get_coordinate (ydim, y0);
    x1 = ds_get_coordinate (xdim, x0 + xstep);
    y1 = ds_get_coordinate (ydim, y0 + ystep);
    wcs_astro_transform3 (ap, FALSE, 1,
			  xdim->name, &x0, FALSE,
			  ydim->name, &y0, FALSE,
			  NULL, NULL, FALSE,
			  0, NULL, NULL);
    wcs_astro_transform3 (ap, FALSE, 1,
			  xdim->name, &x1, FALSE,
			  ydim->name, &y1, FALSE,
			  NULL, NULL, FALSE,
			  0, NULL, NULL);
    /*  Compute offset in RA and DEC  */
    x1 -= x0;
    y1 -= y0;
    if (is_sky) x1 *= cos (PION180 * y0);  /*  Apply cos (dec) correction  */
    wstep = sqrt (x1 * x1 + y1 * y1);
    /*  Set co-ordinates such that the centre of the axis is 0.0  */
    x = (double) (length - 1) * wstep * 0.5;
    iarray_set_world_coords (profile_arr, 0, -x, x);
    XkwTracePopupShowTrace (itrace_winpopup, NULL, NULL);
}   /*  End Function update_trace_profile  */


/*  Private routines follow  */

static void image_mode_cbk (Widget w, XtPointer client_data,
			    XtPointer call_data)
/*  [SUMMARY] Image mode change callback.
    <w> The widget the event occurred on.
    <client_data> Client data.
    <call_data> Call data.
    [RETURNS] Nothing.
*/
{
    unsigned int code = *(int *) call_data;
    extern ViewableImage image_vimage;
    extern unsigned int image_mode;
    /*static char function_name[] = "image_mode_cbk";*/

    image_mode = code;
    switch (code)
    {
      case IMAGE_MODE_LOADED:
	show_image (image_vimage);
	break;
      case IMAGE_MODE_MOM0:
	XtPopup (moment_winpopup, XtGrabNone);
	show_mom0 ();
	break;
      case IMAGE_MODE_MOM1:
	XtPopup (moment_winpopup, XtGrabNone);
	show_mom1 ();
	break;
      case IMAGE_MODE_CHANNEL0:
	show_channel0 ();
	break;
    }
}   /*  End Function image_mode_cbk  */

static void show_image (ViewableImage vimage)
/*  [SUMMARY] Show the loaded image.
    <vimage> The image.
    [RETURNS] Nothing.
*/
{
    /*static char function_name[] = "show_image";*/

    if (vimage == NULL) return;
    viewimg_make_active (vimage);
}   /*  End Function show_image  */

static void show_mom0 ()
/*  [SUMMARY] Show the 0th moment map.
    [RETURNS] Nothing.
*/
{
    double min, max;
    extern ViewableImage mom0_vimage;
    /*static char function_name[] = "show_mom0";*/

    if (mom0_vimage == NULL) return;
    viewimg_get_attributes (mom0_vimage,
			    VIEWIMG_VATT_VALUE_MIN, &min,
			    VIEWIMG_VATT_VALUE_MAX, &max,
			    VIEWIMG_VATT_END);
    if (min >= max)
    {
	fprintf (stderr, "0th moment image contains only: %e: ignoring\n",
		 min);
	return;
    }
    viewimg_make_active (mom0_vimage);
}   /*  End Function show_mom0  */

static void show_mom1 ()
/*  [SUMMARY] Show the 1st moment map.
    [RETURNS] Nothing.
*/
{
    double min, max;
    extern ViewableImage mom1_vimage;
    /*static char function_name[] = "show_mom1";*/

    if (mom1_vimage == NULL) return;
    viewimg_get_attributes (mom1_vimage,
			    VIEWIMG_VATT_VALUE_MIN, &min,
			    VIEWIMG_VATT_VALUE_MAX, &max,
			    VIEWIMG_VATT_END);
    if (min >= max)
    {
	fprintf (stderr, "1st moment image contains only: %e: ignoring\n",
		 min);
	return;
    }
    viewimg_make_active (mom1_vimage);
}   /*  End Function show_mom1  */

static void moment_cbk (Widget w, XtPointer client_data, XtPointer call_data)
/*  [SUMMARY] Moment computation callback.
    <w> The widget the event occurred on.
    <client_data> Client data.
    <call_data> Call data.
    [RETURNS] Nothing.
*/
{
    SliceDisplay sd;
    double min, max;
    extern iarray mom0_arr, mom1_arr;
    extern ViewableImage mom0_vimage, mom1_vimage;
    extern SliceDisplay first_slice_display;
    extern unsigned int image_mode;
    /*static char function_name[] = "moment_cbk";*/

    iarray_min_max (mom0_arr, CONV_CtoR_REAL, &min, &max);
    viewimg_set_attributes (mom0_vimage,
			    VIEWIMG_VATT_VALUE_MIN, min,
			    VIEWIMG_VATT_VALUE_MAX, max,
			    VIEWIMG_VATT_END);
    viewimg_register_data_change (mom0_vimage);
    iarray_min_max (mom1_arr, CONV_CtoR_REAL, &min, &max);
    viewimg_set_attributes (mom1_vimage,
			    VIEWIMG_VATT_VALUE_MIN, min,
			    VIEWIMG_VATT_VALUE_MAX, max,
			    VIEWIMG_VATT_END);
    viewimg_register_data_change (mom1_vimage);
    if (image_mode == IMAGE_MODE_MOM0) show_mom0 ();
    if (image_mode == IMAGE_MODE_MOM1) show_mom1 ();
    for (sd = first_slice_display; sd; sd = sd->next)
	XkwImageDisplayRefresh (sd->image_display, TRUE);
}   /*  End Function moment_cbk  */

static void cubes_button_cbk (Widget w, XtPointer client_data,
			      XtPointer call_data)
/*  [SUMMARY] "New Cube" button callback.
    <w> The widget the event occurred on.
    <client_data> Client data.
    <call_data> Call data.
    [RETURNS] Nothing.
*/
{
    extern KWorldCanvas main_canvas;

    create_slice_win (client_data, main_canvas, TRUE);
}   /*  End Function cubes_button_cbk  */

static void destroy_func (iarray array, void *info)
/*  [SUMMARY] Register destruction of an Intelligent Array.
    <array> The Intelligent Array.
    <info> A pointer to the arbitrary information.
    [RETURNS] Nothing.
*/
{
    ViewableImage *vimage = info;

    viewimg_destroy (*vimage);
    *vimage = NULL;
}   /*  End Function destroy_func  */

static void show_channel0 ()
/*  [SUMMARY] Show the 0th channel map.
    [RETURNS] Nothing.
*/
{
    /*static char function_name[] = "show_channel0";*/

    if (channel0_vimage == NULL) return;
    viewimg_make_active (channel0_vimage);
}   /*  End Function show_channel0  */

void new_primary_cube (iarray cube, double min, double max)
/*  [SUMMARY] Process a new primary cube.
    <cube> The primary cube.
    <min> The minimum value in the cube.
    <max> The maximum value in the cube.
    [RETURNS] Nothing.
*/
{
    extern KWorldCanvas main_canvas;
    extern iarray mom0_arr, mom1_arr;
    extern ViewableImage mom0_vimage, mom1_vimage;
    static char function_name[] = "new_primary_cube";

    XkwMomentGeneratorNewArray (moment_winpopup, cube, min, max);
    if (channel0_vimage) viewimg_destroy (channel0_vimage);
    channel0_vimage = NULL;
    if (cube)
    {
	if ( ( channel0_vimage = viewimg_create_restr (main_canvas,
						       cube->multi_desc,
						       cube->arr_desc,
						       cube->data,
						       2, 1, 0,
						       0, NULL, NULL) )
	     == NULL )
	{
	    m_abort (function_name, "0th channel ViewableImage");
	}
	viewimg_set_attributes (channel0_vimage,
				VIEWIMG_VATT_VALUE_MIN, min,
				VIEWIMG_VATT_VALUE_MAX, max,
				VIEWIMG_VATT_END);
	if (mom0_vimage)
	{
	    fprintf (stderr,
		     "WARNING: 0th moment ViewableImage not deallocated\n");
	}
	if (mom1_vimage)
	{
	    fprintf (stderr,
		     "WARNING: 1st moment ViewableImage not deallocated\n");
	}
	XtVaGetValues (moment_winpopup,
		       XkwNmom0Array, &mom0_arr,
		       XkwNmom1Array, &mom1_arr,
		       NULL);
	if ( ( mom0_vimage = viewimg_create_from_iarray (main_canvas, mom0_arr,
							 FALSE) ) == NULL )
	{
	    m_abort (function_name, "0th moment ViewableImage");
	}
	if ( ( mom1_vimage = viewimg_create_from_iarray (main_canvas, mom1_arr,
							 FALSE) ) == NULL )
	{
	    m_abort (function_name, "1st moment ViewableImage");
	}
	iarray_register_destroy_func (mom0_arr, ( flag (*) () ) destroy_func,
				      &mom0_vimage);
	iarray_register_destroy_func (mom1_arr, ( flag (*) () ) destroy_func,
				      &mom1_vimage);
    }
}   /*  End Function new_primary_cube  */

static flag image_event_func (void *object, multi_array *multi_desc,
			      CONST char *domain, CONST char *name)
/*  [SUMMARY] Process a new image event.
    <object> The object information pointer.
    <multi_desc> The multi_array descriptor.
    <domain> The domain from where the event was generated.
    <name> The name of the event in its domain.
    [RETURNS] TRUE if further callbacks should not be called, else FALSE.
*/
{
    KWorldCanvas mag_pseudo_canvas;
    Widget izoomwinpopup, magnifier;
    Widget shell = object;
    char filename[STRING_LENGTH];
    extern KWorldCanvas main_canvas;
    extern ViewableImage image_vimage;
    extern iarray image_arr;
    extern char title_name[STRING_LENGTH];

    if (strcmp (domain, "CUBE") == 0) return (FALSE);
    if (strncmp (domain, "SliceDisplay: ", 14) == 0) return (FALSE);
    if ( (strcmp (domain, "GLISH") == 0) && (strcmp (name, "image") != 0) )
	return (FALSE);
    if (strcmp (domain, "IMAGE") == 0) filename[0] = '\0';
    else
    {
	strcpy (filename, domain);
	strcat (filename, "::");
    }
    strcat (filename, name);
    XtVaGetValues (main_image_display,
		   XkwNmagnifier, &magnifier,
		   NULL);
    XtVaGetValues (magnifier, XkwNvisibleCanvas, &mag_pseudo_canvas, NULL);
    izoomwinpopup = XtNameToWidget (main_image_display, "izoomwinpopup");
    if ( !process_image (multi_desc, &image_arr, &image_min, &image_max,
			 main_canvas, mag_pseudo_canvas,
			 &image_vimage, &magnified_image) ) return (FALSE);
    if (strcmp (name, "UNLOAD") == 0) XtVaSetValues (shell,
						     XtNtitle, title_name,
						     NULL);
    else
    {
	char *ptr;
	char stripped_filename[STRING_LENGTH], title[STRING_LENGTH];

	strcpy (stripped_filename, filename);
	if ( ( ptr = strrchr (stripped_filename, '.') ) != NULL )
	{
	    if (strcmp (ptr, ".kf") == 0) *ptr = '\0';
	}
	sprintf (title, "%s  file: %s\n", title_name, filename);
	XtVaSetValues (main_image_display,
		       XkwNimageName, stripped_filename,
		       NULL);
	XtVaSetValues (shell,
		       XtNtitle, title,
		       NULL);
    }
    XkwDataclipNewArray (izoomwinpopup, image_arr, image_min, image_max, TRUE);
    return (TRUE);
}   /*  End Function image_event_func  */
