/*  main_xt.c

    Slice file for  kpvslice  (X11 Position-Velocity slice tool for Karma).

    Copyright (C) 1996-2004  Richard Gooch

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Richard Gooch may be reached by email at  karma-request@atnf.csiro.au
    The postal address is:
      Richard Gooch, c/o ATNF, P. O. Box 76, Epping, N.S.W., 2121, Australia.
*/

/*
    This Karma module will enable interactive selection of a Position-Velocity
    slice through a data cube.
    This module runs on an X11 server.


    Written by      Richard Gooch   19-MAR-2001: Moved parts from main_xt.c.

    Updated by      Richard Gooch   21-MAR-2001: First working version.

    Updated by      Richard Gooch   22-MAR-2001: Changed "Unload" to "Discard".
  Discard on failed load if no cube was ever loaded for window.

    Updated by      Richard Gooch   23-MAR-2001: Made use of new function
  <xtmisc_init_get_pc_colourmap> to insure slice window has enough colours.

    Updated by      Richard Gooch   14-AUG-2002: Added #include <stdlib.h>.

    Updated by      Richard Gooch   22-DEC-2002: Stupid compiler warning fix.

    Updated by      Richard Gooch   14-JUN-2003: #include cleanups.

    Updated by      Richard Gooch   4-OCT-2003: Added #include <Xkw/Dataclip.h>

    Last updated by Richard Gooch   6-MAR-2004: Return domain name in
  <create_slice_win> and added <<popup_filebrowser>> parameter.


*/
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <X11/Xatom.h>
#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/Xaw/Command.h>
#include <X11/Xaw/Form.h>
#include <X11/Xaw/Dialog.h>
#include <karma.h>
#define X11
#include <k_event_codes.h>
#include <karma_viewimg.h>
#include <karma_xtmisc.h>
#include <karma_kwin.h>
#include <karma_ds.h>
#include <karma_m.h>
#include <karma_c.h>
#include <Xkw/ImageDisplay.h>
#include <Xkw/Dialogpopup.h>
#include <Xkw/Filewin.h>
#include <Xkw/Ktoggle.h>
#include <Xkw/ExclusiveMenu.h>
#include <Xkw/MomentGenerator.h>
#include <Xkw/TracePopup.h>
#include <Xkw/Dataclip.h>
#define WIDGET_TYPE Widget
#include "kpvslice.h"


/*  Private functions  */
STATIC_FUNCTION (void show_trace_toggle_cbk,
		 (Widget w, XtPointer client_data, XtPointer call_data) );
STATIC_FUNCTION (flag main_canvas_event,
		 (ViewableImage vimage, double x, double y,
		  void *value, unsigned int event_code,
		  void *e_info, void **f_info,
		  double x_lin, double y_lin, unsigned int value_type,
		  double x_im, double y_im, int x_pix, int y_pix) );
STATIC_FUNCTION (flag cube_event_func,
		 (void *object, multi_array *multi_desc,
		  CONST char *domain, CONST char *name) );
STATIC_FUNCTION (void unload_cbk,
		 (Widget w, XtPointer client_data, XtPointer call_data) );


/*  Private data  */


/*  Public functions follow  */

CONST char *create_slice_win (Widget main_shell, KWorldCanvas main_canvas,
			      flag popup_filebrowser)
/*  [SUMMARY] Create a cube slice window.
    <main_shell> The main_shell widget.
    <main_canvas> The main KWorldCanvas.
    <popup_filebrowser> If TRUE, the file browser is poped up.
    [RETURNS] A pointer to the domain name for the file browser.
*/
{
    KWorldCanvas pc_canvas, tc_canvas;
    SliceDisplay sd;
    Widget w, filepopup, box, ztrace_btn;
    Colormap xcmap;
    extern SliceDisplay first_slice_display, last_slice_display;
    static char function_name[] = "cubes_button_cbk";

    if ( ( sd = m_calloc (sizeof *sd) ) == NULL )
	m_abort (function_name, "SliceDisplay object");
    sd->show_trace = TRUE;
    sd->cube_min = TOOBIG;
    sd->cube_max = -TOOBIG;
    sd->cube_callback = ds_event_register_func (cube_event_func, sd);
    sd->main_canvas_event_callback =
	viewimg_register_position_event_func (main_canvas,
					      ( flag (*) () )main_canvas_event,
					      sd);
    xcmap = xtmisc_init_get_pc_colourmap (XtScreen (main_shell), 100, 100,
					  FALSE, NULL);
    sd->shell = XtVaCreatePopupShell ("auxShell", topLevelShellWidgetClass,
				      main_shell,
				      XtNtitle, "Slice Window",
				      XtNcolormap, xcmap,
				      NULL);
    sd->image_display = XtVaCreateManagedWidget ("topForm",
						 imageDisplayWidgetClass,
						 sd->shell,
						 XtNborderWidth, 0,
						 XkwNcreateFilepopup, TRUE,
                                                 XkwNnumTrackLabels, 3,
						 XkwNshowQuitButton, FALSE,
						 NULL);
    if ( ( w = XtNameToWidget (sd->image_display, "filesButton") ) != NULL )
    {
	/*  Force "Files" button to be re-labelled as "New Cube"  */
	XtVaSetValues (w,
		       XtNlabel, "Reload Cube",
		       NULL);
    }
    sprintf (sd->domain_name, "SliceDisplay: %p", (void *) sd);
    filepopup = XtNameToWidget (sd->image_display, "filewinPopup");
    XtVaSetValues (filepopup,
		   XkwNfilenameTester, XkwFilewinStandardFileTester_3D,
		   XkwNpreferFloat, TRUE,
		   XkwNdomainName, sd->domain_name,
		   NULL);
    XtVaGetValues (filepopup, XkwNfilewinWidget, &w, NULL);
    XtVaSetValues (w,
		   XkwNtrapDirectoryDatasets, True,
		   NULL);
    /*  Get application box  */
    box = XtNameToWidget (sd->image_display, "applicationBox");
    /*  Create the show trace toggle  */
    w = XtVaCreateManagedWidget ("showTraceToggle", ktoggleWidgetClass, box,
				 XtNlabel, "Show Trace",
				 XtNstate, sd->show_trace,
				 XkwNcrosses, FALSE,
				 NULL);
    XtAddCallback (w, XtNcallback, show_trace_toggle_cbk, sd);
    w = XtVaCreateManagedWidget ("traceButton", commandWidgetClass, box,
				 XtNlabel, "Z-Profile",
				 NULL);
    ztrace_btn = w;
    w = XtVaCreateManagedWidget ("closeButton", commandWidgetClass, box,
				 XtNlabel, "Discard",
				 NULL);
    XtAddCallback (w, XtNcallback, unload_cbk, sd);
    XtRealizeWidget (sd->shell);
    XtVaGetValues (sd->image_display,
		   XkwNpseudoColourCanvas, &pc_canvas,
		   XkwNtrueColourCanvas, &tc_canvas,
		   NULL);
    sd->canvas = (pc_canvas == NULL) ? tc_canvas : pc_canvas;
    if (sd->canvas == NULL)
    {
	fprintf (stderr, "PseudoColour or TrueColour canvas not found!\n");
	exit (RV_UNDEF_ERROR);
    }
    XtVaSetValues (sd->image_display, XkwNvisibleCanvas, sd->canvas, NULL);
    /*  This reads back cursor coords in 'slice' window  */
    /* how to do it otherwise?
       I want to  have the extra header info transferred only once.
       I also want to be prompted for a fits file name.
       I suppose fix_slice_headers could be called when cursor enters the
       aux window, or the export menu is selected, so the file is ready to go
       as soon as the file name has been entered.
       Put in a request for a fits_file_name from the aux_display widget, and
       when that request is fulfilled, run the fix_slice_headers function?
    */
    viewimg_register_position_event_func (sd->canvas,
					  ( flag (*) () ) track_aux_canvas_event,
					  sd);
					  
    canvas_register_refresh_func (sd->canvas, trace_refresh_func, sd);
    w = XtVaCreatePopupShell ("tracewinpopup", tracePopupWidgetClass,sd->shell,
			      XtNtitle, "Z-Profile Window",
			      NULL);
    sd->ztrace_winpopup = w;
    XtAddCallback (ztrace_btn, XtNcallback, xtmisc_popup_cbk, w);
    XtPopup (sd->shell, XtGrabNone);
    if (popup_filebrowser) XtPopup (filepopup, XtGrabNone);
    /*  Append to list  */
    sd->prev = last_slice_display;
    if (first_slice_display == NULL) first_slice_display = sd;
    else last_slice_display->next = sd;
    last_slice_display = sd;
    return (sd->domain_name);
}   /*  End Function create_slice_win  */

flag track_aux_canvas_xtcoord (SliceDisplay sd, CONST char *track_txt,
			       int track_num)
/*  [PURPOSE] Handles the Xt part of outputting coords while tracking cursor
    in auxilary window.
    NB: The first track coord is slice centre & PA, which is sent from
    main_canvas_event, as the slice is moved around in the main window.
    This routine updates the second or third track coord lines, which have
    the offset arcsec & velocity coords, and
    the RA & DEC of that offset, respectively.
    This routine should only be called by track_aux_canvas_event.
    [RETURNS] FALSE if an error, TRUE otherwise.
*/
{
    Widget track_label;
    char trackLabeln[STRING_LENGTH];
    /*static char function_name[] = "track_aux_canvas_xtcoord";*/

    if ( track_num < 0 || track_num > 2 ) {
      fprintf (stderr, "TRACK_AUX_CANVAS_XTSTUFF: Bad track label number\n");
      return(FALSE);
    }
    sprintf (trackLabeln, "%s%d", "trackLabel", track_num);

    track_label = XtNameToWidget (sd->image_display, trackLabeln);
    XtVaSetValues (track_label, XtNlabel, track_txt, NULL);

    return(TRUE);
}   /*  End Function track_aux_canvas_xtcoord  */


/*  Private routines follow  */

static void show_trace_toggle_cbk (Widget w, XtPointer client_data,
				   XtPointer call_data)
/*  [PURPOSE] This routine is called when the user clicks the show trace toggle
*/
{
    SliceDisplay sd = client_data;

    sd->show_trace = (iaddr) call_data ? TRUE : FALSE;
    XkwImageDisplayRefresh (sd->image_display, TRUE);
}   /*  End Function show_trace_toggle_cbk  */

static flag main_canvas_event (ViewableImage vimage, double x, double y,
			       void *value, unsigned int event_code,
			       void *e_info, void **f_info,
			       double x_lin, double y_lin,
			       unsigned int value_type,
			       double x_im, double y_im, int x_pix, int y_pix)
/*  [PURPOSE] This routine is a position event consumer for a world canvas
    which has a number of ViewableImage objects associated with it.
    <viewimg> The active viewable image.
    <x> The horizontal world co-ordinate of the event.
    <y> The vertical world co-ordinate of the event.
    These values will have been transformed by the registered transform
    function (see  canvas_register_transform_func  ) for the associated
    world canvas.
    <value> A pointer to the data value in the viewable image corresponding
    to the event co-ordinates.
    <event_code> The arbitrary event code.
    <e_info> The arbitrary event information.
    <f_info> The arbitrary function information pointer.
    <x_lin> The linear horizontal world co-ordinate (the co-ordinate prior
    to the transform function being called).
    <y_lin> The linear vertical world co-ordinate (the co-ordinate prior
    to the transform function being called).
    <value_type> The type of the data value. This may be K_DCOMPLEX or
    K_UB_RGB.
    <x_im> The horizontal image pixel co-ordinate. 0.0 is the pixel centre.
    <y_im> The vertical image pixel co-ordinate. 0.0 is the pixel centre.
    <x_pix> The horizontal screen pixel co-ordinate.
    <y_pix> The vertical screen pixel co-ordinate.
    [RETURNS] TRUE if the event was consumed, else FALSE indicating that
    the event is still to be processed.
*/
{
    SliceDisplay sd = *f_info;
    KwcsAstro main_ap, cube_ap;
    double cube_x, cube_y, min, max;
    dim_desc *dim;
    uaddr coords[2];
    unsigned int dim_indices[2];
    /*static char function_name[] = "main_canvas_event";*/

    if (event_code == K_CANVAS_EVENT_POINTER_ENTER) return (FALSE);
    if (sd->cube == NULL) return (FALSE);
    cube_ap = iarray_get_associated_wcs_astro (sd->cube);
    viewimg_get_attributes (vimage,
			    VIEWIMG_VATT_WCS_ASTRO, &main_ap,
			    VIEWIMG_VATT_END);
    convert_lin_world_coords (&cube_x, &cube_y, cube_ap,
			      &x_lin, &y_lin, main_ap, 1);
    dim_indices[0] = 2;
    dim_indices[1] = 1;
    dim = iarray_get_dim_desc (sd->cube, 2);
    if (dim->first_coord < dim->last_coord)
    {
	min = dim->first_coord;
	max = dim->last_coord;
    }
    else
    {
	min = dim->last_coord;
	max = dim->first_coord;
    }
    if ( (cube_x < min) || (cube_x > max) )
    {
	XkwTracePopupInvalidate (sd->ztrace_winpopup);
	return (FALSE);
    }
    coords[0] = ds_get_coord_num (dim, cube_x, SEARCH_BIAS_CLOSEST);
    dim = iarray_get_dim_desc (sd->cube, 1);
    if (dim->first_coord < dim->last_coord)
    {
	min = dim->first_coord;
	max = dim->last_coord;
    }
    else
    {
	min = dim->last_coord;
	max = dim->first_coord;
    }
    if ( (cube_y < min) || (cube_y > max) )
    {
	XkwTracePopupInvalidate (sd->ztrace_winpopup);
	return (FALSE);
    }
    coords[1] = ds_get_coord_num (dim, cube_y, SEARCH_BIAS_CLOSEST);
    XkwTracePopupShowTrace (sd->ztrace_winpopup, dim_indices, coords);
    return (FALSE);
}   /*  End Function main_canvas_event  */

static flag cube_event_func (void *object, multi_array *multi_desc,
			     CONST char *domain, CONST char *name)
/*  [SUMMARY] Process a new cube event.
    <object> The object information pointer.
    <multi_desc> The multi_array descriptor.
    <domain> The domain from where the event was generated.
    <name> The name of the event in its domain.
    [RETURNS] TRUE if further callbacks should not be called, else FALSE.
*/
{
    SliceDisplay sd = object;
    Widget izoomwinpopup;
    char filename[STRING_LENGTH];
    /*static char function_name[] = "cube_event_func";*/

    if (strcmp (domain, "IMAGE") == 0) return (FALSE);
    if ( (strcmp (domain, "GLISH") == 0) && (strcmp (name, "cube") != 0) )
	return (FALSE);
    if (strncmp (domain, "SliceDisplay: ", 14) == 0)
    {
	if (strcmp (domain, sd->domain_name) == 0) filename[0] = '\0';
	else return (FALSE);
    }
    else
    {
	if (sd->prev != NULL) return (FALSE);
	if (strcmp (domain, "CUBE") == 0) filename[0] = '\0';
	else
	{
	    strcpy (filename, domain);
	    strcat (filename, "::");
	}
    }
    strcat (filename, name);
    if ( !process_cube (multi_desc, &sd->cube, &sd->cube_min, &sd->cube_max) )
    {
	if (!sd->had_cube) unload_cbk (NULL, sd, NULL);
	return (FALSE);
    }
    sd->had_cube = TRUE;
    canvas_set_attributes (sd->canvas,
			   CANVAS_ATT_VALUE_MIN, sd->cube_min,
			   CANVAS_ATT_VALUE_MAX, sd->cube_max,
			   CANVAS_ATT_END);
    kwin_resize (canvas_get_pixcanvas (sd->canvas), TRUE, 0, 0, 0, 0);
    izoomwinpopup = XtNameToWidget (sd->image_display, "izoomwinpopup");
    XkwDataclipNewArray (izoomwinpopup, sd->cube, sd->cube_min, sd->cube_max,
			 TRUE);
    XkwTracePopupNewArray (sd->ztrace_winpopup, sd->cube,
			   sd->cube_min, sd->cube_max);
    if (multi_desc == NULL) XtVaSetValues (sd->shell,
					   XtNtitle, "Slice Window",
					   NULL);
    else
    {
	char *ptr;
	char stripped_filename[STRING_LENGTH], title[STRING_LENGTH];

	strcpy (stripped_filename, filename);
	if ( ( ptr = strrchr (stripped_filename, '.') ) != NULL )
	{
	    if (strcmp (ptr, ".kf") == 0) *ptr = '\0';
	}
	sprintf (title, "Slice Window for file: %s\n", filename);
	XtVaSetValues (sd->image_display,
		       XkwNimageName, stripped_filename,
		       NULL);
	XtVaSetValues (sd->shell,
		       XtNtitle, title,
		       NULL);
    }
    if (sd->prev == NULL)
	new_primary_cube (sd->cube, sd->cube_min, sd->cube_max);
    return (TRUE);
}   /*  End Function cube_event_func  */

static void unload_cbk (Widget w, XtPointer client_data, XtPointer call_data)
/*  [PURPOSE] This routine is called when the user clicks the unload button.
*/
{
    SliceDisplay sd = client_data;
    extern SliceDisplay first_slice_display, last_slice_display;

    c_unregister_callback (sd->cube_callback);
    c_unregister_callback (sd->main_canvas_event_callback);
    XtDestroyWidget (sd->shell);
    if (!sd->prev) new_primary_cube (NULL, TOOBIG, -TOOBIG);
    if (sd->cube) iarray_dealloc (sd->cube);
    if (sd->image) iarray_dealloc (sd->image);
    if (sd->offset_coords) m_free (sd->offset_coords);
    if (sd->x_locus) m_free (sd->x_locus);
    if (sd->y_locus) m_free (sd->y_locus);
    if (sd->ra_coord) m_free (sd->ra_coord);
    if (sd->dec_coord) m_free (sd->dec_coord);
    if (!sd->prev) first_slice_display = sd->next;
    else sd->prev->next = sd->next;
    if (!sd->next) last_slice_display = sd->prev;
    else sd->next->prev = sd->prev;
    m_free (sd);
}   /*  End Function unload_cbk  */
