/*  main_xt.c

    Main file for  kshell  (X11 ellipse integrator tool for Karma).

    Copyright (C) 1996-2003  Richard Gooch

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Richard Gooch may be reached by email at  karma-request@atnf.csiro.au
    The postal address is:
      Richard Gooch, c/o ATNF, P. O. Box 76, Epping, N.S.W., 2121, Australia.
*/

/*
    This Karma module will enable interactive integration along the
    circumference of concentric ellipses.
    This module runs on an X11 server.


    Written by      Richard Gooch   24-SEP-1996: Copied from kpvslice module.

    Updated by      Richard Gooch   26-SEP-1996: Added aux_ap support.

    Updated by      Richard Gooch   1-OCT-1996: Fixed resource names for new
  ExportMenu widget. Added "0th channel" image mode.

    Updated by      Richard Gooch   28-OCT-1996: Added hostname and port number
  to title.

    Updated by      Richard Gooch   3-NOV-1996: Made use of new
  <viewimg_track _compute> routine.

    Updated by      Richard Gooch   1-DEC-1996: Made use of
  <xtmisc_init_app_initialise>.

    Updated by      Richard Gooch   27-DEC-1996: Made use of <ds_event_*>
  system. Switched to <dmx> package. Took account of change to
  <xtmisc_init_app_initialise>.

    Updated by      Richard Gooch   30-DEC-1996: Took account of another change
  to <xtmisc_init_app_initialise>.

    Updated by      Richard Gooch   31-DEC-1996: Took account of change to
  to Dataclip widget.

    Updated by      Richard Gooch   12-JAN-1997: Made use of new
  XkwNtrapDirectoryDatasets resource for FilewinWidgetClass.

    Updated by      Richard Gooch   27-JAN-1997: Took account of change to
  interface for <XkwTracePopupShowTrace> function.

    Updated by      Richard Gooch   28-JAN-1997: Took account of change back to
  previous interface for <XkwTracePopupShowTrace> function.

    Updated by      Richard Gooch   2-FEB-1997: No longer register pointer to
  KwcsAstro object with canvas.

    Updated by      Richard Gooch   12-FEB-1997: Only use GLISH "image" and
  "cube" events.

    Updated by      Richard Gooch   4-MAR-1997: Automatically recompute
  projection upon cube reload.

    Updated by      Richard Gooch   9-MAR-1997: Took account of change to
  <xtmisc_init_app_initialise>.

    Updated by      Richard Gooch   26-MAR-1997: Track display now also updates
  when dragging zoom box.

    Updated by      Richard Gooch   19-APR-1997: Now use internal file loader
  in Filewinpopup widget.

    Updated by      Richard Gooch   22-JUN-1997: Took account of change to
  <xtmisc_init_app_initialise> to double varargs list.

    Updated by      Richard Gooch   19-AUG-1997: Set resources in create call
  for second fileselector.

    Updated by      Richard Gooch   31-OCT-1997: Took account of change to
  <XkwDataclipNewArray>.

    Updated by      Richard Gooch   7-DEC-1997: Added font resource for new
  intensity menu and removed code to set pan centre for magnifier as it is now
  done automatically by the MagnifierPopup widget.

    Updated by      Richard Gooch   1-MAR-1998: Made use of new filewinWidget
  resource.

    Updated by      Richard Gooch   2-SEP-1998: Removed overlay code.

    Updated by      Richard Gooch   5-SEP-1998: Used new comms. setup.

    Updated by      Richard Gooch   17-SEP-1998: Use SimpleMenu class name in
  resources.

    Updated by      Richard Gooch   10-NOV-1998: Added support for TrueColour
  canvas only displays.

    Updated by      Richard Gooch   2-MAR-1999: Fixed support for TrueColour
  canvas only displays.

    Updated by      Richard Gooch   8-APR-1999: Added Dataclip control for
  transform window.

    Updated by      Richard Gooch   10-APR-1999: Added zero clipping.

    Updated by      Richard Gooch   13-APR-1999: Transform canvas event handler

    Updated by      Richard Gooch   16-APR-1999: Added "-image" and "-cube"
  command-line switches.

    Updated by      Richard Gooch   21-APR-1999: Remove region callback for
  transform window as the ImageDisplay widget already manages this.

    Updated by      Richard Gooch   22-APR-1999: Added button to dump
  parameters and command-line options to initialise them.

    Updated by      Richard Gooch   30-AUG-1999: Added dirname font resource.

    Updated by      Richard Gooch   31-AUG-1999: Took account of interface
  change to function <viewimg_track _compute>. Added 'l' key binding in
  function <track_canvas_event>.

    Updated by      Richard Gooch   9-SEP-1999: Removed track display code and
  made use of XkwNupdateTrackLabels resource for ImageDisplay widget.

    Updated by      Richard Gooch   11-DEC-1999: Removed unnecessary externs.

    Updated by      Richard Gooch   24-DEC-1999: Select visible canvases at
  initialisation.

    Updated by      Richard Gooch   22-JUN-2000: Update Z profile on all events
  rather than limiting to pointer move events. Invalidate Z profile when event
  lies outside cube boundaries.

    Updated by      Richard Gooch   26-JUN-2000: Do not update Z profile on
  pointer enter events.

    Updated by      Richard Gooch   11-MAY-2001: Changed dirname font resource.

    Updated by      Richard Gooch   14-JUN-2001: Changed <print_transform_info>
  to display in shell (aux) window as well. Increased number of track labels
  for shell window to accommodate change.

    Updated by      Richard Gooch   3-OCT-2001: Added translations for Filewin.

    Updated by      Richard Gooch   21-DEC-2001: Included k_version.h directly.

    Updated by      Richard Gooch   22-DEC-2002: Added #include <stdlib.h>

    Updated by      Richard Gooch   14-JUN-2003: #include cleanups.

    Last updated by Richard Gooch   4-OCT-2003: Added #include <Xkw/Dataclip.h>


*/
#include <stdlib.h>
#include <stdio.h>
#include <X11/Xatom.h>
#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/Xaw/Command.h>
#include <X11/Xaw/Form.h>
#include <X11/Xaw/Dialog.h>
#include <karma.h>
#include <k_version.h>
#define X11
#include <k_event_codes.h>
#include <karma_viewimg.h>
#include <karma_foreign.h>
#include <karma_xtmisc.h>
#include <karma_dsxfr.h>
#include <karma_ds.h>
#include <karma_im.h>
#include <karma_ic.h>
#include <karma_m.h>
#include <karma_c.h>
#include <Xkw/ImageDisplay.h>
#include <Xkw/Dialogpopup.h>
#include <Xkw/Filewin.h>
#include <Xkw/Ktoggle.h>
#include <Xkw/ExclusiveMenu.h>
#include <Xkw/MomentGenerator.h>
#include <Xkw/TracePopup.h>
#include <Xkw/Dataclip.h>
#ifdef OS_SunOS
#  include <floatingpoint.h>
#endif
#define NEED_IMAGE_MODE_ALTERNATIVES
#include "kshell.h"


#define VERSION "1.5.10"


/*  Private functions  */
STATIC_FUNCTION (flag main_canvas_event,
		 (ViewableImage vimage, double x, double y,
		  void *value, unsigned int event_code,
		  void *e_info, void **f_info,
		  double x_lin, double y_lin, unsigned int value_type,
		  double x_im, double y_im, int x_pix, int y_pix) );
STATIC_FUNCTION (void image_mode_cbk,
		 (Widget w, XtPointer client_data, XtPointer call_data) );
STATIC_FUNCTION (void show_image,
		 (KWorldCanvas canvas, iarray image,ViewableImage vimage) );
STATIC_FUNCTION (void show_mom0, (KWorldCanvas canvas) );
STATIC_FUNCTION (void show_mom1, (KWorldCanvas canvas) );
STATIC_FUNCTION (void moment_cbk,
		 (Widget w, XtPointer client_data, XtPointer call_data) );
STATIC_FUNCTION (void destroy_func, (iarray array, void *info) );
STATIC_FUNCTION (void show_channel0, (KWorldCanvas canvas) );
STATIC_FUNCTION (flag cube_event_func,
		 (void *object, multi_array *multi_desc,
		  CONST char *domain, CONST char *name) );
STATIC_FUNCTION (flag image_event_func,
		 (void *object, multi_array *multi_desc,
		  CONST char *domain, CONST char *name) );
STATIC_FUNCTION (void zero_clip_cbk,
		 (Widget w, XtPointer client_data, XtPointer call_data) );
STATIC_FUNCTION (void dump_cbk,
		 (Widget w, XtPointer client_data, XtPointer call_data) );


/*  Private data  */
String fallback_resources[] =
{
    "Kshell.geometry:                                      +0+0",
    "Kshell*pseudoColourCanvas*background:                 black",
    "Kshell*directColourCanvas*background:                 black",
    "Kshell*trueColourCanvas*background:                   black",
    "Kshell*pseudoColourStereoCanvas*background:           black",
    "Kshell*directColourStereoCanvas*background:           black",
    "Kshell*trueColourStereoCanvas*background:             black",
    "Kshell*pseudoColourCanvas*foreground:                 white",
    "Kshell*Command*background:                            grey70",
    "Kshell*Repeater*background:                           grey70",
    "Kshell*Ktoggle*background:                            grey80",
    "Kshell*closeButton*background:                        grey90",
    "Kshell*ChoiceMenu.background:                         turquoise",
    "Kshell*ExclusiveMenu.background:                      turquoise",
    "Kshell*Value*background:                              #d0a0a0",
    "Kshell*ImageDisplay*quit*background:                  orange",
    "Kshell*SimpleSlider.foreground:                       Sea Green",
    "Kshell*ScrollableList.highlightPixel:                 Sea Green",
    "Kshell*SimpleProgressMeter.background:                turquoise3",
    "Kshell*SimpleProgressMeter.foreground:                grey60",
    "Kshell*auxShell.geometry:                             -0+0",
    "Kshell*ImageDisplay*trackLabel0*font:                 8x13bold",
    "Kshell*ImageDisplay*trackLabel1*font:                 8x13bold",
    "Kshell*ImageDisplay*trackLabel2*font:                 8x13bold",
    "Kshell*ImageDisplay*trackLabel3*font:                 8x13bold",
    "Kshell*ExclusiveMenu.font:                            8x13bold",
    "Kshell*SimpleMenu*font:                               10x20",
    "Kshell*ImageDisplay.cmapSize:                         100",
    "Kshell*ImageDisplay.Command.font:                     7x13bold",
    "Kshell*ImageDisplay.applicationBox.Command.font:      7x13bold",
    "Kshell*ImageDisplay.applicationBox.Ktoggle.font:      7x13bold",
    "Kshell*ImageDisplay.ChoiceMenu.font:                  7x13bold",
    "Kshell*ImageDisplay.exportMenu.menuButton.font:       7x13bold",
    "Kshell*ImageDisplay.overlayMenu.menuButton.font:      7x13bold",
    "Kshell*ImageDisplay*zoomMenu*Unzoom*foreground:       red",
    "Kshell*Dataclip.AutoValueScale:                       False",
    "Kshell*SimpleSlider.borderWidth:                      0",
    "Kshell*IntensityPolicy*autoIntensityScale:            False",
    "Kshell*ZoomPolicy*integerXZoom:                       False",
    "Kshell*ZoomPolicy*integerYZoom:                       False",
    "Kshell*Filewin.dirname.textSink.font:                 6x13",
    "Kshell*font:                                          9x15bold",
    "Kshell*borderColor:                                   black",
    "Kshell*background:                                    aquamarine",
    "Kshell*foreground:                                    black",
    NULL
};
static XrmOptionDescRec Options[] =
{
    {"-private_cmap", ".topForm.multiCanvas.pseudoColourCanvas.forceNewCmap",
     XrmoptionNoArg, (XPointer) "True"},
    {"-num_colours", ".topForm.cmapSize", XrmoptionSepArg, (XPointer) NULL},
    {"-fullscreen", ".topForm.fullscreen", XrmoptionNoArg, (XPointer) "True"},
    {"-verbose", "*verbose", XrmoptionNoArg, (XPointer) "True"},
};
static Widget main_shell = NULL;
static Widget main_image_display = NULL;
static Widget aux_image_display = NULL;
static ViewableImage magnified_image = NULL;
static KWorldCanvas main_canvas = NULL;
static Widget moment_winpopup = NULL;
static Widget trace_winpopup = NULL;
static ViewableImage channel0_vimage = NULL;
static double cube_min = TOOBIG;
static double cube_max = -TOOBIG;


/*  Public functions follow  */

int main (int argc, char **argv)
{
    KWorldCanvas pc_canvas, tc_canvas, aux_canvas;
    int argcount;
    XtAppContext app_context;
    Widget filewin, filepopup, w, box, files_btn;
    Widget aux_shell, zero_clip_tgl;
    Widget menu_btn, trace_btn;
    extern flag zero_clip;
    extern double centre_x, centre_y, radius;
    extern char title_name[STRING_LENGTH];

#ifdef SIGFPE_ABORT
	ieee_handler ("set", "division", SIGFPE_ABORT);
#endif
    /*  Initialise module  */
    im_register_module_name ("kshell");
    im_register_module_version_date (VERSION);
    im_register_lib_version (KARMA_VERSION);
    /*  Start up Xt  */
    main_shell = xtmisc_init_app_initialise (&app_context, "Kshell",
					     Options, XtNumber (Options),
					     &argc, argv, fallback_resources,
					     XTMISC_INIT_ATT_MIN_CCELLS, 200,
					     XTMISC_INIT_ATT_COMMS_SETUP, TRUE,
					     XTMISC_INIT_ATT_BECOME_SERVER,
					     TRUE,
					     XTMISC_INIT_ATT_TITLE, title_name,
					     XTMISC_INIT_ATT_END,
					     NULL);
    dsxfr_register_connection_limits (1, -1);
    xtmisc_set_icon (main_shell, ic_write_kimage_icon);
    main_image_display = XtVaCreateManagedWidget ("topForm",
						  imageDisplayWidgetClass,
						  main_shell,
						  XtNborderWidth, 0,
						  XkwNnumTrackLabels, 4,
						  XkwNupdateTrackLabels, TRUE,
						  NULL);
    if ( ( w = XtNameToWidget (main_image_display, "filesButton") ) != NULL )
    {
	/*  Force main "Files" button to be re-labelled as "Cubes"  */
	XtVaSetValues (w,
		       XtNlabel, "Cubes",
		       NULL);
    }
    filepopup = XtNameToWidget (main_image_display, "filewinPopup");
    XtVaGetValues (filepopup, XkwNfilewinWidget, &filewin, NULL);
    XtVaSetValues (filepopup,
		   XkwNfilenameTester, XkwFilewinStandardFileTester_3D,
		   XkwNpreferFloat, TRUE,
		   XkwNdomainName, "CUBE",
		   NULL);
    XtVaSetValues (filewin,
		   XkwNtrapDirectoryDatasets, True,
		   NULL);
    /*  Get the application box  */
    box = XtNameToWidget (main_image_display, "applicationBox");
    /*  Create the Image file selector and popup button  */
    filepopup = XtVaCreatePopupShell ("filewinPopup", filepopupWidgetClass,
				      main_shell,
				      XtNtitle, "Kshell Image Selector",
				      XtNx, 0,
				      XtNy, 0,
				      XkwNautoPopdown, True,
				      XkwNfilenameTester,
				      XkwFilewinStandardFileTester_nD,
				      XkwNpreferFloat, TRUE,
				      XkwNdomainName, "IMAGE",
				      XkwNloadFiles, TRUE,
				      NULL);
    files_btn = XtVaCreateManagedWidget ("filesButton", commandWidgetClass,box,
					 XtNlabel, "Images",
					 NULL);
    XtAddCallback (files_btn, XtNcallback, xtmisc_popup_cbk, filepopup);
    XtVaGetValues (filepopup, XkwNfilewinWidget, &filewin, NULL);
    XtVaSetValues (filewin,
		   XkwNtrapDirectoryDatasets, True,
		   NULL);
    ds_event_register_func (image_event_func, NULL);
    /*  Create the image mode menu  */
    w = XtVaCreateManagedWidget ("menuButton", exclusiveMenuWidgetClass, box,
				 XtNmenuName, "imageModeMenu",
				 XkwNchoiceName, "Image Mode",
				 XkwNnumItems, NUM_IMAGE_MODE_ALTERNATIVES,
				 XkwNitemStrings, image_mode_alternatives,
				 NULL);
    menu_btn = w;
    w = XtVaCreateManagedWidget ("traceButton", commandWidgetClass, box,
				 XtNlabel, "Profile",
				 NULL);
    trace_btn = w;
    w = XtVaCreateManagedWidget ("button", commandWidgetClass, box,
				 XtNlabel, "Dump Parameters",
				 NULL);
    XtAddCallback (w, XtNcallback, dump_cbk, NULL);
    XtRealizeWidget (main_shell);
    XtVaGetValues (main_image_display,
		   XkwNpseudoColourCanvas, &pc_canvas,
		   XkwNtrueColourCanvas, &tc_canvas,
		   NULL);
    main_canvas = (pc_canvas == NULL) ? tc_canvas : pc_canvas;
    if (main_canvas == NULL)
    {
	fprintf (stderr, "PseudoColour or TrueColour canvas not found!\n");
	exit (RV_UNDEF_ERROR);
    }
    XtVaSetValues (main_image_display, XkwNvisibleCanvas, main_canvas, NULL);
    XtAddCallback (menu_btn, XkwNselectCallback, image_mode_cbk, main_canvas);
    viewimg_register_position_event_func (main_canvas,
					  ( flag (*) () ) main_canvas_event,
					  (void *) main_image_display);
    aux_shell = XtVaCreatePopupShell ("auxShell", topLevelShellWidgetClass,
				      main_shell,
				      XtNtitle, "Slice Window",
				      NULL);
    aux_image_display = XtVaCreateManagedWidget ("topForm",
						 imageDisplayWidgetClass,
						 aux_shell,
						 XtNborderWidth, 0,
						 XkwNcreateFilepopup, FALSE,
                                                 XkwNnumTrackLabels, 4,
						 XkwNupdateTrackLabels, TRUE,
						 XkwNshowQuitButton, FALSE,
						 NULL);
    /*  Get the application box  */
    box = XtNameToWidget (aux_image_display, "applicationBox");
    w = XtVaCreateManagedWidget ("zeroToggle", ktoggleWidgetClass, box,
				 XtNlabel, "Zero Clip",
				 XkwNcrosses, False,
				 XtNstate, zero_clip,
				 NULL);
    zero_clip_tgl = w;
    XtRealizeWidget (aux_shell);
    XtVaGetValues (aux_image_display,
		   XkwNpseudoColourCanvas, &pc_canvas,
		   XkwNtrueColourCanvas, &tc_canvas,
		   NULL);
    aux_canvas = (pc_canvas == NULL) ? tc_canvas : pc_canvas;
    if (aux_canvas == NULL)
    {
	fprintf (stderr, "PseudoColour or TrueColour canvas not found!\n");
	exit (RV_UNDEF_ERROR);
    }
    XtVaSetValues (aux_image_display, XkwNvisibleCanvas, aux_canvas, NULL);
    ds_event_register_func (cube_event_func, aux_canvas);
    canvas_register_position_event_func (aux_canvas,
					 ( flag (*) () ) transform_position_func,
					 NULL);
    /*  This draws the ellipse & calls transform */
    canvas_register_position_event_func (main_canvas,
					 ( flag (*) () ) main_position_func,
					 aux_canvas);
    canvas_register_refresh_func (main_canvas, main_refresh_func, NULL);
    XtPopup (aux_shell, XtGrabNone);
    moment_winpopup = XtVaCreatePopupShell ("momentwinpopup",
					    momentGeneratorWidgetClass,
					    main_shell,
					    XtNtitle,
					    "Moment Generator Control Window",
					    NULL);
    XtAddCallback (moment_winpopup, XkwNmomentCallback, moment_cbk, NULL);
    XtRealizeWidget (moment_winpopup);
    trace_winpopup = XtVaCreatePopupShell ("tracewinpopup",
					   tracePopupWidgetClass, main_shell,
					   XtNtitle, "Profile Window",
					   NULL);
    XtAddCallback (trace_btn, XtNcallback, xtmisc_popup_cbk, trace_winpopup);
    XtAddCallback (zero_clip_tgl, XtNcallback, zero_clip_cbk, aux_canvas);
    /*  Option processing after widgets created and realised  */
    for (argcount = 1; argcount < argc; ++argcount)
    {
	if ( (strcmp (argv[argcount], "-image") == 0) ||
	     (strcmp (argv[argcount], "-cube") == 0) )
	{
	    multi_array *multi_desc;
	    char txt[STRING_LENGTH], fname[STRING_LENGTH];

	    if (argcount + 1 >= argc)
	    {
		fprintf (stderr, "File missing for dataset\n");
		continue;
	    }
	    if (strcmp (argv[argcount], "-image") == 0) strcpy (txt, "IMAGE");
	    else strcpy (txt, "CUBE");
	    ++argcount;
	    if (strncmp (argv[argcount], "./", 2) == 0)
	    {
		strcpy (fname, argv[argcount] + 2);
	    }
	    else strcpy (fname, argv[argcount]);
	    if ( ( multi_desc =
		   foreign_guess_and_read (fname, K_CH_MAP_LOCAL,
					   FALSE, NULL,
					   FA_GUESS_READ_END) )
		 == NULL )
	    {
		fprintf (stderr, "Error reading file: \"%s\"\n", fname);
		continue;
	    }
	    ds_event_dispatch (multi_desc, txt, fname);
	    ds_dealloc_multi (multi_desc);
	    m_clear (fname, STRING_LENGTH);  /*  Force me to take care  */
	}
	else if (strcmp (argv[argcount], "-centre_x") == 0)
	{
	    if (argcount + 1 >= argc)
	    {
		fprintf (stderr, "Value missing for option: \"%s\"\n",
			 argv[argcount]);
		continue;
	    }
	    sscanf (argv[++argcount], "%lg", &centre_x);
	}
	else if (strcmp (argv[argcount], "-centre_y") == 0)
	{
	    if (argcount + 1 >= argc)
	    {
		fprintf (stderr, "Value missing for option: \"%s\"\n",
			 argv[argcount]);
		continue;
	    }
	    sscanf (argv[++argcount], "%lg", &centre_y);
	}
	else if (strcmp (argv[argcount], "-radius") == 0)
	{
	    if (argcount + 1 >= argc)
	    {
		fprintf (stderr, "Value missing for option: \"%s\"\n",
			 argv[argcount]);
		continue;
	    }
	    sscanf (argv[++argcount], "%lg", &radius);
	}
	else fprintf (stderr, "Option: \"%s\" ignored\n", argv[argcount]);
    }
    XtAppMainLoop (app_context);
    return (RV_OK);
}   /*  End Function main   */

void print_transform_info (CONST char *window_string)
/*  [SUMMARY] Print information about the slice.
    <window_string> The string to display.
    [RETURNS] Nothing.
*/
{
    Widget label;

    label = XtNameToWidget (main_image_display, "trackLabel3");
    XtVaSetValues (label,
		   XtNlabel, window_string,
		   NULL);
    label = XtNameToWidget (aux_image_display, "trackLabel3");
    XtVaSetValues (label,
		   XtNlabel, window_string,
		   NULL);
}   /*  End Function print_transform_info  */

void update_transform_dataclip (iarray array, double min, double max)
/*  [SUMMARY] Update the transform Dataclip widget.
    <array> The new transform image.
    <min> The minimum image value.
    <max> The maximum image value.
    [RETURNS] Nothing.
*/
{
    Widget izoomwinpopup;

    izoomwinpopup = XtNameToWidget (aux_image_display, "izoomwinpopup");
    XkwDataclipNewArray (izoomwinpopup, array, min, max, TRUE);
}   /*  End Function update_transform_dataclip  */


/*  Private functions follow  */

static flag main_canvas_event (ViewableImage vimage, double x, double y,
			       void *value, unsigned int event_code,
			       void *e_info, void **f_info,
			       double x_lin, double y_lin,
			       unsigned int value_type,
			       double x_im, double y_im, int x_pix, int y_pix)
/*  [PURPOSE] This routine is a position event consumer for a world canvas
    which has a number of ViewableImage objects associated with it.
    <viewimg> The active viewable image.
    <x> The horizontal world co-ordinate of the event.
    <y> The vertical world co-ordinate of the event.
    These values will have been transformed by the registered transform
    function (see  canvas_register_transform_func  ) for the associated
    world canvas.
    <value> A pointer to the data value in the viewable image corresponding
    to the event co-ordinates.
    <event_code> The arbitrary event code.
    <e_info> The arbitrary event information.
    <f_info> The arbitrary function information pointer.
    <x_lin> The linear horizontal world co-ordinate (the co-ordinate prior
    to the transform function being called).
    <y_lin> The linear vertical world co-ordinate (the co-ordinate prior
    to the transform function being called).
    <value_type> The type of the data value. This may be K_DCOMPLEX or
    K_UB_RGB.
    <x_im> The horizontal image pixel co-ordinate. 0.0 is the pixel centre.
    <y_im> The vertical image pixel co-ordinate. 0.0 is the pixel centre.
    <x_pix> The horizontal screen pixel co-ordinate.
    <y_pix> The vertical screen pixel co-ordinate.
    [RETURNS] TRUE if the event was consumed, else FALSE indicating that
    the event is still to be processed.
*/
{
    KwcsAstro main_ap, cube_ap;
    double cube_x, cube_y, min, max;
    dim_desc *dim;
    uaddr coords[2];
    unsigned int dim_indices[2];
    extern iarray cube_arr;
    /*static char function_name[] = "main_canvas_event";*/

    if (event_code == K_CANVAS_EVENT_POINTER_ENTER) return (FALSE);
    if (cube_arr == NULL) return (FALSE);
    cube_ap = iarray_get_associated_wcs_astro (cube_arr);
    viewimg_get_attributes (vimage,
			    VIEWIMG_VATT_WCS_ASTRO, &main_ap,
			    VIEWIMG_VATT_END);
    convert_lin_world_coords (&cube_x, &cube_y, cube_ap,
			      &x_lin, &y_lin, main_ap, 1);
    dim_indices[0] = 2;
    dim_indices[1] = 1;
    dim = iarray_get_dim_desc (cube_arr, 2);
    if (dim->first_coord < dim->last_coord)
    {
	min = dim->first_coord;
	max = dim->last_coord;
    }
    else
    {
	min = dim->last_coord;
	max = dim->first_coord;
    }
    if ( (cube_x < min) || (cube_x > max) )
    {
	XkwTracePopupInvalidate (trace_winpopup);
	return (FALSE);
    }
    coords[0] = ds_get_coord_num (dim, cube_x, SEARCH_BIAS_CLOSEST);
    dim = iarray_get_dim_desc (cube_arr, 1);
    if (dim->first_coord < dim->last_coord)
    {
	min = dim->first_coord;
	max = dim->last_coord;
    }
    else
    {
	min = dim->last_coord;
	max = dim->first_coord;
    }
    if ( (cube_y < min) || (cube_y > max) )
    {
	XkwTracePopupInvalidate (trace_winpopup);
	return (FALSE);
    }
    coords[1] = ds_get_coord_num (dim, cube_y, SEARCH_BIAS_CLOSEST);
    XkwTracePopupShowTrace (trace_winpopup, dim_indices, coords);
    return (FALSE);
}   /*  End Function main_canvas_event  */

static void image_mode_cbk (Widget w, XtPointer client_data,
			    XtPointer call_data)
/*  [SUMMARY] Image mode change callback.
    <w> The widget the event occurred on.
    <client_data> Client data.
    <call_data> Call data.
    [RETURNS] Nothing.
*/
{
    KWorldCanvas canvas = (KWorldCanvas) client_data;
    unsigned int code = *(int *) call_data;
    extern ViewableImage image_vimage;
    extern iarray image_arr;
    extern KwcsAstro main_ap, loaded_image_ap, cube_ap;
    extern unsigned int image_mode;
    /*static char function_name[] = "image_mode_cbk";*/

    image_mode = code;
    switch (code)
    {
      case IMAGE_MODE_LOADED:
	show_image (canvas, image_arr, image_vimage);
	main_ap = loaded_image_ap;
	break;
      case IMAGE_MODE_MOM0:
	XtPopup (moment_winpopup, XtGrabNone);
	show_mom0 (canvas);
	main_ap = cube_ap;
	break;
      case IMAGE_MODE_MOM1:
	XtPopup (moment_winpopup, XtGrabNone);
	show_mom1 (canvas);
	main_ap = cube_ap;
	break;
      case IMAGE_MODE_CHANNEL0:
	show_channel0 (canvas);
	main_ap = cube_ap;
	break;
    }
}   /*  End Function image_mode_cbk  */

static void show_image (KWorldCanvas canvas, iarray image,ViewableImage vimage)
/*  [SUMMARY] Show the loaded image.
    <canvas> The world canvas.
    <image> The image array.
    <vimage> The image.
    [RETURNS] Nothing.
*/
{
    double i_min, i_max;
    /*static char function_name[] = "show_image";*/

    if (vimage == NULL) return;
    if (viewimg_get_active (canvas) == vimage) return;
    /*  Must change the canvas intensity scale  */
    if ( !iarray_min_max (image, CONV1_REAL, &i_min, &i_max) )
    {
	fprintf (stderr, "Error computing min-max\n");
    }
    canvas_set_attributes (canvas,
			   CANVAS_ATT_VALUE_MIN, i_min,
			   CANVAS_ATT_VALUE_MAX, i_max,
			   CANVAS_ATT_END);
    viewimg_make_active (vimage);
}   /*  End Function show_image  */

static void show_mom0 (KWorldCanvas canvas)
/*  [SUMMARY] Show the 0th moment map.
    <canvas> The world canvas.
    [RETURNS] Nothing.
*/
{
    double i_min, i_max;
    extern iarray mom0_arr;
    extern ViewableImage mom0_vimage;
    /*static char function_name[] = "show_mom0";*/

    if (mom0_vimage == NULL) return;
    if (viewimg_get_active (canvas) == mom0_vimage) return;
    /*  Must change the canvas intensity scale  */
    if ( !iarray_min_max (mom0_arr, CONV1_REAL, &i_min, &i_max) )
    {
	fprintf (stderr, "Error computing min-max\n");
    }
    if (i_min >= i_max)
    {
	fprintf (stderr, "Computed 0th moment map is blank: ignoring\n");
	return;
    }
    canvas_set_attributes (canvas,
			   CANVAS_ATT_VALUE_MIN, i_min,
			   CANVAS_ATT_VALUE_MAX, i_max,
			   CANVAS_ATT_END);
    viewimg_make_active (mom0_vimage);
}   /*  End Function show_mom0  */

static void show_mom1 (KWorldCanvas canvas)
/*  [SUMMARY] Show the 1st moment map.
    <canvas> The world canvas.
    [RETURNS] Nothing.
*/
{
    double i_min, i_max;
    extern iarray mom1_arr;
    extern ViewableImage mom0_vimage, mom1_vimage;
    /*static char function_name[] = "show_mom1";*/

    if (mom1_vimage == NULL) return;
    if ( (viewimg_get_active (canvas) == NULL) ||
	 (viewimg_get_active (canvas) == mom0_vimage) )
    {
	/*  Must change the canvas intensity scale  */
	if ( !iarray_min_max (mom1_arr, CONV1_REAL, &i_min, &i_max) )
	{
	    fprintf (stderr, "Error computing min-max\n");
	}
	if (i_min >= i_max)
	{
	    fprintf (stderr, "Computed 1st moment map is blank: ignoring\n");
	    return;
	}
	canvas_set_attributes (canvas,
			       CANVAS_ATT_VALUE_MIN, i_min,
			       CANVAS_ATT_VALUE_MAX, i_max,
			       CANVAS_ATT_END);
    }
    viewimg_make_active (mom1_vimage);
}   /*  End Function show_mom1  */

static void moment_cbk (Widget w, XtPointer client_data, XtPointer call_data)
/*  [SUMMARY] Moment computation callback.
    <w> The widget the event occurred on.
    <client_data> Client data.
    <call_data> Call data.
    [RETURNS] Nothing.
*/
{
    extern iarray cube_arr;
    extern ViewableImage mom0_vimage, mom1_vimage;
    extern unsigned int image_mode;
    /*static char function_name[] = "moment_cbk";*/

    if (cube_arr == NULL) return;
    viewimg_register_data_change (mom0_vimage);
    viewimg_register_data_change (mom1_vimage);
    if (image_mode == IMAGE_MODE_MOM0) show_mom0 (main_canvas);
    if (image_mode == IMAGE_MODE_MOM1) show_mom1 (main_canvas);
}   /*  End Function moment_cbk  */

static void destroy_func (iarray array, void *info)
/*  [SUMMARY] Register destruction of an Intelligent Array.
    <array> The Intelligent Array.
    <info> A pointer to the arbitrary information.
    [RETURNS] Nothing.
*/
{
    ViewableImage *vimage = info;

    viewimg_destroy (*vimage);
    *vimage = NULL;
}   /*  End Function destroy_func  */

static void show_channel0 (KWorldCanvas canvas)
/*  [SUMMARY] Show the 0th channel map.
    <canvas> The world canvas.
    [RETURNS] Nothing.
*/
{
    /*static char function_name[] = "show_channel0";*/

    if (channel0_vimage == NULL) return;
    if (viewimg_get_active (canvas) == channel0_vimage) return;
    /*  Must change the canvas intensity scale  */
    canvas_set_attributes (canvas,
			   CANVAS_ATT_VALUE_MIN, cube_min,
			   CANVAS_ATT_VALUE_MAX, cube_max,
			   CANVAS_ATT_END);
    viewimg_make_active (channel0_vimage);
}   /*  End Function show_channel0  */

static flag cube_event_func (void *object, multi_array *multi_desc,
			     CONST char *domain, CONST char *name)
/*  [SUMMARY] Process a new cube event.
    <object> The object information pointer.
    <multi_desc> The multi_array descriptor.
    <domain> The domain from where the event was generated.
    <name> The name of the event in its domain.
    [RETURNS] TRUE if further callbacks should not be called, else FALSE.
*/
{
    KWorldCanvas aux_canvas = object;
    char filename[STRING_LENGTH];
    extern iarray cube_arr, mom0_arr, mom1_arr;
    extern ViewableImage mom0_vimage, mom1_vimage;
    extern KwcsAstro main_ap, cube_ap;
    extern unsigned int image_mode;
    static char function_name[] = "cube_event_func";

    if (strcmp (domain, "IMAGE") == 0) return (FALSE);
    if ( (strcmp (domain, "GLISH") == 0) && (strcmp (name, "cube") != 0) )
	return (FALSE);
    if (strcmp (domain, "CUBE") == 0) filename[0] = '\0';
    else
    {
	strcpy (filename, domain);
	strcat (filename, "::");
    }
    strcat (filename, name);
    if ( !process_cube (multi_desc, &cube_arr, &cube_min, &cube_max) )
	return (FALSE);
    XkwMomentGeneratorNewArray (moment_winpopup, cube_arr, cube_min, cube_max);
    XkwTracePopupNewArray (trace_winpopup, cube_arr, cube_min, cube_max);
    if (channel0_vimage != NULL) viewimg_destroy (channel0_vimage);
    if ( ( channel0_vimage = viewimg_create_restr (main_canvas,
						   cube_arr->multi_desc,
						   cube_arr->arr_desc,
						   cube_arr->data,
						   2, 1, 0,
						   0, NULL, NULL) ) == NULL )
    {
	m_abort (function_name, "0th channel ViewableImage");
    }
    viewimg_set_attributes (channel0_vimage,
			    VIEWIMG_VATT_VALUE_MIN, cube_min,
			    VIEWIMG_VATT_VALUE_MAX, cube_max,
			    VIEWIMG_VATT_END);
    if (mom0_vimage != NULL)
    {
	fprintf (stderr,"WARNING: 0th moment ViewableImage not deallocated\n");
    }
    if (mom1_vimage != NULL)
    {
	fprintf (stderr,"WARNING: 1st moment ViewableImage not deallocated\n");
    }
    XtVaGetValues (moment_winpopup,
		   XkwNmom0Array, &mom0_arr,
		   XkwNmom1Array, &mom1_arr,
		   NULL);
    if ( ( mom0_vimage = viewimg_create_from_iarray (main_canvas, mom0_arr,
						     FALSE) ) == NULL )
    {
	m_abort (function_name, "0th moment ViewableImage");
    }
    if ( ( mom1_vimage = viewimg_create_from_iarray (main_canvas, mom1_arr,
						     FALSE) ) == NULL )
    {
	m_abort (function_name, "1st moment ViewableImage");
    }
    iarray_register_destroy_func (mom0_arr, ( flag (*) () ) destroy_func,
				  &mom0_vimage);
    iarray_register_destroy_func (mom1_arr, ( flag (*) () ) destroy_func,
				  &mom1_vimage);
    if ( (image_mode == IMAGE_MODE_MOM0) ||
	 (image_mode == IMAGE_MODE_MOM1) )
    {
	main_ap = cube_ap;
    }
    update_projection (aux_canvas);
    return (TRUE);
}   /*  End Function cube_event_func  */

static flag image_event_func (void *object, multi_array *multi_desc,
			      CONST char *domain, CONST char *name)
/*  [SUMMARY] Process a new image event.
    <object> The object information pointer.
    <multi_desc> The multi_array descriptor.
    <domain> The domain from where the event was generated.
    <name> The name of the event in its domain.
    [RETURNS] TRUE if further callbacks should not be called, else FALSE.
*/
{
    KWorldCanvas canvas, mag_canvas;
    double min, max;
    Widget izoomwinpopup, magnifier;
    char *ptr;
    char filename[STRING_LENGTH], stripped_filename[STRING_LENGTH];
    char title[STRING_LENGTH];
    extern ViewableImage image_vimage;
    extern iarray image_arr;
    extern char title_name[STRING_LENGTH];

    if (strcmp (domain, "CUBE") == 0) return (FALSE);
    if ( (strcmp (domain, "GLISH") == 0) && (strcmp (name, "image") != 0) )
	return (FALSE);
    if (strcmp (domain, "IMAGE") == 0) filename[0] = '\0';
    else
    {
	strcpy (filename, domain);
	strcat (filename, "::");
    }
    strcat (filename, name);
    strcpy (stripped_filename, filename);
    if ( ( ptr = strrchr (stripped_filename, '.') ) != NULL )
    {
	if (strcmp (ptr, ".kf") == 0) *ptr = '\0';
    }
    XtVaGetValues (main_image_display,
		   XkwNvisibleCanvas, &canvas,
		   XkwNmagnifier, &magnifier,
		   NULL);
    XtVaGetValues (magnifier, XkwNvisibleCanvas, &mag_canvas, NULL);
    izoomwinpopup = XtNameToWidget (main_image_display, "izoomwinpopup");
    if ( !process_image (multi_desc, &image_arr, &min, &max,
			 canvas, mag_canvas,
			 &image_vimage, &magnified_image) ) return (FALSE);
    sprintf (title, "%s  file: %s\n", title_name, filename);
    XtVaSetValues (main_shell,
		   XtNtitle, title,
		   NULL);
    XtVaSetValues (main_image_display,
		   XkwNimageName, stripped_filename,
		   NULL);
    XkwDataclipNewArray (izoomwinpopup, image_arr, min, max, TRUE);
    return (TRUE);
}   /*  End Function image_event_func  */

static void zero_clip_cbk (Widget w, XtPointer client_data,XtPointer call_data)
/*  This is the zero clip callback.
*/
{
    KWorldCanvas aux_canvas = client_data;
    extern flag zero_clip;

    zero_clip = (uaddr) call_data;
    update_projection (aux_canvas);
}   /*  End Function zero_clip_cbk   */

static void dump_cbk (Widget w, XtPointer client_data,XtPointer call_data)
/*  This is the dump parameters callback.
*/
{
    extern double centre_x, centre_y, radius;

    printf ("CX: %g  CY: %g  R: %g\n", centre_x, centre_y, radius);
}   /*  End Function dump_cbk   */
