/*  kspherical-grid.c

    Source file for  kspherical-grid  (generate a spherical grid).

    Copyright (C) 1997-1999  Richard Gooch

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Richard Gooch may be reached by email at  karma-request@atnf.csiro.au
    The postal address is:
      Richard Gooch, c/o ATNF, P. O. Box 76, Epping, N.S.W., 2121, Australia.
*/

/*  This Karma module will generate an ASCII file containing line segments
    which make up a spherical grid. This file may be read into the <xray>
    programme as a segment file.


    Written by      Richard Gooch   3-SEP-1997

    Updated by      Richard Gooch   8-SEP-1997

    Updated by      Richard Gooch   17-SEP-1998: Took account of some CONST
  declarations in <ex> package.

    Last updated by Richard Gooch   15-AUG-1999: Made egcs happy.


*/
#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <karma.h>
#include <karma_ch.h>
#include <karma_ex.h>

int main (int argc, char **argv)
{
    Channel ch;
    unsigned int lon_count, lat_count /*, seg_count*/;
    unsigned int num_lon = 12;
    /*unsigned int num_lat = 6;*/
    unsigned int num_seg = 50;
    float cx, cy, cz, rx, ry, rz;
    float longitude, latitude, seg_step;
    float x1, y1, z1, x2, y2, z2, tmp;
    CONST char *p;
    static char usage_string[] =
	"Usage:\tkspherical-grid outfile Cx Cy Cz Rx Ry Rz";

    if (argc != 8)
    {
	fprintf (stderr, "%s\n", usage_string);
	exit (RV_BAD_PARAM);
    }
    cx = ex_float (argv[2], &p);
    cy = ex_float (argv[3], &p);
    cz = ex_float (argv[4], &p);
    rx = ex_float (argv[5], &p);
    ry = ex_float (argv[6], &p);
    rz = ex_float (argv[7], &p);
    if ( ( ch = ch_open_file (argv[1], "w") ) == NULL )
    {
	fprintf (stderr, "Error opening file: \"%s\"\t%s\n",
		 argv[1], ERRSTRING);
	exit (RV_CANNOT_OPEN);
    }
    /*  Write lines of longitude  */
    seg_step = PI / (float) num_seg;
    for (lon_count = 0; lon_count < num_lon; ++lon_count)
    {
	longitude = TWOPI * (float) lon_count / (float) num_lon;
	for (lat_count = 0; lat_count < num_seg; ++lat_count)
	{
	    latitude = PI_ON_2 - seg_step * (float) lat_count;
	    tmp = cos (latitude);
	    x1 = cx + tmp * rx * cos (longitude);
	    y1 = cy + tmp * ry * sin (longitude);
	    z1 = cz + rz * sin (latitude);
	    tmp = cos (latitude + seg_step);
	    x2 = cx + tmp * rx * cos (longitude);
	    y2 = cy + tmp * ry * sin (longitude);
	    z2 = cz + rz * sin (latitude + seg_step);
	    ch_printf (ch, "%g %g %g  %g %g %g\n",
		       x1, y1, z1, x2, y2, z2);
	}
    }
    return (RV_OK);
}   /*  End Function main  */
