/*LINTLIBRARY*/
/*  RenderDisplayPopup.c

    This code provides a render display widget for Xt.

    Copyright (C) 1997-2000  Richard Gooch

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Richard Gooch may be reached by email at  karma-request@atnf.csiro.au
    The postal address is:
      Richard Gooch, c/o ATNF, P. O. Box 76, Epping, N.S.W., 2121, Australia.
*/

/*

    This file contains all routines needed for a render display widget for Xt.


    Written by      Richard Gooch   19-SEP-1997

    Updated by      Richard Gooch   17-JUN-1998: First cut at generic
  implementation.

    Updated by      Richard Gooch   28-JUL-1998: Grab min&max from parent
  canvas for PseudoColour rendering.

    Updated by      Richard Gooch   31-AUG-1998: Supported delayed creation.

    Last updated by Richard Gooch   12-JUL-2000: Added #include <karma_kwin.h>.


*/
#include <stdio.h>
#include <stdlib.h>
#include <X11/Xos.h>
#include <X11/IntrinsicP.h>
#include <X11/StringDefs.h>
#ifndef X11
#  define X11
#endif
#include <karma.h>
#include <karma_vrender.h>
#include <karma_viewimg.h>
#include <karma_iarray.h>
#include <karma_kwin.h>
#include <karma_ds.h>
#include <karma_a.h>
#include <karma_m.h>
#include <karma_c.h>
/*#include <Xkw/RenderDisplayPopupP.h>*/
#include "RenderDisplayPopupP.h"
#include "xraycast_3d.h"


/*  Private functions  */
STATIC_FUNCTION (void RenderDisplayPopup__Initialise,
		 (Widget request, Widget new) );
STATIC_FUNCTION (void realised_cbk,
		 (Widget parent, XtPointer client_data, XtPointer call_data) );
STATIC_FUNCTION (void popup_cbk,
		 (Widget w, XtPointer client_data, XtPointer call_data) );
STATIC_FUNCTION (Boolean work_proc, (XtPointer client_data) );
STATIC_FUNCTION (void image_desc_notify_func,
		 (KVolumeRenderContext context, void **info) );
STATIC_FUNCTION (void RenderDisplayPopup__render,
		 (RenderDisplayPopupWidget top, flag stereo) );
STATIC_FUNCTION (flag XkwRenderDisplayPopupChangeCallback,
		 (void *object, void *client1_data,
		  void *call_data, void *client2_data) );


#define offset(field) XtOffset(RenderDisplayPopupWidget, renderDisplayPopup.field)

static XtResource RenderDisplayPopupResources[] = 
{
    {XkwNfastPanner, XkwCFastPanner, XtRBool, sizeof (Bool),
     XtOffset (RenderDisplayPopupWidget, slaveImageDisplayPopup.fastPanner),
     XtRImmediate, (XtPointer) False},
    {XkwNvolumeRenderContext, XkwCVolumeRenderContext, XtRPointer,
     sizeof (XtPointer), offset (vrend_ctx), XtRImmediate, NULL},
    {XkwNpreRenderFunction, XkwCPreRenderFunction, XtRPointer,
     sizeof (XtPointer), offset (preRenderFunction), XtRImmediate, NULL},
};

#undef offset

RenderDisplayPopupClassRec renderDisplayPopupClassRec = 
{
  {
      /* CoreClassPart */
    (WidgetClass) &slaveImageDisplayPopupClassRec, /* superclass */
    "RenderDisplayPopup",                          /* class_name */
    sizeof (RenderDisplayPopupRec),                /* widget_size */
    NULL,                                          /* class_initialise */
    NULL,                                          /* class_part_initialise */
    FALSE,                                         /* class_init */
    (XtInitProc) RenderDisplayPopup__Initialise,   /* initialise */
    NULL,                                          /* initialise_hook */
    XtInheritRealize,                              /* realise */
    NULL,                                          /* actions */
    0,                                             /* num_actions */
    RenderDisplayPopupResources,                   /* resources */
    XtNumber (RenderDisplayPopupResources),        /* num_resources */
    NULLQUARK,                                     /* xrm_class */
    TRUE,                                          /* compress_motion */
    TRUE,                                          /* compress_exposure */
    TRUE,                                          /* compress_enterleave */
    TRUE,                                          /* visible_interest */
    NULL,                                          /* destroy */
    XtInheritResize,                               /* resize */
    NULL,                                          /* expose */
    NULL,                                          /* set_values */
    NULL,                                          /* set_values_hook */
    XtInheritSetValuesAlmost,                      /* set_values_almost */
    NULL,                                          /* get_values_hook */
    NULL,                                          /* accept_focus */
    XtVersion,                                     /* version */
    NULL,                                          /* callback_private */
    NULL,                                          /* tm_translations */
    NULL,
    NULL,
    NULL,
  },
  {     /* CompositeClassPart */
      /* geometry_manager */	XtInheritGeometryManager,
      /* change_managed	  */	XtInheritChangeManaged,
      /* insert_child	  */	XtInheritInsertChild,
      /* delete_child	  */	XtInheritDeleteChild,
      /* extension	  */	NULL
  },
  { /* Shell */
      /* extension	  */	NULL
  },
  { /* WMShell */
      /* extension	  */	NULL
  },
  { /* VendorShell */
      /* extension	  */	NULL
  },
  { /* TopLevelShell */
      /* extension	  */	NULL
  },
  {  /* SlaveImageDisplayPopupClassPart */
      0                   /* empty */
  },
  {  /* RenderDisplayPopupClassPart */
      0                   /* empty */
  },
};

WidgetClass renderDisplayPopupWidgetClass = (WidgetClass) &renderDisplayPopupClassRec;


static void RenderDisplayPopup__Initialise (Widget Request, Widget New)
{
    KVolumeRenderContext vrend_ctx;
    iarray cube;
    unsigned int projection;
    double esep;
    /*RenderDisplayPopupWidget request = (RenderDisplayPopupWidget) Request;*/
    RenderDisplayPopupWidget new = (RenderDisplayPopupWidget) New;
    extern KVolumeRenderContext main_vrend_ctx;
    extern KCallbackList new_byte_cube_event_list;
    extern KCallbackList cursor3d_event_list;
    extern KCallbackList rotation_event_list;
    extern KCallbackList shader_event_list;
    extern KCallbackList projection_event_list;
    extern KCallbackList eye_separation_event_list;
    extern CONST char *current_shader_name;
    extern rotation_specification rotation;
    /*static char function_name[] = "RenderDisplayPopupWidget::Initialise";*/

    new->renderDisplayPopup.left_buffer = NULL;
    new->renderDisplayPopup.right_buffer = NULL;
    new->renderDisplayPopup.buffer_bytes = 0;
    new->renderDisplayPopup.left_vimage = NULL;
    new->renderDisplayPopup.right_vimage = NULL;
    new->renderDisplayPopup.last_was_stereo = FALSE;
    if (new->renderDisplayPopup.vrend_ctx == NULL)
    {
	new->renderDisplayPopup.vrend_ctx =
	    vrender_create_context (new, VRENDER_CONTEXT_ATT_END);
    }
    vrend_ctx = new->renderDisplayPopup.vrend_ctx;
    XtAddCallback (New, XkwNrealisedCallback, realised_cbk, New);
    XtAddCallback (New, XtNpopupCallback, popup_cbk, New);
    vrender_image_desc_notify_func (vrend_ctx, image_desc_notify_func, New);
    /*  Initialise rotation  */
    generic_rotation_callback (vrend_ctx, NULL, &rotation, NULL);
    /*  Initialise other things using main context  */
    vrender_get_context_attributes (main_vrend_ctx,
				    VRENDER_CONTEXT_ATT_CUBE, &cube,
				    VRENDER_CONTEXT_ATT_PROJECTION,&projection,
				    VRENDER_CONTEXT_ATT_EYE_SEPARATION, &esep,
				    VRENDER_CONTEXT_ATT_END);
    vrender_set_context_attributes (vrend_ctx,
				    VRENDER_CONTEXT_ATT_CUBE, cube,
				    VRENDER_CONTEXT_ATT_SHADER,
				    current_shader_name,
				    VRENDER_CONTEXT_ATT_PROJECTION, projection,
				    VRENDER_CONTEXT_ATT_EYE_SEPARATION, esep,
				    VRENDER_CONTEXT_ATT_END);
    /*  Events to capture to store data  */
    c_register_callback (&new_byte_cube_event_list, generic_new_cube_callback,
			 vrend_ctx, NULL, FALSE, NULL, FALSE, FALSE);
    c_register_callback (&shader_event_list, generic_shader_callback,
			 vrend_ctx, NULL, FALSE, NULL, FALSE, FALSE);
    c_register_callback (&projection_event_list, generic_projection_callback,
			 vrend_ctx, NULL, FALSE, NULL, FALSE, FALSE);
    c_register_callback (&eye_separation_event_list,
			 generic_eye_separation_callback,
			 vrend_ctx, NULL, FALSE, NULL, FALSE, FALSE);
    /*  Events to capture to render  */
    c_register_callback (&cursor3d_event_list,
			 XkwRenderDisplayPopupChangeCallback,
			 new, NULL, FALSE, NULL, FALSE, FALSE);
    c_register_callback (&rotation_event_list,
			 XkwRenderDisplayPopupChangeCallback,
			 new, NULL, FALSE, NULL, FALSE, FALSE);
    c_register_callback (&shader_event_list,
			 XkwRenderDisplayPopupChangeCallback,
			 new, NULL, FALSE, NULL, FALSE, FALSE);
    c_register_callback (&projection_event_list,
			 XkwRenderDisplayPopupChangeCallback,
			 new, NULL, FALSE, NULL, FALSE, FALSE);
}   /*  End Function Initialise  */

static void realised_cbk (Widget parent, XtPointer client_data,
			  XtPointer call_data)
{
#if 0
    RenderDisplayPopupWidget top = (RenderDisplayPopupWidget) client_data;
    /*Widget TOP = (Widget) client_data;*/
    KWorldCanvas *wc_ptr;
    SlaveImageDisplayPopupPart *slave =
	(SlaveImageDisplayPopupPart *) &top->slaveImageDisplayPopup;
#endif
}   /*  End Function realised_cbk  */

static void popup_cbk (Widget w, XtPointer client_data, XtPointer call_data)
{
    
    XtAppAddWorkProc (XtDisplayToApplicationContext ( XtDisplay (w) ),
		      work_proc, client_data);
}   /*  End Function popup_cbk   */

static Boolean work_proc (XtPointer client_data)
/*  [SUMMARY] Render.
    [RETURNS] True if the work procedure is finished, else False to indicate
    that it should be called again.
*/
{
    RenderDisplayPopupWidget top = (RenderDisplayPopupWidget) client_data;
    extern flag stereo;

    RenderDisplayPopup__render (top, stereo);
    return (True);
}   /*  End Function work_proc  */

static void image_desc_notify_func (KVolumeRenderContext context, void **info)
/*  [SUMMARY] Image descriptor changed callback.
    [PURPOSE] This routine is called when the output image descriptor is
    changed.
    <context> The volume render context.
    <info> A pointer to the arbitrary information pointer.
    [RETURNS] Nothing.
*/
{
    RenderDisplayPopupWidget top = (RenderDisplayPopupWidget) *info;

    /*  This is called multiple times when a new cube is received, so this must
	do nothing more than invalidate old data  */
    if (top->renderDisplayPopup.left_vimage != NULL)
    {
	viewimg_destroy (top->renderDisplayPopup.left_vimage);
	top->renderDisplayPopup.left_vimage = NULL;
    }
    if (top->renderDisplayPopup.right_vimage != NULL)
    {
	viewimg_destroy (top->renderDisplayPopup.right_vimage);
	top->renderDisplayPopup.right_vimage = NULL;
    }
}   /*  End Function image_desc_notify_func  */

static void RenderDisplayPopup__render (RenderDisplayPopupWidget top,
					flag stereo)
/*  [SUMMARY] Render.
    <top> The RenderDisplayPopupWidget.
    [RETURNS] Nothing.
*/
{
    flag visible;
    uaddr image_bytes;
    KVolumeRenderContext vrend_ctx = top->renderDisplayPopup.vrend_ctx;
    KWorldCanvas mono_pc, left_pc, right_pc;
    KWorldCanvas mono_tc, left_tc, right_tc;
    KWorldCanvas mono_dc, left_dc, right_dc;
    double min, max;
    char *right_buffer;
    array_desc *arr_desc;
    flag (*func) (KVolumeRenderContext vrend_ctx);
    static char function_name[] = "RenderDisplayPopupWidget::render";

    kwin_get_attributes (canvas_get_pixcanvas
			 (top->slaveImageDisplayPopup.visibleCanvas),
			 KWIN_ATT_VISIBLE, &visible,
			 KWIN_ATT_END);
    if (!visible) return;
    func = ( flag (*) (KVolumeRenderContext vrend_ctx) )
	top->renderDisplayPopup.preRenderFunction;
    if (func != NULL) if ( !(*func) (vrend_ctx) ) return;
    vrender_get_context_attributes (vrend_ctx,
				    VRENDER_CONTEXT_ATT_IMAGE_DESC, &arr_desc,
				    VRENDER_CONTEXT_ATT_END);
    image_bytes = ds_get_array_size (arr_desc) *
	ds_get_packet_size (arr_desc->packet);
    if (top->renderDisplayPopup.buffer_bytes < image_bytes)
    {
	if (top->renderDisplayPopup.left_buffer != NULL)
	    m_free (top->renderDisplayPopup.left_buffer);
	if (top->renderDisplayPopup.right_buffer != NULL)
	    m_free (top->renderDisplayPopup.right_buffer);
	if ( ( top->renderDisplayPopup.left_buffer = m_alloc (image_bytes) )
	     == NULL ) m_abort (function_name, "left buffer");
	if ( ( top->renderDisplayPopup.right_buffer = m_alloc (image_bytes) )
	     == NULL ) m_abort (function_name, "right buffer");
	top->renderDisplayPopup.buffer_bytes = image_bytes;
    }
    right_buffer = stereo ? top->renderDisplayPopup.right_buffer : NULL;
    if ( !vrender_to_buffer (vrend_ctx, top->renderDisplayPopup.left_buffer,
			     right_buffer, &min, &max, NULL, NULL) )
    {
	fprintf (stderr, "%s: error rendering\n", function_name);
	return;
    }
    if ( (top->renderDisplayPopup.left_vimage != NULL) &&
	 (top->renderDisplayPopup.last_was_stereo == stereo) )
    {
	viewimg_register_data_change (top->renderDisplayPopup.left_vimage);
	if (top->renderDisplayPopup.right_vimage != NULL)
	    viewimg_register_data_change(top->renderDisplayPopup.right_vimage);
	top->renderDisplayPopup.last_was_stereo = stereo;
	return;
    }
    top->renderDisplayPopup.last_was_stereo = stereo;
    if (top->renderDisplayPopup.left_vimage != NULL)
    {
	viewimg_destroy (top->renderDisplayPopup.left_vimage);
	top->renderDisplayPopup.left_vimage = NULL;
    }
    if (top->renderDisplayPopup.right_vimage != NULL)
    {
	viewimg_destroy (top->renderDisplayPopup.right_vimage);
	top->renderDisplayPopup.right_vimage = NULL;
    }
    XtVaGetValues ( (Widget) top,
		    XkwNpseudoColourCanvas, &mono_pc,
		    XkwNpseudoColourLeftCanvas, &left_pc,
		    XkwNpseudoColourRightCanvas, &right_pc,
		    XkwNtrueColourCanvas, &mono_tc,
		    XkwNtrueColourLeftCanvas, &left_tc,
		    XkwNtrueColourRightCanvas, &right_tc,
		    XkwNdirectColourCanvas, &mono_dc,
		    XkwNdirectColourLeftCanvas, &left_dc,
		    XkwNdirectColourRightCanvas, &right_dc,
		    NULL );
    if (mono_pc == NULL)
    {
	mono_pc = mono_tc;
	left_pc = left_tc;
	right_pc = right_tc;
    }
    if (mono_tc == NULL)
    {
	mono_tc = mono_dc;
	left_tc = left_dc;
	right_tc = right_dc;
    }
    if (mono_tc == NULL)
    {
	mono_tc = mono_pc;
	left_tc = left_pc;
	right_tc = right_pc;
    }
    if (arr_desc->packet->num_elements == 1)
    {
	/*  PseudoColour  */
	KWorldCanvas parent_wc;

	top->renderDisplayPopup.left_vimage =
	    viewimg_create (stereo ? left_pc : mono_pc, NULL, arr_desc,
			    top->renderDisplayPopup.left_buffer, 1, 0, 0);
	if (top->renderDisplayPopup.left_vimage == NULL)
	{
	    fprintf (stderr, "%s: error creating left/cyclops vimage\n",
		     function_name);
	    return;
	}
	XtVaGetValues (XtParent (top), XkwNvisibleCanvas, &parent_wc, NULL);
	canvas_get_attributes (parent_wc,
			       CANVAS_ATT_VALUE_MIN, &min,
			       CANVAS_ATT_VALUE_MAX, &max,
			       CANVAS_ATT_END);
	viewimg_set_attributes (top->renderDisplayPopup.left_vimage,
				VIEWIMG_VATT_VALUE_MIN, min,
				VIEWIMG_VATT_VALUE_MAX, max,
				VIEWIMG_VATT_END);
	if (stereo)
	{
	    top->renderDisplayPopup.right_vimage =
	    viewimg_create (right_pc, NULL, arr_desc,
			    top->renderDisplayPopup.right_buffer, 1, 0, 0);
	    if (top->renderDisplayPopup.right_vimage == NULL)
	    {
		fprintf (stderr, "%s: error creating right vimage\n",
			 function_name);
		return;
	    }
	    viewimg_set_attributes (top->renderDisplayPopup.right_vimage,
				    VIEWIMG_VATT_VALUE_MIN, min,
				    VIEWIMG_VATT_VALUE_MAX, max,
				    VIEWIMG_VATT_END);
	    viewimg_set_active (top->renderDisplayPopup.right_vimage, FALSE);
	}
	viewimg_make_active (top->renderDisplayPopup.left_vimage);
	return;
    }
    /*  TrueColour  */
    top->renderDisplayPopup.left_vimage =
	viewimg_create_rgb (stereo ? left_tc : mono_tc, NULL, arr_desc,
			    top->renderDisplayPopup.left_buffer, 1, 0, 0, 1, 2,
			    0, NULL, NULL);
    if (top->renderDisplayPopup.left_vimage == NULL)
    {
	fprintf (stderr, "%s: error creating left/cyclops vimage\n",
		 function_name);
	return;
    }
    if (stereo)
    {
	top->renderDisplayPopup.right_vimage =
	    viewimg_create_rgb (right_tc, NULL, arr_desc,
				top->renderDisplayPopup.right_buffer, 1, 0,
				0, 1, 2, 0, NULL, NULL);
	if (top->renderDisplayPopup.right_vimage == NULL)
	{
	    fprintf (stderr, "%s: error creating right vimage\n",
		     function_name);
	    return;
	}
	viewimg_set_active (top->renderDisplayPopup.right_vimage, FALSE);
    }
    viewimg_make_active (top->renderDisplayPopup.left_vimage);
}   /*  End Function render  */


/*  Public functions follow  */

flag XkwRenderDisplayPopupChangeCallback (void *object, void *client1_data,
					  void *call_data, void *client2_data)
/*  [SUMMARY] Callback for changing a render popup display.
    [PURPOSE] This routine is called when object callbacks are called.
    <object> The object information pointer.
    <client1_data> The first client information pointer.
    <call_data> The call information pointer.
    <client2_data> The second client information pointer.
    [RETURNS] TRUE if further callbacks should not be called, else FALSE.
*/
{
    RenderDisplayPopupWidget top = (RenderDisplayPopupWidget) object;
    extern flag stereo;

    RenderDisplayPopup__render (top, stereo);
    return (FALSE);
}   /*  End Function XkwRenderDisplayPopupChangeCallback  */
