#include <math.h>
#include <karma.h>
#include <karma_a.h>
#include "xraycast_3d.h"
/*
    extern uaddr offset_pixel1;
    extern uaddr offset_pixel2;
    extern uaddr offset_pixel3;
    extern uaddr offset_depth;
    extern uaddr offset_cumulative_opacity;
    extern uaddr offset_opacity_rolloff;
    extern uaddr offset_depth_offset;
    extern uaddr offset_depth_factor;
*/

SHADER DEPTH_max_voxel
    signed short depth = -1;
    int pixel = -129;
    float alpha, beta, tmp;
    double d_pixel;
    extern int voxel_sum_position;
    extern uaddr offset_pixel1;
    extern uaddr offset_depth;
    extern uaddr offset_depth_offset;
    extern uaddr offset_depth_factor;
CODE
	if (voxel > pixel) pixel = voxel, depth = plane_pos;
CLEANUP
    if (pixel < -127) pixel = voxel_sum_position;
    else
    {
	d_pixel = (double) pixel;
	if (d_pixel < *minimum_image_value) *minimum_image_value = d_pixel;
	if (d_pixel > *maximum_image_value) *maximum_image_value = d_pixel;
    }
    *(signed char *) (pixel_ptr + offset_pixel1) = pixel;
    *(float *) (pixel_ptr + offset_depth) = depth;
    /*  Compute values which will allow a fast translation from distance of
	cut plane to raster plane and depth of cut plane down ray.  */
    /*  First shift ray start to first voxel  */
    _sh_ray_start_h += _sh_ray_direction_h * t_enter;
    _sh_ray_start_v += _sh_ray_direction_v * t_enter;
    _sh_ray_start_d += _sh_ray_direction_d * t_enter;
    tmp = (normal.h * _sh_ray_direction_h + normal.v * _sh_ray_direction_v +
	   normal.d * _sh_ray_direction_d);
    alpha = (normal.h * (vpc.h - _sh_ray_start_h) +
	     normal.v * (vpc.v - _sh_ray_start_v) +
	     normal.d * (vpc.d - _sh_ray_start_d) ) / tmp;
    beta = (normal.h * normal.h + normal.v * normal.v + normal.d * normal.d) /
	   tmp;
    /*  Scale these two to plane co-ordinates rather than units of
	ray_direction  */
/*
(void) fprintf (stderr, "%e  %e\n", alpha * _sh_ray_direction_d,
		beta * _sh_ray_direction_d);
*/
    *(float *) (pixel_ptr + offset_depth_offset) = alpha * _sh_ray_direction_d;
    *(float *) (pixel_ptr + offset_depth_factor) = beta * _sh_ray_direction_d;


SHADER DEPTH_sum_voxels
    int pixel = 0;
    double d_pixel;
    extern int voxel_sum_position;
CODE
   	pixel += voxel - voxel_sum_position;
CLEANUP
    d_pixel = (double) pixel;
    if (d_pixel < *minimum_image_value) *minimum_image_value = d_pixel;
    if (d_pixel > *maximum_image_value) *maximum_image_value = d_pixel;
    *(float *) pixel_ptr = pixel;


SHADER DEPTH_min_voxel
    int pixel = 128;
    double d_pixel;
    extern int voxel_sum_position;
CODE
	if (voxel < pixel) pixel = voxel;
CLEANUP
    if (pixel > 127) pixel = voxel_sum_position;
    else
    {
	d_pixel = (double) pixel;
	if (d_pixel < *minimum_image_value) *minimum_image_value = d_pixel;
	if (d_pixel > *maximum_image_value) *maximum_image_value = d_pixel;
    }
    *(signed char *) pixel_ptr = pixel;


SHADER DEPTH_front_voxel
    int pixel = -128;
    double d_pixel;
    extern int voxel_threshold;
CODE
	if (voxel >= voxel_threshold)
	{
	    pixel = voxel;
	    RAYCAST_TERMINATE;
	}
CLEANUP
    if (pixel < -127) pixel = -128;
    else
    {
	d_pixel = (double) pixel;
	if (d_pixel < *minimum_image_value) *minimum_image_value = d_pixel;
	if (d_pixel > *maximum_image_value) *maximum_image_value = d_pixel;
    }
    *(signed char *) pixel_ptr = pixel;


SHADER DEPTH_hot_gas_subs
    int index;
    float transparency;
    float emissivity;
    float visible_threshold = 0.02;
    float one = 1.0;
    float old_opacity = 0.0;
    float old_red = 0.0;
    float old_green = 0.0;
    float old_blue = 0.0;
    int old_index = -1;
    extern int num_substances;
    float red, green, blue, cumulative_transparency;
    extern unsigned int bad_substance_count;
    extern char voxel_to_substance_indices[256];
    extern substance substances[MAX_SUBSTANCES];

    red = 0.0;
    green = 0.0;
    blue = 0.0;
    cumulative_transparency = 1.0;
CODE
	if ( (index = voxel_to_substance_indices[voxel + 128])
	    >= num_substances )
	{
	    /*  Outside substance table: skip  */
	    ++bad_substance_count;
	    continue;
	}
	if (old_index != index)
	{
	    old_opacity = substances[index].opacity;
	    old_red = substances[index].red;
	    old_green = substances[index].green;
	    old_blue = substances[index].blue;
	    old_index = index;
	}
	transparency = one - old_opacity;
	emissivity = old_opacity * cumulative_transparency;
	red += old_red * emissivity;
	green += old_green * emissivity;
	blue += old_blue * emissivity;
	cumulative_transparency *= transparency;
	if (cumulative_transparency < visible_threshold)
	{
	    RAYCAST_TERMINATE;
	}
CLEANUP
    emissivity = 255.0;
    *(unsigned char *) pixel_ptr = (int) (red * emissivity);
    *( (unsigned char *) pixel_ptr + 1 ) = (int) (green * emissivity);
    *( (unsigned char *) pixel_ptr + 2 ) = (int) (blue * emissivity);


SHADER DEPTH_hot_gas_fullcolour
    int index;
    float transparency;
    float emissivity;
    float visible_threshold = 0.02;
    float one = 1.0;
    float opacity;
    float red, green, blue, cumulative_transparency;
    extern float fullcolour_hg_reds[FULLCOLOUR_HOT_GAS_CMAP_SIZE];
    extern float fullcolour_hg_greens[FULLCOLOUR_HOT_GAS_CMAP_SIZE];
    extern float fullcolour_hg_blues[FULLCOLOUR_HOT_GAS_CMAP_SIZE];
    extern float fullcolour_hg_opacities[FULLCOLOUR_HOT_GAS_CMAP_SIZE];

   red = 0.0;
   green = 0.0;
   blue = 0.0;
   cumulative_transparency = 1.0;
CODE
	index = voxel + 127;
	opacity = fullcolour_hg_opacities[index];
	transparency = one - opacity;
	emissivity = opacity * cumulative_transparency;
	red += fullcolour_hg_reds[index] * emissivity;
	green += fullcolour_hg_greens[index] * emissivity;
	blue += fullcolour_hg_blues[index] * emissivity;
	cumulative_transparency *= transparency;
	if (cumulative_transparency < visible_threshold)
	{
	    RAYCAST_TERMINATE;
	}
CLEANUP
    emissivity = 255.0;
    *(unsigned char *) pixel_ptr = (int) (red * emissivity);
    *( (unsigned char *) pixel_ptr + 1 ) = (int) (green * emissivity);
    *( (unsigned char *) pixel_ptr + 2 ) = (int) (blue * emissivity);


SHADER DEPTH_hot_gas_mono
    int index;
    float transparency;
    float emissivity;
    float visible_threshold = 0.02;
    float one = 1.0;
    float opacity;
    float intensity, cumulative_transparency;
    double d_pixel;
    extern float mono_hg_intensities[MONO_HOT_GAS_CMAP_SIZE];
    extern float mono_hg_opacities[MONO_HOT_GAS_CMAP_SIZE];

    intensity = 0.0;
    cumulative_transparency = 1.0;
CODE
	index = voxel + 127;
	opacity = mono_hg_opacities[index];
	transparency = one - opacity;
	emissivity = opacity * cumulative_transparency;
	intensity += mono_hg_intensities[index] * emissivity;
	cumulative_transparency *= transparency;
	if (cumulative_transparency < visible_threshold)
	{
	    RAYCAST_TERMINATE;
	}
CLEANUP
    d_pixel = (double) intensity;
    if (d_pixel < *minimum_image_value) *minimum_image_value = d_pixel;
    if (d_pixel > *maximum_image_value) *maximum_image_value = d_pixel;
    *(float *) pixel_ptr = intensity;
