/*  main_xt.c

    Source file for  xray  (Karma volume rendering tool).

    Copyright (C) 1993-2004  Richard Gooch

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Richard Gooch may be reached by email at  karma-request@atnf.csiro.au
    The postal address is:
      Richard Gooch, c/o ATNF, P. O. Box 76, Epping, N.S.W., 2121, Australia.
*/

/*  This file contains all the Xt interface code.


    Written by      Richard Gooch   30-JUL-1994

    Updated by      Richard Gooch   2-AUG-1994

    Updated by      Richard Gooch   18-SEP-1994: Added "Stereo" toggle.

    Updated by      Richard Gooch   21-SEP-1994: Stopped rendering every time
  distance changes.

    Updated by      Richard Gooch   25-OCT-1994: Moved some #include's around.

    Updated by      Richard Gooch   31-OCT-1994: Changed from  sun4sol2  to
  sparc_Solaris

    Updated by      Richard Gooch   3-NOV-1994: Switched to OS_ and MACHINE_
  macros for machine determination.

    Updated by      Richard Gooch   12-NOV-1994: Added  #ifdef HAVE_SPACEBALL

    Updated by      Richard Gooch   14-NOV-1994: Changed some resource names.

    Updated by      Richard Gooch   19-NOV-1994: Added mono hot gas support.

    Updated by      Richard Gooch   1-DEC-1994: Moved over to  chx_  package.

    Updated by      Richard Gooch   10-DEC-1994: Made use of new ExclusiveMenu
  widget.

    Updated by      Richard Gooch   14-DEC-1994: Fixed background colour
  resource for ExclusiveMenu widget.

    Updated by      Richard Gooch   4-JAN-1995: Use  XtNameToWidget  to get
  child widgets.

    Updated by      Richard Gooch   8-JAN-1995: Took account of change to
  <panel_> package in <raycast_3d>.

    Updated by      Richard Gooch   12-JAN-1995: Create display window last so
  that it can't chew up all available colours (a problem if something else is
  also using lots of colourcells).

    Updated by      Richard Gooch   20-JAN-1995: Use TrueColour visual rather
  than DirectColour (no need to worry about colourmap contents).

    Updated by      Richard Gooch   1-APR-1995: Register cube filename with
  ImageDisplay widget.

    Updated by      Richard Gooch   15-APR-1995: Added call to
  <im_register_lib_version>.

    Updated by      Richard Gooch   23-APR-1995: Added support for FITS files.

    Updated by      Richard Gooch   24-APR-1995: Prevent conversion of raw cube
  to float if raw cube is float.

    Updated by      Richard Gooch   9-JUL-1995: Added internal movie saving.

    Updated by      Richard Gooch   17-JUL-1995: Added byte cube save facility.

    Updated by      Richard Gooch   22-JUL-1995: Removed smooth toggle and
  added cursor menu.

    Updated by      Richard Gooch   24-JUL-1995: Added ray display window for
  2-D active cursor.

    Updated by      Richard Gooch   26-JUL-1995: Added support for RGB values
  in position event callback.

    Updated by      Richard Gooch   31-JUL-1995: Added support for Slicer
  window.

    Updated by      Richard Gooch   28-AUG-1995: Added raycast mode menu.

    Updated by      Richard Gooch   29-AUG-1995: Added eye separation slider.

    Updated by      Richard Gooch   1-SEP-1995: Changed to new colourmap
  creation routine.

    Updated by      Richard Gooch   7-SEP-1995: Register image name for images/
  movies too.

    Updated by      Richard Gooch   28-SEP-1995: Added Miriad Image support.

    Updated by      Richard Gooch   22-OCT-1995: Support stereo display.

    Updated by      Richard Gooch   10-DEC-1995: Fixed "Rescale" button
  callback: now actually pops up the Dataclip widget.

    Updated by      Richard Gooch   3-JAN-1996: Completed port to <vrender_>
  package.

    Updated by      Richard Gooch   18-JAN-1996: Added -no_work_funcs option.

    Updated by      Richard Gooch   8-FEB-1996: Changed distance units to cm.

    Updated by      Richard Gooch   21-FEB-1996: Added -no_slice_win and
  -no_images options and created <refresh_display_window> routine.

    Updated by      Richard Gooch   29-FEB-1996: Made background colour for
  both mono and stereo canvases black and created <move_slicer_cursor>.

    Updated by      Richard Gooch   4-MAR-1996: Added opacities to active 2D
  cursor ray display.

    Updated by      Richard Gooch   22-MAR-1996: Added -fullscreen option.

    Updated by      Richard Gooch   4-MAY-1996: Switched to KtoggleWidget.

    Updated by      Richard Gooch   5-MAY-1996: Added -test option and tied
  experimental SimpleSlider widget use and cut plane control to it.

    Updated by      Richard Gooch   6-MAY-1996: Changed so SimpleSlider widget
  is always used (no -test needed). Moved more Value widgets over to
  SimpleSlider widget. Added support for "hot_gas_three_colour" shader.

    Updated by      Richard Gooch   17-MAY-1996: Removed "Save Cube" button and
  made use of XkwNfixedOutputType resource for Dataclip widget.

    Updated by      Richard Gooch   24-MAY-1996: Set XtNresizable resource to
  TRUE for dialog widgets so that longer names may be typed and viewed.

    Updated by      Richard Gooch   26-MAY-1996: Cleaned code to keep
  gcc -Wall -pedantic-errors happy.

    Updated by      Richard Gooch   22-JUN-1996: Made use of
  <XkwFilewinStandardFileTester_nD> and <XkwFilewinStandardFileTester_3D>
  routines.

    Updated by      Richard Gooch   29-JUL-1996: Added 2D cursor "active_write"
  mode for Sung Kim.

    Updated by      Richard Gooch   31-JUL-1996: Now use filtered/unfiltered
  float cubes.

    Updated by      Richard Gooch   23-AUG-1996: Removed <move_slicer_cursor>
  and placed cursor event callback into slice_3d_xt.c

    Updated by      Richard Gooch   15-SEP-1996: Made use of new <xv_*>
  routines.

    Updated by      Richard Gooch   1-OCT-1996: Fixed resource names for new
  ExportMenu widget.

    Updated by      Richard Gooch   28-OCT-1996: Added hostname and port number
  to title.

    Updated by      Richard Gooch   9-NOV-1996: Set font for slicer widget
  track display.

    Updated by      Richard Gooch   21-NOV-1996: Moved mono Hot Gas Xt code
  to mono_hg_xt.c

    Updated by      Richard Gooch   22-NOV-1996: Moved Hot Gas substances Xt
  code to subs_hg_xt.c

    Updated by      Richard Gooch   28-NOV-1996: Took account of substance
  changes.

    Updated by      Richard Gooch   1-DEC-1996: Made use of
  <xtmisc_init_app_initialise>.

    Updated by      Richard Gooch   6-DEC-1996: Added -segments option.

    Updated by      Richard Gooch   7-DEC-1996: Changed subcube sliders to be
  in pixel units, rather than percent.

    Updated by      Richard Gooch   12-DEC-1996: Removed -segments option and
  switched to "Load Segments" button and full control panel.

    Updated by      Richard Gooch   16-DEC-1996: Switched to <dmx> package.

    Updated by      Richard Gooch   17-DEC-1996: Switched to <conn_initialise>
  function.

    Updated by      Richard Gooch   22-DEC-1996: No longer change size of
  orientation window after new cube is loaded.

    Updated by      Richard Gooch   27-DEC-1996: Made use of <ds_event_*>
  system. Switched to <dmx> package. Took account of change to
  <xtmisc_init_app_initialise>.

    Updated by      Richard Gooch   30-DEC-1996: Took account of another change
  to <xtmisc_init_app_initialise>. Added -cmap_master option.

    Updated by      Richard Gooch   12-JAN-1997: Made use of new
  XkwNtrapDirectoryDatasets resource for FilewinWidgetClass.

    Updated by      Richard Gooch   3-FEB-1997: Fixed bug in
  <cubefileselect_cbk> where obsolete XkwNiarray resource was being used to
  update the Dataclip widget. Now using <XkwDataclipNewArray> function.

    Updated by      Richard Gooch   12-FEB-1997: Only use GLISH "cube" events.

    Updated by      Richard Gooch   9-MAR-1997: Took account of change to
  <xtmisc_init_app_initialise>.

    Updated by      Richard Gooch   12-APR-1997: Ensure ThreeDeeSlice widget
  knows that byte cube data is different in <region_cbk>.

    Updated by      Richard Gooch   13-APR-1997: Made use of
  <iarray_format_value> function in <offset_position_func> function.

    Updated by      Richard Gooch   19-APR-1997: Now use internal file loader
  in Filewinpopup widget.

    Updated by      Richard Gooch   23-APR-1997: Added "Show Statistics" button

    Updated by      Richard Gooch   29-MAY-1997: Added -show_timings option.

    Updated by      Richard Gooch   22-JUN-1997: Took account of change to
  <xtmisc_init_app_initialise> to double varargs list.

    Updated by      Richard Gooch   1-MAR-1998: Made use of new filewinWidget
  resource.

    Updated by      Richard Gooch   19-MAY-1998: Trap illegal values.

    Updated by      Richard Gooch   11-JUN-1998: Add voxel threshold for
  front voxel shader.

    Updated by      Richard Gooch   28-JUL-1998: Added region control.

    Updated by      Richard Gooch   12-AUG-1998: Update cursor position on all
  canvas events.

    Updated by      Richard Gooch   9-SEP-1998: Moved subcube slider variables
  to globals_xt.c

    Updated by      Richard Gooch   17-SEP-1998: Use SimpleMenu class name in
  resources.

    Updated by      Richard Gooch   8-OCT-1998: No longer freeze controls when
  network data arrives. Set subcube slider with global values in
  <cube_event_func>.

    Updated by      Richard Gooch   10-DEC-1998: Added support for TrueColour
  only displays for ThreeDeeSlice widget.

    Updated by      Richard Gooch   16-JAN-1999: Fixed <shader_cbk> to not call
  <register_shader> twice.

    Updated by      Richard Gooch   20-FEB-1999: Added "Log" button to log
  statistics to a file.

    Updated by      Richard Gooch   19-MAR-1999: Removed restrictions when
  PseudoColour visual not available.

    Updated by      Richard Gooch   6-JUL-1999: Added -volume option.

    Udated by       Richard Gooch   30-AUG-1999: Added dirname font resource.

    Uupdated by     Richard Gooch   10-SEP-1999: Made use of initialisation
  resource <XTMISC_INIT_ATT_TITLE>.

    Updated by      Richard Gooch   11-DEC-1999: Removed unnecessary externs.

    Updated by      Richard Gooch   12-JUN-2000: Moved movie control code to
  movie_xt.c

    Updated by      Richard Gooch   11-MAY-2001: Changed dirname font resource.

    Updated by      Richard Gooch   3-OCT-2001: Added translations for Filewin.

    Updated by      Richard Gooch   21-DEC-2001: Included k_version.h directly.

    Updated by      Richard Gooch   22-DEC-2002: Added #include <stdlib.h>

    Updated by      Richard Gooch   17-SEP-2003: Removed -no_work_funcs option
  since it is now handled by the <xtmisc_init_app_initialise> function.

    Updated by      Richard Gooch   4-OCT-2003: Added #include <Xkw/Dataclip.h>

    Last updated by Richard Gooch   13-MAR-2004: Fixed bug in function
  <offset_position_func>: was consuming all events.


*/
#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <errno.h>
#include <sys/time.h>
#include <X11/Xatom.h>
#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/Xaw/Label.h>
#include <X11/Xaw/Command.h>
#include <X11/Xaw/Form.h>
#include <X11/Xaw/Dialog.h>
#include <karma.h>
#define X11
#include <k_event_codes.h>
#include <k_version.h>
#include <karma_vrender.h>
#include <karma_overlay.h>
#include <karma_viewimg.h>
#include <karma_foreign.h>
#include <karma_iarray.h>
#include <karma_xtmisc.h>
#include <karma_dsxfr.h>
#include <karma_kcmap.h>
#include <karma_kwin.h>
#include <karma_conn.h>
#include <karma_wcs.h>
#include <karma_dir.h>
#include <karma_xv.h>
#include <karma_im.h>
#include <karma_hi.h>
#include <karma_xc.h>
#include <karma_ds.h>
#include <karma_ic.h>
#include <karma_a.h>
#include <karma_m.h>
#include <Xkw/Canvas.h>
#include <Xkw/Filepopup.h>
#include <Xkw/Cmapwin.h>
#include <Xkw/Value.h>
#include <Xkw/SimpleSlider.h>
#include <Xkw/Twodpos.h>
#include <Xkw/ExclusiveMenu.h>
#include <Xkw/ImageDisplay.h>
#include <Xkw/Dialogpopup.h>
#include <Xkw/ThreeDeeSlice.h>
#include <Xkw/Filewin.h>
#include <Xkw/Ktoggle.h>
#include <Xkw/Dataclip.h>
#define NEED_PROJECTION_ALTERNATIVES
#define NEED_MODE_ALTERNATIVES
#define NEED_CURSOR_2D_ALTERNATIVES
#include "xraycast_3d.h"


#define VERSION "1.8.16"

#define ANGLE_MIN -179
#define ANGLE_MAX 180

#define DISTANCE_MAX 500  /*  5 meters from screen (cube centre)  */


/*  File: display_xt.c  */
EXTERN_FUNCTION (Widget create_display_win,
		 (Widget main_shell, Widget stereo_tgl) );

/*  File: slice_3d_xt.c  */
EXTERN_FUNCTION (Widget create_3d_slice_win,
		 (Widget main_shell, Kcolourmap kcmap, Widget *slicer,
		  Visual *visual, Colormap xcmap) );

/*  File: cont_hg_xt.c  */
EXTERN_FUNCTION (Widget create_cont_hg_win,
		 (Widget main_shell, Visual *visual) );

/*  File: threecolour_hg_xt.c  */
EXTERN_FUNCTION (Widget create_threecolour_hg_win, (Widget main_shell) );

/*  File: filter_xt.c  */
EXTERN_FUNCTION (Widget create_filter_win, (Widget main_shell) );

/*  File: cursor_3d_xt.c  */
EXTERN_FUNCTION (Widget create_cursor_3d_menu,
		 (Widget form, Widget vert, Widget horiz) );

/*  File: mono_hg_xt.c  */
EXTERN_FUNCTION (Widget create_mono_hg_win, (Widget main_shell) );

/*  File: subs_hg_xt.c  */
EXTERN_FUNCTION (Widget create_substance_win, (Widget main_shell,
					       Visual *visual) );

/*  File: segments_xt.c  */
EXTERN_FUNCTION (void load_segments_cbk,
		 (Widget w, XtPointer client_data, XtPointer call_data) );

/*  File: region_xt.c  */
EXTERN_FUNCTION (void region_button_cbk,
		 (Widget w, XtPointer client_data, XtPointer call_data) );

/*  File: movie_xt.c  */
EXTERN_FUNCTION (void create_movie_win, (Widget main_shell) );

/*  Private functions  */
STATIC_FUNCTION (void quiescent_func, () );
STATIC_FUNCTION (void quit_cbk, (Widget w, XtPointer client_data,
				 XtPointer call_data) );
STATIC_FUNCTION (void shader_cbk, (Widget w, XtPointer client_data,
				   XtPointer call_data) );
STATIC_FUNCTION (void projection_cbk, (Widget w, XtPointer client_data,
				       XtPointer call_data) );
STATIC_FUNCTION (Widget create_orient_win, (Widget main_shell) );
STATIC_FUNCTION (void slider_view_cbk, (Widget w, XtPointer client_data,
					XtPointer call_data) );
STATIC_FUNCTION (void toggle_cbk, (Widget w, XtPointer client_data,
				   XtPointer call_data) );
STATIC_FUNCTION (Widget create_histogram_win,
		 (Widget main_shell, void (*x_func) (signed char x),
		  int *x_ptr) );
STATIC_FUNCTION (void region_cbk, (Widget w, XtPointer client_data,
				   XtPointer call_data) );
STATIC_FUNCTION (void cursor_2d_cbk, (Widget w, XtPointer client_data,
				      XtPointer call_data) );
STATIC_FUNCTION (void create_ray_win, (Widget main_shell) );
STATIC_FUNCTION (void mode_cbk, (Widget w, XtPointer client_data,
				 XtPointer call_data) );
STATIC_FUNCTION (void eye_sep_cbk, (Widget w, XtPointer client_data,
				    XtPointer call_data) );
STATIC_FUNCTION (void hide_wireframe_cbk, (Widget w, XtPointer client_data,
					   XtPointer call_data) );
STATIC_FUNCTION (void depth_cbk, (Widget w, XtPointer client_data,
				  XtPointer call_data) );
STATIC_FUNCTION (flag cube_event_func,
		 (void *object, multi_array *multi_desc,
		  CONST char *domain, CONST char *name) );
STATIC_FUNCTION (void show_statistics_cbk,
		 (Widget w, XtPointer client_data, XtPointer call_data) );
STATIC_FUNCTION (void register_voxel_threshold, (signed char threshold) );


/*  Private data  */
String fallback_resources[] =
{
    "Xray.geometry:                                 +0+0",
    "Xray*Canvas.background:                        black",
    "Xray*Canvas.foreground:                        white",
    "Xray*Command*background:                       grey70",
    "Xray*Repeater*background:                      grey70",
    "Xray*Ktoggle*background:                       grey80",
    "Xray*closeButton*background:                   grey90",
    "Xray*ChoiceMenu.background:                    turquoise",
    "Xray*ExclusiveMenu.background:                 turquoise",
    "Xray*Value*background:                         #d0a0a0",
    "Xray*SimpleSlider.foreground:                  Sea Green",
    "Xray*ScrollableList*foreground:                black",
    "Xray*ScrollableList.highlightPixel:            Sea Green",
    "Xray*topForm*quit*background:                  orange",
    "Xray*displaywinpopup.geometry:                 -0+0",
    "Xray*ImageDisplay.trackLabel0.font:            8x13",
    "Xray*ImageDisplay.trackLabel1.font:            8x13",
    "Xray*ImageDisplay.trackLabel2.font:            8x13",
    "Xray*ImageDisplay*trackOutput*font:            8x13bold",
    "Xray*SimpleMenu*font:                          10x20",
    "Xray*ImageDisplay.Command.font:                7x13bold",
    "Xray*ImageDisplay.Ktoggle.font:                7x13bold",
    "Xray*ImageDisplay.ChoiceMenu.font:             7x13bold",
    "Xray*ImageDisplay.exportMenu.menuButton.font:  7x13bold",
    "Xray*ImageDisplay.overlayMenu.menuButton.font: 7x13bold",
    "Xray*ImageDisplay*zoomMenu.Unzoom.foreground:  red",
    "Xray*slicewinpopup.geometry:                   +0-0",
    "Xray*slicewinpopup*trackLabel0*font:           8x13bold",
    "Xray*slicewinpopup*trackLabel1*font:           8x13bold",
    "Xray*SimpleSlider.borderWidth:                 0",
    "Xray*sectionLabel.foreground:                  red",
    "Xray*IntensityPolicy*autoIntensityScale:       False",
    "Xray*Filewin.dirname.textSink.font:            6x13",
    "Xray*font:                                     9x15bold",
    "Xray*imageExpandSlider*box.Repeater.initialDelay: 1000",
    "Xray*borderColor:                              black",
    "Xray*background:                               aquamarine",
    "Xray*foreground:                               black",
    NULL
};
static XrmOptionDescRec Options[] =
{
    {"-private_cmap",
     ".displaywinpopup.form.multiCanvas.pseudoColourCanvas.forceNewCmap",
     XrmoptionNoArg, (XtPointer) "True"},
    {"-num_colours", ".displaywinpopup.form.cmapSize", XrmoptionSepArg,
     (XtPointer) NULL},
    {"-cmap_master", ".displaywinpopup.form.cmapMaster", XrmoptionSepArg,
     (XPointer) NULL},
    {"-fullscreen", ".displaywinpopup.form.fullscreen", XrmoptionNoArg,
     (XtPointer) "True"},
    {"-verbose", "*verbose", XrmoptionNoArg, (XtPointer) "True"},
};

static XtAppContext app_context = NULL;
static Widget main_shell = NULL;
static Widget orient_winpopup = NULL;
static Widget roll_sld = NULL;
static Widget pitch_sld = NULL;
static Widget yaw_sld = NULL;
static Widget substance_winpopup = NULL;
static Widget offset_winpopup = NULL;
static Widget threshold_winpopup = NULL;
static Widget cont_hg_winpopup = NULL;
static Widget threecolour_hg_winpopup = NULL;
static Widget mono_hg_winpopup = NULL;
static Widget display_winpopup = NULL;
static Widget rescale_btn = NULL;
static Widget filter_btn = NULL;
static Widget ray_winpopup = NULL;
static Widget slicer_popup = NULL;
static Widget slicer_3d = NULL;
static Widget filter_winpopup = NULL;


/*  Main window info  */

/*  Other global variables  */

/*  These values are in degrees  */
static int roll_value = 0;
static int pitch_value = 0;
static int yaw_value = 0;
/*  This is in centimeters  */
static int distance_value = DEFAULT_DISTANCE;

/*  Shader list  */
static unsigned int num_shaders = 0;
static CONST char **shader_names;


/*  Public routines follow  */

int main (int argc, char **argv)
{
    flag no_slice_win = FALSE;
    flag test_mode = FALSE;
    long count;
    Widget topform;
    Widget filepopup, files_btn, filewin;
    Widget show_orient_btn, make_movie_btn;
    Widget quit_btn;
    Widget hide_wireframe_tgl;
    Widget w_bottom, w_left;
    Widget compute_btn, reset_orient_btn, segments_btn, region_btn;
    Widget distance_sld;
    Widget expand_sld, eye_sep_sld, depth_sld;
    Widget stereo_tgl;
    Widget cursor_2d_menu_btn, shader_menu_btn, proj_menu_btn, mode_menu_btn;
    Display *dpy;
    XVisualInfo *vinfo;
    Visual *pseudocolour_visual, *truecolour_visual, *directcolour_visual;
    Visual *any_visual, *root_visual;
    Screen *screen;
    char title_name[STRING_LENGTH];
    extern flag using_spaceball;
    extern flag hide_wireframe, stereo;
    extern flag verbose, inhibit_images, show_timings;
    extern KWorldCanvas cyclops_image_pc_worldcanvas;
    extern KWorldCanvas cyclops_image_rgb_worldcanvas;
    extern int roll_value, pitch_value, yaw_value;
    extern int image_expand_factor;
    extern int cut_plane_depth;
    extern int voxel_sum_position, voxel_threshold;
    extern Widget start_x_sld, end_x_sld, start_y_sld, end_y_sld;
    extern Widget start_z_sld, end_z_sld;
    extern Widget dataclip_winpopup, slicer_popup, slicer_3d, movie_winpopup;
    extern CONST char *current_shader_name;
    extern substance substances[MAX_SUBSTANCES];
    extern substance temp_substances[MAX_SUBSTANCES];
    extern i_coord_3d subcube_start;
    extern i_coord_3d subcube_end;
    static char function_name[] = "main";

#if defined(SIGFPE_ABORT) && !defined(OS_Solaris)
	ieee_handler ("set", "division", SIGFPE_ABORT);
#endif
    /*  First search for special options  */
    for (count = 1; count < argc; ++count)
    {
	if (strcmp (argv[count], "-show_timings") == 0)
	{
	    show_timings = TRUE;
	}
	else if (strcmp (argv[count], "-no_slice_win") == 0)
	{
	    no_slice_win = TRUE;
	}
	else if (strcmp (argv[count], "-no_images") == 0)
	{
	    inhibit_images = TRUE;
	}
	else if (strcmp (argv[count], "-test") == 0)
	{
	    test_mode = TRUE;
	}
    }
    /*  Initialise module  */
    im_register_module_name ("xray");
    im_register_module_version_date (VERSION);
    im_register_lib_version (KARMA_VERSION);
    /*  Start up Xt  */
    main_shell = xtmisc_init_app_initialise (&app_context, "Xray",
					     Options, XtNumber (Options),
					     &argc, argv, fallback_resources,
					     XTMISC_INIT_ATT_MIN_CCELLS, 100,
					     XTMISC_INIT_ATT_COMMS_SETUP, TRUE,
					     XTMISC_INIT_ATT_BECOME_SERVER,
					     TRUE,
					     XTMISC_INIT_ATT_TITLE, title_name,
					     XTMISC_INIT_ATT_VERBOSE, &verbose,
					     XTMISC_INIT_ATT_END,
					     NULL);
    if (verbose) fprintf (stderr, "Verbose output enabled\n");
    if (show_timings) fprintf (stderr, "Timing display output enabled\n");
    if (no_slice_win) fprintf (stderr, "3D slicer window disabled\n");
    if (inhibit_images) fprintf (stderr, "Rendered images inhibited\n");
    if (test_mode) fprintf (stderr, "Running in test mode\n");
    /*  Initialise communications  */
    dpy = XtDisplay (main_shell);
    screen = XtScreen (main_shell);
    /*  Get visual information  */
    root_visual = XDefaultVisualOfScreen (screen);
    vinfo = xv_get_visinfo_for_visual (dpy, root_visual);
    xv_get_visuals (screen, &pseudocolour_visual,
		    &truecolour_visual, &directcolour_visual);
    if ( (vinfo->depth == 8) && (vinfo->class == PseudoColor) &&
	 (vinfo->colormap_size == 256) )
    {
	/*  Use the default visual  */
	pseudocolour_visual = root_visual;
    }
    /*  Select a "default" visual for windows which don't really care. Choose
	TrueColour by preference, since that requires no colourcells  */
    any_visual = truecolour_visual;
    if (any_visual == NULL) any_visual = pseudocolour_visual;
    if (any_visual == NULL) any_visual = directcolour_visual;
    if (any_visual == NULL)
    {
	fprintf (stderr, "No appropriate visuals found!\n");
	exit (RV_UNDEF_ERROR);
    }
    XFree (vinfo);
    vinfo = NULL;
    initialise ();
    reset_view ();
#ifdef HAVE_SPACEBALL
    using_spaceball = initialise_spaceball ();
#endif
    /*  Create other widgets  */
    topform = XtVaCreateManagedWidget ("topForm", formWidgetClass, main_shell,
				       XtNorientation, XtorientHorizontal,
				       XtNhSpace, 0,
				       XtNborderWidth, 0,
				       NULL);
    filepopup = XtVaCreatePopupShell
	("filewinPopup", filepopupWidgetClass, main_shell,
	 XkwNfilenameTester, XkwFilewinStandardFileTester_3D,
	 XtNtitle, "Xray File Selector",
	 XkwNautoPopdown, TRUE,
	 XkwNloadFiles, TRUE,
	 XkwNdomainName, "CUBE",
	 NULL);
    XtVaGetValues (filepopup, XkwNfilewinWidget, &filewin, NULL);
    XtVaSetValues (filewin,
		   XkwNforwardSyntheticEvents, True,
		   XkwNtrapDirectoryDatasets, True,
		   NULL);
    ds_event_register_func (cube_event_func, NULL);
    dataclip_winpopup = XtVaCreatePopupShell ("dataclip",
					      dataclipWidgetClass, main_shell,
					      XtNtitle,
					      "Xray Range Selector",
					      XkwNshowBlankControl, True,
					      XkwNfixedOutputType, K_BYTE,
					      NULL);
    XtAddCallback (dataclip_winpopup, XkwNrawRegionCallback, region_cbk, NULL);
    XtRealizeWidget (dataclip_winpopup);
    files_btn = XtVaCreateManagedWidget ("button", commandWidgetClass, topform,
					 XtNlabel, "Files",
					 NULL);
    XtAddCallback (files_btn, XtNcallback, xtmisc_popup_cbk, filepopup);
    show_orient_btn = XtVaCreateManagedWidget ("button",
					       commandWidgetClass, topform,
					       XtNlabel, "Show Orientation",
					       XtNfromHoriz, files_btn,
					       NULL);
    make_movie_btn = XtVaCreateManagedWidget ("button",
					      commandWidgetClass, topform,
					      XtNlabel, "Make Movie",
					      XtNfromHoriz, show_orient_btn,
					      NULL);
    rescale_btn = XtVaCreateManagedWidget ("button",commandWidgetClass,topform,
					   XtNlabel, "Rescale",
					   XtNfromHoriz, make_movie_btn,
					   NULL);
    XtSetSensitive (rescale_btn, False);
    filter_btn = XtVaCreateManagedWidget ("button",commandWidgetClass,topform,
					   XtNlabel, "Filter",
					   XtNfromHoriz, rescale_btn,
					   NULL);
    XtSetSensitive (filter_btn, False);
    quit_btn = XtVaCreateManagedWidget ("quit", commandWidgetClass, topform,
					XtNlabel, "Quit",
					XtNfromHoriz, filter_btn,
					NULL);
    XtAddCallback (quit_btn, XtNcallback, quit_cbk, NULL);
    hide_wireframe_tgl = XtVaCreateManagedWidget ("toggle",
						  ktoggleWidgetClass, topform,
						  XtNlabel, "Hide Wireframe",
						  XtNstate, False,
						  XkwNcrosses, False,
						  XtNfromVert, files_btn,
						  NULL);
    XtAddCallback (hide_wireframe_tgl, XtNcallback, hide_wireframe_cbk,
		   (XtPointer) &hide_wireframe);
    /*  Create the stereo toggle  */
    stereo_tgl = XtVaCreateManagedWidget ("toggle", ktoggleWidgetClass,topform,
					  XtNlabel, "Stereo",
					  XtNstate, False,
					  XkwNcrosses, False,
					  XtNfromVert, show_orient_btn,
					  XtNfromHoriz, hide_wireframe_tgl,
					  NULL);
    XtAddCallback (stereo_tgl, XtNcallback, toggle_cbk, (XtPointer) &stereo);
    /*  Create the 2-D cursor menu  */
    cursor_2d_menu_btn = XtVaCreateManagedWidget ("menuButton",
						  exclusiveMenuWidgetClass,
						  topform,
						  XtNmenuName, "2DcursorMenu",
						  XkwNchoiceName, "2-D CURSOR",
						  XtNfromVert, show_orient_btn,
						  XtNfromHoriz, stereo_tgl,
						  XkwNnumItems,
						  NUM_CURSOR_2D_ALTERNATIVES,
						  XkwNitemStrings,
						  cursor_2d_alternatives,
						  NULL);
    XtAddCallback (cursor_2d_menu_btn, XkwNselectCallback, cursor_2d_cbk,NULL);
    /*  Create the shader menu  */
    if ( ( shader_names = vrender_get_shaders (&num_shaders) ) == NULL )
    {
	m_abort (function_name, "array of shaders");
    }
    /*  Now add special (unlisted) depth shaders  */
    init_depth_shaders ();
    current_shader_name = shader_names[0];
    shader_menu_btn = XtVaCreateManagedWidget ("menuButton",
					       exclusiveMenuWidgetClass,
					       topform,
					       XtNmenuName, "shaderMenu",
					       XkwNchoiceName, "SHADER",
					       XtNfromVert, stereo_tgl,
					       XkwNnumItems, num_shaders,
					       XkwNitemStrings, shader_names,
					       NULL);
    XtAddCallback (shader_menu_btn, XkwNselectCallback, shader_cbk, NULL);
    /*  Create the projection menu  */
    proj_menu_btn = XtVaCreateManagedWidget ("menuButton",
					     exclusiveMenuWidgetClass, topform,
					     XtNmenuName, "projectionMenu",
					     XkwNchoiceName, "PROJECTION",
					     XtNfromVert, stereo_tgl,
					     XtNfromHoriz, shader_menu_btn,
					     XkwNnumItems, NUM_PROJECTIONS,
					     XkwNitemStrings,
					     projection_alternatives,
					     NULL);
    XtAddCallback (proj_menu_btn, XkwNselectCallback, projection_cbk, NULL);
    /*  Create the raycast mode menu  */
    mode_menu_btn = XtVaCreateManagedWidget ("menuButton",
					     exclusiveMenuWidgetClass, topform,
					     XtNmenuName, "modeMenu",
					     XkwNchoiceName, "MODE",
					     XtNfromVert, proj_menu_btn,
					     XkwNnumItems, NUM_MODES,
					     XkwNitemStrings,
					     mode_alternatives,
					     NULL);
    XtAddCallback (mode_menu_btn, XkwNselectCallback, mode_cbk, NULL);
    compute_btn = XtVaCreateManagedWidget ("button", commandWidgetClass,
					   topform,
					   XtNlabel, "Compute",
					   XtNfromVert, mode_menu_btn,
					   NULL);
    XtAddCallback (compute_btn, XtNcallback, compute, NULL);
    if (using_spaceball)
    {
	w_bottom = compute_btn;
	w_left = compute_btn;
    }
    else
    {
	reset_orient_btn =
	XtVaCreateManagedWidget ("button", commandWidgetClass, topform,
				 XtNlabel, "Reset Orientation",
				 XtNfromVert, mode_menu_btn,
				 XtNfromHoriz, compute_btn,
				 NULL);
	XtAddCallback (reset_orient_btn, XtNcallback, reset_view, NULL);
	w_left = reset_orient_btn;
	roll_sld =
	XtVaCreateManagedWidget ("slider", simpleSliderWidgetClass, topform,
				 XtNlabel, "Roll",
				 XtNfromVert, compute_btn,
				 XtNorientation, XtorientVertical,
				 XkwNminimum, ANGLE_MIN,
				 XkwNmaximum, ANGLE_MAX,
				 XkwNmodifier, 1,
				 XtNvalue, 0,
				 XkwNvaluePtr, &roll_value,
				 XkwNwrap, True,
				 NULL);
	XtAddCallback (roll_sld, XkwNvalueChangeCallback, slider_view_cbk,
		       NULL);
	pitch_sld =
	XtVaCreateManagedWidget ("slider", simpleSliderWidgetClass, topform,
				 XtNlabel, "Pitch",
				 XtNfromVert, compute_btn,
				 XtNfromHoriz, roll_sld,
				 XtNorientation, XtorientVertical,
				 XkwNminimum, ANGLE_MIN,
				 XkwNmaximum, ANGLE_MAX,
				 XkwNmodifier, 1,
				 XtNvalue, 0,
				 XkwNvaluePtr, &pitch_value,
				 XkwNwrap, True,
				 NULL);
	XtAddCallback (pitch_sld, XkwNvalueChangeCallback, slider_view_cbk,
		       NULL);
	yaw_sld =
	XtVaCreateManagedWidget ("slider", simpleSliderWidgetClass, topform,
				 XtNlabel, "Yaw",
				 XtNfromVert, compute_btn,
				 XtNfromHoriz, pitch_sld,
				 XtNorientation, XtorientVertical,
				 XkwNminimum, ANGLE_MIN,
				 XkwNmaximum, ANGLE_MAX,
				 XkwNmodifier, 1,
				 XkwNvaluePtr, &yaw_value,
				 XtNvalue, 0,
				 XkwNwrap, True,
				 NULL);
	XtAddCallback (yaw_sld, XkwNvalueChangeCallback, slider_view_cbk,NULL);
	w_bottom = yaw_sld;
    }
    segments_btn = XtVaCreateManagedWidget ("button", commandWidgetClass,
					    topform,
					    XtNlabel, "Load Segments",
					    XtNfromVert, mode_menu_btn,
					    XtNfromHoriz, w_left,
					    NULL);
    XtAddCallback (segments_btn, XtNcallback, load_segments_cbk, main_shell);
    region_btn = XtVaCreateManagedWidget ("button", commandWidgetClass,
					  topform,
					  XtNlabel, "Region",
					  XtNfromVert, mode_menu_btn,
					  XtNfromHoriz, segments_btn,
					  NULL);
    XtAddCallback (region_btn, XtNcallback, region_button_cbk, main_shell);
    if (getuid () == 465)
    {
	Widget w;

	w = XtVaCreateManagedWidget ("button", commandWidgetClass, topform,
				     XtNlabel, "Stats",
				     XtNfromVert, mode_menu_btn,
				     XtNfromHoriz, region_btn,
				     NULL);
	XtAddCallback (w, XtNcallback, show_statistics_cbk, NULL);
	w = XtVaCreateManagedWidget ("button", commandWidgetClass, topform,
				     XtNlabel, "Log",
				     XtNfromVert, mode_menu_btn,
				     XtNfromHoriz, w,
				     NULL);
	XtAddCallback (w, XtNcallback, show_statistics_cbk, "stats.log");
    }
    start_x_sld = XtVaCreateManagedWidget ("slider", simpleSliderWidgetClass,
					   topform,
					   XtNlabel, "Start X",
					   XtNfromVert, w_bottom,
					   XtNorientation, XtorientVertical,
					   XkwNminimum, 0,
					   XkwNmaximum, 0,
					   XkwNmodifier, 1,
					   XtNvalue, 0,
					   XkwNvaluePtr, &subcube_start.x,
					   NULL);
    XtAddCallback (start_x_sld, XkwNvalueChangeCallback, register_subcube_size,
		   NULL);
    start_y_sld = XtVaCreateManagedWidget ("slider", simpleSliderWidgetClass,
					   topform,
					   XtNlabel, "Start Y",
					   XtNfromVert, w_bottom,
					   XtNfromHoriz, start_x_sld,
					   XtNorientation, XtorientVertical,
					   XkwNminimum, 0,
					   XkwNmaximum, 0,
					   XkwNmodifier, 1,
					   XtNvalue, 0,
					   XkwNvaluePtr, &subcube_start.y,
					   NULL);
    XtAddCallback (start_y_sld, XkwNvalueChangeCallback, register_subcube_size,
		   NULL);
    start_z_sld = XtVaCreateManagedWidget ("slider", simpleSliderWidgetClass,
					   topform,
					   XtNlabel, "Start Z",
					   XtNfromVert, w_bottom,
					   XtNfromHoriz, start_y_sld,
					   XtNorientation, XtorientVertical,
					   XkwNminimum, 0,
					   XkwNmaximum, 0,
					   XkwNmodifier, 1,
					   XtNvalue, 0,
					   XkwNvaluePtr, &subcube_start.z,
					   NULL);
    XtAddCallback (start_z_sld, XkwNvalueChangeCallback, register_subcube_size,
		   NULL);
    end_x_sld = XtVaCreateManagedWidget ("slider", simpleSliderWidgetClass,
					 topform,
					 XtNlabel, "End X",
					 XtNfromVert, start_x_sld,
					 XtNorientation, XtorientVertical,
					 XkwNminimum, 0,
					 XkwNmaximum, 0,
					 XkwNmodifier, 1,
					 XtNvalue, 0,
					 XkwNvaluePtr, &subcube_end.x,
					 NULL);
    XtAddCallback (end_x_sld, XkwNvalueChangeCallback, register_subcube_size,
		   NULL);
    end_y_sld = XtVaCreateManagedWidget ("slider", simpleSliderWidgetClass,
					 topform,
					 XtNlabel, "End Y",
					 XtNfromVert, start_x_sld,
					 XtNfromHoriz, end_x_sld,
					 XtNorientation, XtorientVertical,
					 XkwNminimum, 0,
					 XkwNmaximum, 0,
					 XkwNmodifier, 1,
					 XkwNvaluePtr, &subcube_end.y,
					 XtNvalue, 0,
					 NULL);
    XtAddCallback (end_y_sld, XkwNvalueChangeCallback, register_subcube_size,
		   NULL);
    end_z_sld = XtVaCreateManagedWidget ("slider", simpleSliderWidgetClass,
					 topform,
					 XtNlabel, "End Z",
					 XtNfromVert, start_x_sld,
					 XtNfromHoriz, end_y_sld,
					 XtNorientation, XtorientVertical,
					 XkwNminimum, 0,
					 XkwNmaximum, 0,
					 XkwNmodifier, 1,
					 XtNvalue, 0,
					 XkwNvaluePtr, &subcube_end.z,
					 NULL);
    XtAddCallback (end_z_sld, XkwNvalueChangeCallback, register_subcube_size,
		   NULL);
    distance_sld = XtVaCreateManagedWidget ("slider", valueWidgetClass,
					    topform,
					    XtNlabel, "Distance (cm)",
					    XtNfromVert, end_z_sld,
					    XtNorientation, XtorientVertical,
					    XkwNminimum, 10,
					    XkwNmaximum, DISTANCE_MAX,
					    XkwNmodifier, 1,
					    XtNvalue, DEFAULT_DISTANCE,
					    XkwNvaluePtr, &distance_value,
					    NULL);
    XtAddCallback (distance_sld, XkwNvalueChangeCallback, slider_view_cbk,
		   NULL);
    expand_sld = XtVaCreateManagedWidget ("imageExpandSlider",
					  valueWidgetClass, topform,
					  XtNlabel, "Image Expand",
					  XtNfromVert, end_z_sld,
					  XtNfromHoriz, distance_sld,
					  XtNorientation, XtorientVertical,
					  XkwNminimum, 1,
					  XkwNmaximum, 10,
					  XkwNmodifier, 1,
					  XtNvalue, 1,
					  XkwNvaluePtr, &image_expand_factor,
					  NULL);
    XtAddCallback (expand_sld, XkwNvalueChangeCallback, compute,
		   NULL);
    eye_sep_sld = XtVaCreateManagedWidget ("eyeSeparationSlider",
					   valueWidgetClass, topform,
					   XtNlabel, "Eye Sep. (mm)",
					   XtNfromVert, end_z_sld,
					   XtNfromHoriz, expand_sld,
					   XtNorientation, XtorientVertical,
					   XkwNminimum, 10,
					   XkwNmaximum, 1000,
					   XkwNmodifier, 1,
					   XtNvalue, DEFAULT_EYE_SEP,
					   NULL);
    XtAddCallback (eye_sep_sld, XkwNvalueChangeCallback, eye_sep_cbk, NULL);
    if (test_mode)
    {
	/*  Hack for cut plane  */
	depth_sld = XtVaCreateManagedWidget ("depthSlider", valueWidgetClass,
					     topform,
					     XtNlabel, "Depth",
					     XtNfromVert, eye_sep_sld,
					     XtNorientation, XtorientVertical,
					     XkwNminimum, -400,
					     XkwNmaximum, 400,
					     XkwNmodifier, 1,
					     XtNvalue, 50,
					     XkwNvaluePtr, &cut_plane_depth,
					     NULL);
	XtAddCallback (depth_sld, XkwNvalueChangeCallback, depth_cbk, NULL);
    }
    XtRealizeWidget (main_shell);
    XtRealizeWidget (filepopup);
    orient_winpopup = create_orient_win (main_shell);
    XtAddCallback (show_orient_btn, XtNcallback, xtmisc_popup_cbk,
		   orient_winpopup);
    /*  Setup substances  */
    for (count = 0; count < MAX_SUBSTANCES; ++count)
    {
	/*  Initialise substance data structure  */
	substances[count].max_voxel = -128;
	substances[count].red = 0.0;
	substances[count].green = 0.0;
	substances[count].blue = 0.0;
	substances[count].opacity = 0.0;
	temp_substances[count].max_voxel = substances[count].max_voxel;
	temp_substances[count].red = substances[count].red;
	temp_substances[count].green = substances[count].green;
	temp_substances[count].blue = substances[count].blue;
	temp_substances[count].opacity = substances[count].opacity;
    }
    substance_winpopup = create_substance_win (main_shell, any_visual);
    offset_winpopup = create_histogram_win (main_shell, register_voxel_offset,
					    &voxel_sum_position);
    threshold_winpopup = create_histogram_win (main_shell,
					       register_voxel_threshold,
					       &voxel_threshold);
    create_movie_win (main_shell);
    cont_hg_winpopup = create_cont_hg_win (main_shell, pseudocolour_visual);
    threecolour_hg_winpopup = create_threecolour_hg_win (main_shell);
    mono_hg_winpopup = create_mono_hg_win (main_shell);
    create_ray_win (main_shell);
    filter_winpopup = create_filter_win (main_shell);
    /*  Create the display window later since it tries to grab 200 colours. If
	you are running out of colours, this could cause problems.  */
    display_winpopup = create_display_win (main_shell, stereo_tgl);
    /*  Create the 3-D cursor menu  */
    create_cursor_3d_menu (topform, proj_menu_btn, mode_menu_btn);
    /*  Create the slicer popup last since it grabs the Kcolourmap of the
	display window.  */
    if (!no_slice_win)
    {
	Kcolourmap kcmap;
	Colormap xcmap;
	Visual *visual;

	if (pseudocolour_visual == NULL)
	{
	    kcmap = canvas_get_cmap (cyclops_image_rgb_worldcanvas);
	    XtVaGetValues (main_shell, XtNcolormap, &xcmap, NULL);
	    visual = root_visual;
	}
	else
	{
	    Kdisplay dpy_handle = NULL;

	    kcmap = canvas_get_cmap (cyclops_image_pc_worldcanvas);
	    kcmap_get_attributes (kcmap,
				  KCMAP_ATT_DPY_HANDLE, &dpy_handle,
				  KCMAP_ATT_END);
	    xcmap = xc_get_cmap (dpy_handle);
	    visual = pseudocolour_visual;
	}
	slicer_popup = create_3d_slice_win (main_shell, kcmap, &slicer_3d,
					    visual, xcmap);
    }
    XtAddCallback (make_movie_btn, XtNcallback, xtmisc_popup_cbk,
		   movie_winpopup);
    XtAddCallback (rescale_btn, XtNcallback, xtmisc_popup_cbk,
		   dataclip_winpopup);
    XtAddCallback (filter_btn, XtNcallback, xtmisc_popup_cbk,
		   filter_winpopup);
    /*  If controlled by CM tool or shell: disable controls and wait  */
    if ( conn_controlled_by_cm_tool () )
    {
	fprintf (stderr, "Controlled by CM tool\n");
	conn_register_cm_quiescent_func (quiescent_func);
	freeze_controls (TRUE);
	fprintf (stderr, "Entering Xt main loop...\n");
    }
    XRaiseWindow ( dpy, XtWindow (main_shell) );
    /*  Option processing after widgets created and realised  */
    for (count = 1; count < argc; ++count)
    {
	if (strcmp (argv[count], "-volume") == 0)
	{
	    multi_array *multi_desc;
	    char fname[STRING_LENGTH];

	    if (count + 1 >= argc)
	    {
		fprintf (stderr, "File missing for volume\n");
		continue;
	    }
	    ++count;
	    if (strncmp (argv[count], "./", 2) == 0)
		strcpy (fname, argv[count] + 2);
	    else strcpy (fname, argv[count]);
	    if ( ( multi_desc =
		   foreign_guess_and_read (fname, K_CH_MAP_LOCAL,
					   FALSE, NULL,
					   FA_GUESS_READ_END) )
		 == NULL )
	    {
		fprintf (stderr, "Error reading file: \"%s\"\n", fname);
		continue;
	    }
	    ds_event_dispatch (multi_desc, "CUBE", fname);
	    ds_dealloc_multi (multi_desc);
	    m_clear (fname, STRING_LENGTH);  /*  Force me to take care  */
	}
	/*else fprintf (stderr, "Option: \"%s\" ignored\n", argv[count]);*/
    }
    XtAppMainLoop (app_context);
    return (RV_OK);
}   /*  End Function main   */

void freeze_controls (freeze)
/*  This routine will freeze or unfreeze the controls.
    If the value of  freeze  is TRUE, the controls will be frozen, else they
    will be unfrozen.
    NOTE: if the global variable  running_movie  is TRUE, then the routine does
    nothing.
    The routine returns nothing.
*/
flag freeze;
{
    Boolean sensitive;
    long frozen_time;
    struct timeval unfreeze_time;
    struct timezone tz;
    extern Widget dataclip_winpopup;
    static flag frozen = FALSE;
    static struct timeval freeze_time;

    if (freeze == frozen) return;
    sensitive = freeze ? False : True;
    if (!freeze)
    {
	if (gettimeofday (&unfreeze_time, &tz) != 0)
	{
	    fprintf (stderr, "Error getting time of day\t%s%c\n",
		     ERRSTRING, BEL);
	    exit (RV_SYS_ERROR);
	}
    }
    XtSetSensitive (main_shell, sensitive);
    XtSetSensitive (substance_winpopup, sensitive);
    XtSetSensitive (cont_hg_winpopup, sensitive);
    XtSetSensitive (threecolour_hg_winpopup, sensitive);
    if (mono_hg_winpopup != NULL) XtSetSensitive (mono_hg_winpopup, sensitive);
    XtSetSensitive (dataclip_winpopup, sensitive);
    XtSetSensitive (filter_winpopup, sensitive);
    frozen = freeze;
    XSync (XtDisplay (main_shell), False);
    if (freeze)
    {
	if (gettimeofday (&freeze_time, &tz) != 0)
	{
	    fprintf (stderr, "Error getting time of day\t%s%c\n",
		     ERRSTRING, BEL);
	    exit (RV_SYS_ERROR);
	}
    }
    else
    {
	frozen_time = (unfreeze_time.tv_sec - freeze_time.tv_sec) * 1000;
	frozen_time += (unfreeze_time.tv_usec - freeze_time.tv_usec) / 1000;
	/*fprintf (stderr, "Freeze time: %ld ms\n", frozen_time);*/
    }
}   /*  End Function freeze_controls  */

void reset_view ()
{
    rotation_specification rotation;
    extern flag using_spaceball;
    extern int roll_value, pitch_value, yaw_value;

    rotation.roll = 0.0;
    rotation.pitch = 0.0;
    rotation.yaw = 0.0;
    rotation.rho = (float) DEFAULT_DISTANCE;
    register_rotation_change (rotation);
    if ( !using_spaceball && (roll_sld != (Widget) NULL) )
    {
	roll_value = rotation.roll / PION180;
	pitch_value = rotation.pitch / PION180;
	yaw_value = rotation.yaw / PION180;
	XtVaSetValues (roll_sld, XtNvalue, roll_value, NULL);
	XtVaSetValues (pitch_sld, XtNvalue, pitch_value, NULL);
	XtVaSetValues (yaw_sld, XtNvalue, yaw_value, NULL);
    }
    compute ();
}   /*  End Function reset_view  */

flag offset_position_func (KWorldCanvas canvas, double x, double y,
			   unsigned int event_code, void *e_info,
			   void **f_info, double x_lin, double y_lin)
/*  [SUMMARY] Position event callback.
    [PURPOSE] This routine is a position event consumer for a world canvas.
    <canvas> The canvas on which the event occurred.
    <x> The horizontal world co-ordinate of the event.
    <y> The vertical world co-ordinate of the event.
    <event_code> The arbitrary event code.
    <e_info> A pointer to arbitrary event information.
    <f_info> A pointer to an arbitrary function information pointer.
    <x_lin> The horizontal linear world co-ordinate prior to the transform
    function being called.
    <y_lin> The vertical linear world co-ordinate prior to the transform
    function being called.
    [NOTE] The world co-ordinate values will have been transformed by the
    registered transform function (see [<canvas_register_transform_func>]).
    [RETURNS] TRUE if the event was consumed, else FALSE indicating that
    the event is still to be processed.
*/
{
    Widget w = (Widget) *f_info;
    char txt[STRING_LENGTH];
    extern iarray byte_cube;
    extern double byte_cube_scale, byte_cube_offset;

    if (byte_cube == NULL) return (FALSE);
    iarray_format_value (byte_cube, txt, x, byte_cube_scale, byte_cube_offset);
    XtVaSetValues (w,
		   XtNlabel, txt,
		   XtNheight, 20,
		   NULL);
    return (FALSE);
}   /*  End Function offset_position_func  */

void select_visual (KWorldCanvas canvas)
/*  This routine will select the visible canvas.
*/
{
    extern KWorldCanvas visible_worldcanvas;
    extern KWorldCanvas cyclops_image_pc_worldcanvas;
    extern KWorldCanvas cyclops_image_rgb_worldcanvas;
    extern KWorldCanvas left_image_pc_worldcanvas, right_image_pc_worldcanvas;
    extern KWorldCanvas left_image_rgb_worldcanvas,right_image_rgb_worldcanvas;
    extern KOverlayList cyclops_olist, left_olist, right_olist;
    extern Widget image_display;

    XtVaSetValues (image_display,
		   XkwNvisibleCanvas, canvas,
		   NULL);
    visible_worldcanvas = canvas;
    if ( (canvas == cyclops_image_pc_worldcanvas) ||
	 (canvas == cyclops_image_rgb_worldcanvas) )
	overlay_specify_canvas (cyclops_olist, canvas);
    if ( (canvas == left_image_pc_worldcanvas) ||
	 (canvas == right_image_pc_worldcanvas) )
    {
	overlay_specify_canvas (left_olist,
				left_image_pc_worldcanvas);
	overlay_specify_canvas (right_olist,
				right_image_pc_worldcanvas);
    }
    if ( (canvas == left_image_rgb_worldcanvas) ||
	 (canvas == right_image_rgb_worldcanvas) )
    {
	overlay_specify_canvas (left_olist,
				left_image_rgb_worldcanvas);
	overlay_specify_canvas (right_olist,
				right_image_rgb_worldcanvas);
    }
}   /*  End Function select_visual  */

void show_animate_control (flag display)
/*  [PURPOSE] This routine will show the animate control window.
    <display> If TRUE, the window is displayed, else it is popped down.
    [RETURNS] Nothing.
*/
{
    Widget animate_winpopup;
    extern Widget image_display;

    animate_winpopup = XtNameToWidget (image_display, "animatepopup");
    if (display) XtPopup (animate_winpopup, XtGrabNone);
    else XtPopdown (animate_winpopup);
}   /*  End Function show_animate_control  */

void refresh_display_window ()
/*  [PURPOSE] This routine will refresh the current display window. It
    determines which canvas is active and then refreshes it (and the
    corresponding right-eye canvas if appropriate).
    [RETURNS] Nothing.
*/
{
    extern Widget image_display;

    XkwImageDisplayRefresh (image_display, FALSE);
}   /*  End Function refresh_display_window  */

void do_rotate (int roll_inc, int yaw_inc, int pitch_inc)
/*  [SUMMARY] Rotate and recompute.
    <roll_inc> The number of degrees to roll by.
    <yaw_inc> The number of degrees to yaw by.
    <pitch_inc> The number of degrees to pitch by.
*/
{
    extern flag using_spaceball;
    extern rotation_specification rotation;

#ifdef HAVE_SPACEBALL
    if (using_spaceball)
    {
	spin_sb (PION180 * (float) roll_inc,
		 PION180 * (float) pitch_inc,
		 PION180 * (float) yaw_inc);
	compute ();
	return;
    }
#endif
    if (roll_inc != 0)
    {
	roll_value += roll_inc;
	if (roll_value < ANGLE_MIN) roll_value += ANGLE_RANGE;
	if (roll_value > ANGLE_MAX) roll_value -= ANGLE_RANGE;
	rotation.roll = PION180 * (float) roll_value;
	XtVaSetValues (roll_sld, XtNvalue, roll_value, NULL);
    }
    if (pitch_inc != 0)
    {
	pitch_value += pitch_inc;
	if (pitch_value < ANGLE_MIN) pitch_value += ANGLE_RANGE;
	if (pitch_value > ANGLE_MAX) pitch_value -= ANGLE_RANGE;
	rotation.pitch = PION180 * (float) pitch_value;
	XtVaSetValues (pitch_sld, XtNvalue, pitch_value, NULL);
    }
    if (yaw_inc != 0)
    {
	yaw_value += yaw_inc;
	if (yaw_value < ANGLE_MIN) yaw_value += ANGLE_RANGE;
	if (yaw_value > ANGLE_MAX) yaw_value -= ANGLE_RANGE;
	rotation.yaw = PION180 * (float) yaw_value;
	XtVaSetValues (yaw_sld, XtNvalue, yaw_value, NULL);
    }
    register_rotation_change (rotation);
    compute ();
}   /*  End Function do_rotate  */


/*  Private routines follow  */

static void quiescent_func ()
/*  This function is called when the CM tool or shell becomes quiescent.
    The routine returns nothing.
*/
{
    freeze_controls (FALSE);
}   /*  End Function quiescent_func  */

static void quit_cbk (Widget w, XtPointer client_data, XtPointer call_data)
/*  This is the quit button callback.
*/
{
    exit (RV_OK);
}   /*  End Function quit_cbk   */

static void shader_cbk (Widget w, XtPointer client_data, XtPointer call_data)
/*  This is the shader menu callback.
*/
{
    unsigned int shadercode = *(int *) call_data;
    CONST char *name;
    extern flag substances_needed;
    extern CONST char *current_shader_name;
    extern CONST char **shader_names;
    static unsigned int old_code = 0;

    name = shader_names[shadercode];
    current_shader_name = name;
    if (shadercode != old_code)
    {
	/*  Determine what should be popped down  */
	if (strcmp (name, "hot_gas_substances") == 0)
	{
	    XtPopdown (cont_hg_winpopup);
	    XtPopdown (threecolour_hg_winpopup);
	    XtPopdown (offset_winpopup);
	    XtPopdown (threshold_winpopup);
	    if (mono_hg_winpopup != NULL) XtPopdown (mono_hg_winpopup);
	}
	else if (strcmp (name, "hot_gas_continuous") == 0)
	{
	    XtPopdown (substance_winpopup);
	    XtPopdown (threecolour_hg_winpopup);
	    XtPopdown (offset_winpopup);
	    XtPopdown (threshold_winpopup);
	    if (mono_hg_winpopup != NULL) XtPopdown (mono_hg_winpopup);
	}
	else if (strcmp (name, "hot_gas_three_colour") == 0)
	{
	    XtPopdown (substance_winpopup);
	    XtPopdown (cont_hg_winpopup);
	    XtPopdown (offset_winpopup);
	    XtPopdown (threshold_winpopup);
	    if (mono_hg_winpopup != NULL) XtPopdown (mono_hg_winpopup);
	}
	else if (strcmp (name, "sum_voxels") == 0)
	{
	    XtPopdown (substance_winpopup);
	    XtPopdown (cont_hg_winpopup);
	    XtPopdown (threecolour_hg_winpopup);
	    XtPopdown (threshold_winpopup);
	    if (mono_hg_winpopup != NULL) XtPopdown (mono_hg_winpopup);
	}
	else if (strcmp (name, "hot_gas_mono") == 0)
	{
	    XtPopdown (substance_winpopup);
	    XtPopdown (offset_winpopup);
	    XtPopdown (threshold_winpopup);
	    XtPopdown (cont_hg_winpopup);
	    XtPopdown (threecolour_hg_winpopup);
	}
	else if (strcmp (name, "front_voxel") == 0)
	{
	    XtPopdown (substance_winpopup);
	    XtPopdown (offset_winpopup);
	    XtPopdown (cont_hg_winpopup);
	    XtPopdown (threecolour_hg_winpopup);
	    if (mono_hg_winpopup != NULL) XtPopdown (mono_hg_winpopup);
	}
	else
	{
	    XtPopdown (substance_winpopup);
	    XtPopdown (cont_hg_winpopup);
	    XtPopdown (threecolour_hg_winpopup);
	    XtPopdown (offset_winpopup);
	    XtPopdown (threshold_winpopup);
	    if (mono_hg_winpopup != NULL) XtPopdown (mono_hg_winpopup);
	}
    }
    if (strcmp (name, "hot_gas_substances") == 0)
    {
	XtPopup (substance_winpopup, XtGrabNone);
	substances_needed = TRUE;
    }
    else
    {
	substances_needed = FALSE;
    }
    if (strcmp (name, "hot_gas_continuous") == 0)
    {
	compute_cont_hg_info ();
	XtPopup (cont_hg_winpopup, XtGrabNone);
    }
    if (strcmp (name, "hot_gas_three_colour") == 0)
    {
	compute_threecolour_hg_info ();
	XtPopup (threecolour_hg_winpopup, XtGrabNone);
    }
    if (strcmp (name, "sum_voxels") == 0)
    {
	XtPopup (offset_winpopup, XtGrabNone);
    }
    if (strcmp (name, "hot_gas_mono") == 0)
    {
	compute_mono_hg_info ();
	XtPopup (mono_hg_winpopup, XtGrabNone);
    }
    if (strcmp (name, "front_voxel") == 0)
    {
	XtPopup (threshold_winpopup, XtGrabNone);
    }
    if (shadercode != old_code)
    {
	old_code = shadercode;
	register_shader (name);
    }
}   /*  End Function shader_cbk   */

static void projection_cbk (Widget w, XtPointer client_data,
			    XtPointer call_data)
/*  This is the projection menu callback.
*/
{
    unsigned int projectioncode = *(int *) call_data;
    static int projection_mode = VRENDER_PROJECTION_PARALLEL;
    /*static char function_name[] = "projection_cbk";*/

    if (projectioncode != projection_mode)
    {
	register_projection (projectioncode);
    }
    if (projectioncode != projection_mode) compute ();
    projection_mode = projectioncode;
}   /*  End Function projection_cbk   */

static Widget create_orient_win (main_shell)
Widget main_shell;
{
    Widget form, cnv, close_btn;
    extern KPixCanvas orientation_pixcanvas;

    orient_winpopup = XtVaCreatePopupShell ("orientwinpopup",
					    topLevelShellWidgetClass,
					    main_shell,
					    XtNtitle, "Orientation Window",
					    NULL);
    form = XtVaCreateManagedWidget ("form", formWidgetClass, orient_winpopup,
				    XtNborderWidth,0,
				    NULL);
    close_btn = XtVaCreateManagedWidget ("closeButton", commandWidgetClass,
					 form,
					 XtNlabel, "Close",
					 XtNtop, XtChainTop,
					 XtNbottom, XtChainTop,
					 XtNleft, XtChainLeft,
					 XtNright, XtChainRight,
					 XtNwidth, 200,
					 XtNheight, 20,
					 NULL);
    XtAddCallback (close_btn, XtNcallback, xtmisc_popdown_cbk,orient_winpopup);
    cnv = XtVaCreateManagedWidget ("pseudocolourCanvas", canvasWidgetClass,
				   form,
				   XtNfromVert, close_btn,
				   XtNleft, XtChainLeft,
				   XtNright, XtChainRight,
				   XtNwidth, 200,
				   XtNheight, 200,
				   XkwNsilenceUnconsumed, True,
				   XtNresizable, True,
				   XtNtop, XtChainTop,
				   XtNbottom, XtChainBottom,
				   NULL);
    XtRealizeWidget (orient_winpopup);
    XtVaGetValues (cnv,
		   XkwNmonoPixCanvas, &orientation_pixcanvas,
		   NULL);
    initialise_orientation_canvas (orientation_pixcanvas);
    return (orient_winpopup);
}   /*  End Function create_orient_win  */

static void slider_view_cbk (Widget w, XtPointer client_data,
			     XtPointer call_data)
/*  This is the slider rotate callback.
*/
{
    rotation_specification rotation;
    extern int roll_value, pitch_value, yaw_value, distance_value;

    rotation.roll = PION180 * (float) roll_value;
    rotation.pitch = PION180 * (float) pitch_value;
    rotation.yaw = PION180 * (float) yaw_value;
    rotation.rho = (float) distance_value;
    register_rotation_change (rotation);
}   /*  End Function slider_view_cbk   */

static void toggle_cbk (Widget w, XtPointer client_data, XtPointer call_data)
/*  This is the toggle callback.
*/
{
    *(flag *) client_data = (iaddr) call_data ? TRUE : FALSE;
    compute ();
}   /*  End Function toggle_cbk   */

static Widget create_histogram_win (Widget main_shell,
				    void (*x_func) (signed char x), int *x_ptr)
{
    KPixCanvas pixcanvas;
    KWorldCanvas worldcanvas;
    Widget form, cnv, close_btn, w, track_label, popup;
    /*static char function_name[] = "create_histogram_win";*/

    popup = XtVaCreatePopupShell ("offsetwinpopup", topLevelShellWidgetClass,
				  main_shell,
				     XtNtitle, "Offset Window",
				     NULL);
    form = XtVaCreateManagedWidget ("form", formWidgetClass, popup,
				    XtNborderWidth, 0,
				    NULL);
    close_btn = XtVaCreateManagedWidget ("closeButton", commandWidgetClass,
					 form,
					 XtNlabel, "Close",
					 XtNtop, XtChainTop,
					 XtNbottom, XtChainTop,
					 XtNleft, XtChainLeft,
					 XtNheight, 20,
					 NULL);
    XtAddCallback (close_btn, XtNcallback, xtmisc_popdown_cbk, popup);
    w = XtVaCreateManagedWidget ("track", labelWidgetClass, form,
				 XtNlabel, "Offset:",
				 XtNborderWidth, 0,
				 XtNfromHoriz, close_btn,
				 XtNtop, XtChainTop,
				 XtNbottom, XtChainTop,
				 XtNheight, 20,
				 XtNresizable, True,
				 NULL);
    track_label = w;
    cnv = XtVaCreateManagedWidget ("pseudocolourCanvas", canvasWidgetClass,
				   form,
				   XtNfromVert, close_btn,
				   XtNleft, XtChainLeft,
				   XtNright, XtChainRight,
				   XtNwidth, 500,
				   XtNheight, 400,
				   XkwNsilenceUnconsumed, True,
				   XtNresizable, True,
				   XtNtop, XtChainTop,
				   XtNbottom, XtChainBottom,
				   NULL);
    XtRealizeWidget (popup);
    XtVaGetValues (cnv,
		   XkwNmonoPixCanvas, &pixcanvas,
		   NULL);
    worldcanvas = setup_offset_control_canvas (pixcanvas, XtDisplay (cnv),
					       track_label, x_func, x_ptr);
    return (popup);
}   /*  End Function create_histogram_win  */

static void region_cbk (Widget w, XtPointer client_data, XtPointer call_data)
/*  This is the region callback.
*/
{
    iarray float_cube;
    int min_clip_val = -127;
    int max_clip_val = 127;
    int x, y, z;
    int xlen, ylen, zlen;
    int val;
    float fval, scale, offset;
    double d_min = -127.0;
    double d_max = 127.0;
    double raw_cube_scale, raw_cube_offset;
    DataclipRegions *regions = (DataclipRegions *) call_data;
    float min, max;
    extern iarray byte_cube, filtered_cube, unfiltered_cube;
    extern flag verbose;
    static char function_name[] = "region_cbk";

    if (regions->num_regions == 0)
    {
	XBell (XtDisplay (w), 100);
	return;
    }
    if (regions->num_regions != 1)
    {
	fprintf (stderr, "num_regions: %u is not 1\n",
			regions->num_regions);
	a_prog_bug (function_name);
    }
    freeze_controls (TRUE);
    if (verbose) fprintf (stderr, "Converting to K_BYTE array...\n");
    float_cube = (filtered_cube == NULL) ? unfiltered_cube : filtered_cube;
    xlen = iarray_dim_length (float_cube, 2);
    ylen = iarray_dim_length (float_cube, 1);
    zlen = iarray_dim_length (float_cube, 0);
    min = regions->minima[0];
    max = regions->maxima[0];
    if (regions->blank_data_outside_regions)
    {
	min_clip_val = -128;
	max_clip_val = -128;
    }
    fprintf (stderr, "Selected minimum: %e  maximum: %e\n", min, max);
    scale = 254.0 / (max - min);
    offset = -127.0 - scale * min;
    for (z = 0; z < zlen; ++z) for (y = 0; y < ylen; ++y)
    {
	for (x = 0; x < xlen; ++x)
	{
	    fval = F3 (float_cube, z, y, x);
	    if ( DOUBLE_IS_BLANK (fval) ) val = -128;
	    else if (fval < min) val = min_clip_val;
	    else if (fval > max) val = max_clip_val;
	    else val = (int) (fval * scale + offset);
	    B3 (byte_cube, z, y, x) = val;
	}
    }
    raw_cube_scale = 1.0;
    raw_cube_offset = 0.0;
    register_new_byte_cube (byte_cube, raw_cube_scale / scale,
			    raw_cube_offset - offset * raw_cube_scale / scale);
    if (slicer_3d != NULL)
    {
	XtVaSetValues (slicer_3d,
		       XkwNiarray, NULL,
		       NULL);
	XtVaSetValues (slicer_3d,
		       XkwNiarray, byte_cube,
		       XkwNminPtr, &d_min,
		       XkwNmaxPtr, &d_max,
		       NULL);
    }
    if (slicer_popup != NULL) XtPopup (slicer_popup, XtGrabNone);
}   /*  End Function region_cbk   */

static void cursor_2d_cbk (Widget w, XtPointer client_data,XtPointer call_data)
{
    unsigned int code = *(int *) call_data;
    extern unsigned int cursor_2d_mode;
    static char function_name[] = "cursor_2d_cbk";

    switch (code)
    {
      case CURSOR_2D_DUMB:
	XtPopdown (ray_winpopup);
	break;
      case CURSOR_2D_ACTIVE:
	XtPopup (ray_winpopup, XtGrabNone);
	break;
      default:
	fprintf (stderr, "Illegal cursor mode: %u\n", code);
	a_prog_bug (function_name);
	break;
    }
    cursor_2d_mode = code;
}   /*  End Function cursor_2d_cbk  */

static void create_ray_win (main_shell)
Widget main_shell;
{
    KPixCanvas pixcanvas;
    Widget form, cnv, close_btn;
    /*static char function_name[] = "create_ray_win";*/

    ray_winpopup = XtVaCreatePopupShell ("offsetwinpopup",
					 topLevelShellWidgetClass, main_shell,
					 XtNtitle, "Ray Window",
					 NULL);
    form = XtVaCreateManagedWidget ("form", formWidgetClass, ray_winpopup,
				    XtNborderWidth, 0,
				    NULL);
    close_btn = XtVaCreateManagedWidget ("closeButton", commandWidgetClass,
					 form,
					 XtNlabel, "Close",
					 XtNtop, XtChainTop,
					 XtNbottom, XtChainTop,
					 XtNleft, XtChainLeft,
					 XtNheight, 20,
					 NULL);
    XtAddCallback (close_btn, XtNcallback, xtmisc_popdown_cbk, ray_winpopup);
    cnv = XtVaCreateManagedWidget ("pseudocolourCanvas", canvasWidgetClass,
				   form,
				   XtNfromVert, close_btn,
				   XtNleft, XtChainLeft,
				   XtNright, XtChainRight,
				   XtNwidth, 500,
				   XtNheight, 400,
				   XkwNsilenceUnconsumed, True,
				   XtNresizable, True,
				   XtNtop, XtChainTop,
				   XtNbottom, XtChainBottom,
				   NULL);
    XtRealizeWidget (ray_winpopup);
    XtVaGetValues (cnv,
		   XkwNmonoPixCanvas, &pixcanvas,
		   NULL);
    setup_ray_display_canvas ( pixcanvas, XtDisplay (cnv) );
}   /*  End Function create_ray_win  */

static void mode_cbk (Widget w, XtPointer client_data, XtPointer call_data)
/*  This is the mode menu callback.
*/
{
    unsigned int mode = *(int *) call_data;
    extern unsigned int raycast_mode;
    /*static char function_name[] = "mode_cbk";*/

    if (mode != raycast_mode)
    {
	raycast_mode = mode;
	compute ();
    }
}   /*  End Function mode_cbk   */

static void eye_sep_cbk (Widget w, XtPointer client_data, XtPointer call_data)
/*  [PURPOSE] This routine is the eye separation callback.
*/
{
    double sep = (double) *(int *) call_data;

    /*  Compute distance independent of subcube size
	Assume 1 voxel corresponds to 1 pixel (invalid if image expanded)
	Assume screen is 305 mm across (17 inch MAG) with 960 pixels
	Hence 1 mm is 960 / 305 pixels
	If image size is doubled in x and y then 1 voxel corresponds to
	2 pixels
	*/
    sep *= 960.0 / 305.0;        /*  Distance in pixels   */
    /*  TODO: sep /= image_expand_factor  */
    register_eye_separation (sep);
}   /*  End Function eye_sep_cbk  */

static void hide_wireframe_cbk (Widget w, XtPointer client_data,
				XtPointer call_data)
/*  [PURPOSE] This routine is called when the user clicks the wireframe toggle.
*/
{
    *(flag *) client_data = (iaddr) call_data ? TRUE : FALSE;
    compute ();
}   /*  End Function hide_wireframe_cbk   */

EXTERN_FUNCTION (void draw_cut_plane, (float depth) );

static void depth_cbk (Widget w, XtPointer client_data, XtPointer call_data)
/*  [PURPOSE] This routine is called when the user changes the cut plane depth.
*/
{
    draw_cut_plane (*(int *) call_data);
}   /*  End Function depth_cbk  */

static flag cube_event_func (void *object, multi_array *multi_desc,
			     CONST char *domain, CONST char *name)
/*  [SUMMARY] Process a new cube event.
    <object> The object information pointer.
    <multi_desc> The multi_array descriptor.
    <domain> The domain from where the event was generated.
    <name> The name of the event in its domain.
    [RETURNS] TRUE if further callbacks should not be called, else FALSE.
*/
{
    iarray cube;
    double min = -127.0;
    double max = 127.0;
    char *ptr;
    char filename[STRING_LENGTH], stripped_filename[STRING_LENGTH];
    extern iarray unfiltered_cube, byte_cube;
    extern i_coord_3d subcube_start, subcube_end;
    extern Widget image_display, slicer_3d;
    extern Widget dataclip_winpopup, slicer_popup, filter_winpopup;
    extern Widget start_x_sld, end_x_sld, start_y_sld, end_y_sld;
    extern Widget start_z_sld, end_z_sld;
    static flag first_time = TRUE;
    /*static char function_name[] = "cube_event_func";*/

    if (strcmp (domain, "IMAGE") == 0) return (FALSE);
    if ( (strcmp (domain, "GLISH") == 0) && (strcmp (name, "cube") != 0) )
	return (FALSE);
    if (strcmp (domain, "CUBE") == 0)
    {
	filename[0] = '\0';
	freeze_controls (TRUE);
    }
    else
    {
	strcpy (filename, domain);
	strcat (filename, "::");
    }
    strcat (filename, name);
    if ( !process_cube (multi_desc) )
    {
	freeze_controls (FALSE);
	return (FALSE);
    }
    XtSetSensitive (dataclip_winpopup, True);
    if (unfiltered_cube == NULL)
    {
	/*  No unfiltered float cube: no clipping possible  */
	XtPopdown (dataclip_winpopup);
	XtPopdown (filter_winpopup);
	XtSetSensitive (rescale_btn, False);
	XtSetSensitive (filter_btn, False);
	if (slicer_popup != NULL) XtPopup (slicer_popup, XtGrabNone);
	if (slicer_3d != NULL) XtVaSetValues (slicer_3d,
					      XkwNiarray, byte_cube,
					      XkwNminPtr, &min,
					      XkwNmaxPtr, &max,
					      NULL);
	cube = byte_cube;
    }
    else
    {
	/*  Have unfiltered float cube: require clipping  */
	XtPopup (dataclip_winpopup, XtGrabNone);
	XkwDataclipNewArray (dataclip_winpopup, unfiltered_cube,
			     TOOBIG, -TOOBIG, TRUE);
	XtSetSensitive (rescale_btn, True);
	XtSetSensitive (filter_btn, True);
	cube = unfiltered_cube;
    }
    XtVaSetValues (start_x_sld,
		   XkwNmaximum, iarray_dim_length (cube, 2) - 1,
		   XtNvalue, subcube_start.x,
		   NULL);
    XtVaSetValues (start_y_sld,
		   XkwNmaximum, iarray_dim_length (cube, 1) - 1,
		   XtNvalue, subcube_start.y,
		   NULL);
    XtVaSetValues (start_z_sld,
		   XkwNmaximum, iarray_dim_length (cube, 0) - 1,
		   XtNvalue, subcube_start.z,
		   NULL);
    XtVaSetValues (end_x_sld,
		   XkwNmaximum, iarray_dim_length (cube, 2) - 1,
		   XtNvalue, subcube_end.x,
		   NULL);
    XtVaSetValues (end_y_sld,
		   XkwNmaximum, iarray_dim_length (cube, 1) - 1,
		   XtNvalue, subcube_end.y,
		   NULL);
    XtVaSetValues (end_z_sld,
		   XkwNmaximum, iarray_dim_length (cube, 0) - 1,
		   XtNvalue, subcube_end.z,
		   NULL);
    strcpy (stripped_filename, filename);
    if ( ( ptr = strrchr (stripped_filename, '.') ) != NULL )
    {
	if (strcmp (ptr, ".kf") == 0) *ptr = '\0';
    }
    XtVaSetValues (image_display,
		   XkwNimageName, stripped_filename,
		   NULL);
    if (!first_time) return (TRUE);
    first_time = FALSE;
    XtPopup (display_winpopup, XtGrabNone);
    return (TRUE);
}   /*  End Function cube_event_func  */

static void show_statistics_cbk (Widget w, XtPointer client_data,
				 XtPointer call_data)
/*  [PURPOSE] This routine is called when the user asks for statistics.
*/
{
    CONST char *filename = client_data;
    FILE *fp;
    extern KVolumeRenderContext main_vrend_ctx;
    extern long compute_time;

    if (filename == NULL)
    {
	vrender_show_stats (main_vrend_ctx, stderr);
	return;
    }
    if ( ( fp = fopen (filename, "w") ) == NULL )
    {
	fprintf (stderr, "Error opening file: \"%s\"\t%s\n",
		 filename, ERRSTRING);
	return;
    }
    vrender_show_stats (main_vrend_ctx, fp);
    fprintf (fp, "Compute time: %ld ms\n", compute_time);
    fclose (fp);
}   /*  End Function show_statistics_cbk  */

static void register_voxel_threshold (signed char threshold)
{
    extern int voxel_threshold;

    voxel_threshold = threshold;
}   /*  End Function register_voxel_threshold  */
