/*  mono_hg_xt.c

    Source file for  xray  (Karma volume rendering tool).

    Copyright (C) 1996-1997  Richard Gooch

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Richard Gooch may be reached by email at  karma-request@atnf.csiro.au
    The postal address is:
      Richard Gooch, c/o ATNF, P. O. Box 76, Epping, N.S.W., 2121, Australia.
*/

/*
    This file contains the code necessary for creating the mono Hot Gas
  control window.


    Written by      Richard Gooch   21-NOV-1996: Moved from main_xt.c

    Last updated by Richard Gooch   25-MAY-1997: Relabled apply button to
  "Compute".


*/
#include <stdio.h>
#include <errno.h>
#include <X11/Xatom.h>
#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/Xaw/Form.h>
#include <X11/Xaw/Command.h>
#include <karma.h>
#ifndef X11
#  define X11
#endif
#include <karma_xtmisc.h>
#include <karma_iarray.h>
#include <karma_c.h>
#include <Xkw/Canvas.h>
#include <Xkw/SimpleSlider.h>
#include "xraycast_3d.h"


/*  Private functions  */
STATIC_FUNCTION (void slider_mono_hg_cbk,
		 (Widget w, XtPointer client_data, XtPointer call_data) );
STATIC_FUNCTION (void mono_hg_apply_cbk, (Widget w, XtPointer client_data,
					  XtPointer call_data) );
STATIC_FUNCTION (flag new_cube_func,
		 (void *object, void *client1_data, void *call_data,
		  void *client2_data) );


/*  Private data  */
static Widget blank_min_sld = NULL;
static Widget blank_max_sld = NULL;


/*  Public functions follow  */

Widget create_mono_hg_win (Widget main_shell)
{
    Widget popup, form, cnv, close_btn, apply_btn;
    Widget Ialpha_sld, Oalpha_sld;
    extern KPixCanvas mono_hg_pixcanvas;
    extern KCallbackList new_byte_cube_event_list;
    extern int mono_hg_intensity_alpha_tenths,  mono_hg_opacity_alpha_tenths;
    extern int mono_hg_blank_min, mono_hg_blank_max;
    /*static char function_name[] = "create_mono_hg_win";*/

    popup = XtVaCreatePopupShell ("mono_hgwinpopup",
				  topLevelShellWidgetClass, main_shell,
				  XtNtitle, "Mono Hot Gas Window",
				  NULL);
    form = XtVaCreateManagedWidget ("form", formWidgetClass, popup,
				    XtNborderWidth, 0,
				    NULL);
    close_btn = XtVaCreateManagedWidget ("closeButton", commandWidgetClass,
					 form,
					 XtNlabel, "Close",
					 NULL);
    XtAddCallback (close_btn, XtNcallback, xtmisc_popdown_cbk, popup);
    apply_btn = XtVaCreateManagedWidget ("button", commandWidgetClass, form,
					 XtNlabel, "Compute",
					 XtNfromHoriz, close_btn,
					 NULL);
    XtAddCallback (apply_btn, XtNcallback, mono_hg_apply_cbk, NULL);
    Ialpha_sld = XtVaCreateManagedWidget ("slider", simpleSliderWidgetClass,
					  form,
					  XtNlabel, "Intensity Alpha",
					  XtNfromVert, apply_btn,
					  XtNwidth, 250,
					  XkwNminimum, 0,
					  XkwNmaximum, 50,
					  XkwNmodifier, 1,
					  XtNvalue,
					  mono_hg_intensity_alpha_tenths,
					  XkwNscaledFormat, "%.1f",
					  NULL);
    XkwSimpleSliderSetScale (Ialpha_sld, 0.1, 0.0, FALSE, TRUE);
    XtAddCallback (Ialpha_sld, XkwNvalueChangeCallback, slider_mono_hg_cbk,
		   (XtPointer) &mono_hg_intensity_alpha_tenths);
    Oalpha_sld = XtVaCreateManagedWidget ("slider", simpleSliderWidgetClass,
					  form,
					  XtNlabel, "Opacity Alpha",
					  XtNfromVert, apply_btn,
					  XtNfromHoriz, Ialpha_sld,
					  XtNwidth, 250,
					  XkwNminimum, 0,
					  XkwNmaximum, 50,
					  XkwNmodifier, 1,
					  XtNvalue,
					  mono_hg_opacity_alpha_tenths,
					  XkwNscaledFormat, "%.1f",
					  NULL);
    XkwSimpleSliderSetScale (Oalpha_sld, 0.1, 0.0, FALSE, TRUE);
    XtAddCallback (Oalpha_sld, XkwNvalueChangeCallback, slider_mono_hg_cbk,
		   (XtPointer) &mono_hg_opacity_alpha_tenths);
    blank_min_sld = XtVaCreateManagedWidget ("slider",
					     simpleSliderWidgetClass, form,
					     XtNlabel, "Blank Start",
					     XtNfromVert, Oalpha_sld,
					     XtNwidth, 500,
					     XkwNminimum, -128,
					     XkwNmaximum, 127,
					     XkwNmodifier, 1,
					     XtNvalue, -128,
					     NULL);
    XtAddCallback (blank_min_sld, XkwNvalueChangeCallback, slider_mono_hg_cbk,
		   (XtPointer) &mono_hg_blank_min);
    blank_max_sld = XtVaCreateManagedWidget ("slider",
					     simpleSliderWidgetClass, form,
					     XtNlabel, "Blank End",
					     XtNfromVert, blank_min_sld,
					     XtNwidth, 500,
					     XkwNminimum, -128,
					     XkwNmaximum, 127,
					     XkwNmodifier, 1,
					     XtNvalue, -128,
					     NULL);
    XtAddCallback (blank_max_sld, XkwNvalueChangeCallback, slider_mono_hg_cbk,
		   (XtPointer) &mono_hg_blank_max);
    cnv = XtVaCreateManagedWidget ("pseudocolourCanvas",canvasWidgetClass,form,
				   XtNfromVert, blank_max_sld,
				   XtNwidth, 500,
				   XtNheight, 200,
				   XkwNsilenceUnconsumed, True,
				   NULL);
    XtRealizeWidget (popup);
    XtVaGetValues (cnv,
		   XkwNmonoPixCanvas, &mono_hg_pixcanvas,
		   NULL);
    setup_mono_hg_canvas (mono_hg_pixcanvas);
    c_register_callback (&new_byte_cube_event_list, new_cube_func,
			 NULL, NULL, FALSE, NULL, FALSE, FALSE);
    return (popup);
}   /*  End Function create_mono_hg_win  */


/*  Private functions follow  */

static void slider_mono_hg_cbk (Widget w, XtPointer client_data,
				XtPointer call_data)
/*  This is the slider mono Hot Gas callback.
*/
{
    int *new_value = (int *) call_data;
    int *global_value = (int *) client_data;
    extern KPixCanvas mono_hg_pixcanvas;

    *global_value = *new_value;
    compute_mono_hg_info ();
    kwin_resize (mono_hg_pixcanvas, TRUE, 0, 0, 0, 0);
}   /*  End Function slider_mono_hg_cbk   */

static void mono_hg_apply_cbk (Widget w, XtPointer client_data,
			       XtPointer call_data)
/*  This is the mono Hot Gas apply callback.
*/
{
    extern KPixCanvas mono_hg_pixcanvas;

    compute_mono_hg_info ();
    kwin_resize (mono_hg_pixcanvas, TRUE, 0, 0, 0, 0);
    compute ();
}   /*  End Function mono_hg_apply_cbk   */

static flag new_cube_func (void *object, void *client1_data, void *call_data,
			   void *client2_data)
/*  [SUMMARY] Generic callback.
    [PURPOSE] This routine is called when object callbacks are called.
    <object> The object information pointer.
    <client1_data> The first client information pointer.
    <call_data> The call information pointer.
    <client2_data> The second client information pointer.
    [RETURNS] TRUE if further callbacks should not be called, else FALSE.
*/
{
    iarray cube = (iarray) call_data;
    extern double byte_cube_scale, byte_cube_offset;

    XtVaSetValues (blank_min_sld,
		   XkwNscaledUnit, iarray_value_name (cube),
		   NULL);
    XkwSimpleSliderSetScale (blank_min_sld, byte_cube_scale, byte_cube_offset,
			     TRUE, TRUE);
    XtVaSetValues (blank_max_sld,
		   XkwNscaledUnit, iarray_value_name (cube),
		   NULL);
    XkwSimpleSliderSetScale (blank_max_sld, byte_cube_scale, byte_cube_offset,
			     TRUE, TRUE);
    return (FALSE);
}   /*  End Function new_cube_func  */
