/*  save_movie.c

    Source file for  xray  (Karma volume rendering tool).

    Copyright (C) 1993-1996  Richard Gooch

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Richard Gooch may be reached by email at  karma-request@atnf.csiro.au
    The postal address is:
      Richard Gooch, c/o ATNF, P. O. Box 76, Epping, N.S.W., 2121, Australia.
*/

/*  This file contains code to save movies.


    Written by      Richard Gooch   10-JUL-1995

    Last updated by Richard Gooch   10-JUL-1995


*/

#include <stdio.h>
#include <math.h>
#include <errno.h>
#include <karma.h>
#include <karma_iarray.h>
#include <karma_dsxfr.h>
#include <karma_ds.h>
#include <karma_a.h>
#include <karma_m.h>
#include "xraycast_3d.h"

flag save_movie (CONST char *file, array_desc *im_arr_desc, CONST char *movie,
		 unsigned int num_frames)
/*  [PURPOSE] This routine will save a movie.
    <file> The filename to save the movie to.
    <im_arr_desc> A pointer to the image array descriptor.
    <cube> The movie data.
    <num_frames> The number of frames.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    flag retval;
    uaddr lengths[3];
    unsigned int data_type = K_FLOAT;
    multi_array *multi_desc;
    array_desc *arr_desc;
    dim_desc *xdim, *ydim;
    packet_desc *arr_pack_desc, *top_pack_desc;
    static CONST char *dim_names[] = {"Frame Number", "y", "x"};
    static CONST char *elem_name = "Data Value";

    lengths[0] = num_frames;
    lengths[1] = im_arr_desc->dimensions[0]->length;
    lengths[2] = im_arr_desc->dimensions[1]->length;
    if ( ( multi_desc =
	  ds_wrap_preallocated_n_element_array ( (char *) movie, 3,
						lengths, (double *) NULL,
						(double *) NULL,
						NULL,
						dim_names,
						1, &data_type, &elem_name ) )
	== NULL ) return (FALSE);
    top_pack_desc = multi_desc->headers[0];
    arr_desc = (array_desc *) top_pack_desc->element_desc[0];
    /*  Save information  */
    ydim = arr_desc->dimensions[1];
    xdim = arr_desc->dimensions[2];
    arr_pack_desc = arr_desc->packet;
    /*  Immoral hacks  */
    arr_desc->dimensions[1] = im_arr_desc->dimensions[0];
    arr_desc->dimensions[2] = im_arr_desc->dimensions[1];
    arr_desc->packet = im_arr_desc->packet;
    retval = dsxfr_put_multi (file, multi_desc);
    /*  Restore information  */
    arr_desc->dimensions[1] = ydim;
    arr_desc->dimensions[2] = xdim;
    arr_desc->packet = arr_pack_desc;
    ds_dealloc_multi (multi_desc);
    return (retval);
}   /*  End Function save_movie  */
