/*  slice_3d_xt.c

    Source file for  xray  (Karma volume rendering tool).

    Copyright (C) 1993-2000  Richard Gooch

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Richard Gooch may be reached by email at  karma-request@atnf.csiro.au
    The postal address is:
      Richard Gooch, c/o ATNF, P. O. Box 76, Epping, N.S.W., 2121, Australia.
*/

/*
    This file contains the code necessary for slicing the cube.


    Written by      Richard Gooch   31-JUL-1995: Copied from
  kslice_3d/main_xt.c

    Updated by      Richard Gooch   28-OCT-1995: Specified visual.

    Updated by      Richard Gooch   21-FEB-1996: Moved over to call to
  <refresh_display_window>.

    Updated by      Richard Gooch   27-MAY-1996: Cleaned code to keep
  gcc -Wall -pedantic-errors happy.

    Updated by      Richard Gooch   25-JUL-1996: Took account of change to
  <move_3d_cursor>.

    Updated by      Richard Gooch   23-AUG-1996: Took account of change to
  <move_3d_cursor> and added cursor event callback.

    Updated by      Richard Gooch   29-AUG-1996: Moved to <change_3d_cursor>

    Updated by      Richard Gooch   9-NOV-1996: Took account of changes to
  ThreeDeeSlice widget.

    Updated by      Richard Gooch   10-DEC-1998: Added support for TrueColour
  only displays.

    Updated by      Richard Gooch   11-DEC-1999: Removed unnecessary externs.

    Last updated by Richard Gooch   26-JAN-2000: Removed unnecessary includes.


*/
#include <stdio.h>
#include <errno.h>
#include <X11/Xatom.h>
#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/Xaw/Form.h>
#include <X11/Xaw/Command.h>
#include <karma.h>
#ifndef X11
#  define X11
#endif
#include <karma_iarray_def.h>
#include <karma_kcmap.h>
#include <karma_a.h>
#include <karma_c.h>
#include <Xkw/ThreeDeeSlice.h>
#include <Xkw/Filepopup.h>
#include <Xkw/Cmapwinpopup.h>
#include "xraycast_3d.h"


/*  Private functions  */
STATIC_FUNCTION (void cursor_cbk,
		 (Widget w, XtPointer client_data, XtPointer call_data) );
STATIC_FUNCTION (flag cursor_callback,
		 (void *object, void *client1_data, void *call_data,
		  void *client2_data) );


/*  Private data  */
static flag quench = FALSE;


Widget create_3d_slice_win (Widget main_shell, Kcolourmap kcmap,
			    Widget *slicer, Visual *visual, Colormap xcmap)
/*  [PURPOSE] This routine will create the slicer window.
    <main_shell> The application main shell.
    <kcmap> The Kcolourmap to use.
    <slicer> The ThreeDeeSlice widget will be written here.
    <visual> The visual to be used.
    <xcmap> The X colourmap to use.
    [RETURNS] The slicer popup.
*/
{
    Widget popup, form;
    Widget precomp_btn;
    extern KCallbackList cursor3d_event_list;
    /*static char function_name[] = "create_3d_slice_win";*/

    popup = XtVaCreatePopupShell ("slicewinpopup",
				  topLevelShellWidgetClass, main_shell,
				  XtNtitle, "Slice_3d Window",
				  NULL);
    form = XtVaCreateManagedWidget ("topForm", formWidgetClass, popup,
				    XtNborderWidth, 0,
				    NULL);
    precomp_btn = XtVaCreateManagedWidget ("button", commandWidgetClass, form,
					   XtNlabel, "Precompute",
					   NULL);
    *slicer = XtVaCreateManagedWidget ("slicer", threeDeeSliceWidgetClass,form,
				       XtNborderWidth, 0,
				       XtNhorizDistance, 0,
				       XtNvertDistance, 0,
				       XtNfromVert, precomp_btn,
				       XkwNkarmaColourmap, kcmap,
				       XkwNcanvasVisual, visual,
				       XkwNcanvasColourmap, xcmap,
				       NULL);
    XtAddCallback (*slicer, XkwNcursorCallback, cursor_cbk, NULL);
    XtAddCallback (precomp_btn, XtNcallback, XkwThreeDeeSlicePrecompute,
		   *slicer);
    c_register_callback (&cursor3d_event_list, cursor_callback, *slicer,
			 NULL, FALSE, NULL, FALSE, FALSE);
    XtRealizeWidget (popup);
    return (popup);
}   /*  End Function create_3d_slice_win   */


/*  Private routines follow  */

static void cursor_cbk (Widget w, XtPointer client_data, XtPointer call_data)
/*  [PURPOSE] This routine is the cursor callback. It is called whenever the
    cursor position changes.
*/
{
    XkwThreeDeeSliceCallbackData cursor =
	(XkwThreeDeeSliceCallbackData) call_data;

    quench = TRUE;
    change_3d_cursor (cursor->pixel, TRUE);
    quench = FALSE;
}   /*  End Function cursor_cbk  */

static flag cursor_callback (void *object, void *client1_data, void *call_data,
			     void *client2_data)
/*  [SUMMARY] Generic callback.
    [PURPOSE] This routine is called when object callbacks are called.
    <object> The object information pointer.
    <client1_data> The first client information pointer.
    <call_data> The call information pointer.
    <client2_data> The second client information pointer.
    [RETURNS] TRUE if further callbacks should not be called, else FALSE.
*/
{
    Widget slicer_3d = (Widget) object;
    Kcoord_3d *cursor = (Kcoord_3d *) call_data;
    extern iarray byte_cube;

    if (quench) return (FALSE);
    if (byte_cube == NULL) return (TRUE);
    XkwThreeDeeSliceSetCursor (slicer_3d, XkwThreeDeeSlicePixelCursor,
			       cursor->x, cursor->y, cursor->z);
    return (FALSE);
}   /*  End Function cursor_callback  */
