/*  subs_hg_xt.c

    Source file for  xray  (Karma volume rendering tool).

    Copyright (C) 1996-2000  Richard Gooch

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Richard Gooch may be reached by email at  karma-request@atnf.csiro.au
    The postal address is:
      Richard Gooch, c/o ATNF, P. O. Box 76, Epping, N.S.W., 2121, Australia.
*/

/*
    This file contains the code necessary for creating the Hot Gas substances
  control window.


    Written by      Richard Gooch   22-NOV-1996: Moved from main_xt.c

    Updated by      Richard Gooch   28-NOV-1996: Took account of substance
  changes.

    Updated by      Richard Gooch   13-APR-1997: Moved offset track label down
  to second row.

    Updated by      Richard Gooch   25-MAY-1997: Changed to
  <process_substances>.

    Updated by      Richard Gooch   11-DEC-1999: Removed unnecessary externs.

    Last updated by Richard Gooch   12-JUL-2000: Added #include <karma_kcmap.h>


*/
#include <stdio.h>
#include <errno.h>
#include <X11/Xatom.h>
#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/Xaw/Form.h>
#include <X11/Xaw/Command.h>
#include <X11/Xaw/Dialog.h>
#include <karma.h>
#ifndef X11
#  define X11
#endif
#include <karma_xtmisc.h>
#include <karma_iarray.h>
#include <karma_canvas.h>
#include <karma_kcmap.h>
#include <karma_dsxfr.h>
#include <karma_ds.h>
#include <karma_xc.h>
#include <karma_a.h>
#include <karma_m.h>
#include <Xkw/Canvas.h>
#include <Xkw/Filepopup.h>
#include <Xkw/Filewin.h>
#include <Xkw/SimpleSlider.h>
#include "xraycast_3d.h"


/*  Private functions  */
STATIC_FUNCTION (void subsfileselect_cbk, (Widget w, XtPointer client_data,
					   XtPointer call_data) );
STATIC_FUNCTION (void save_substances_cbk, (Widget w, XtPointer client_data,
					    XtPointer call_data) );
STATIC_FUNCTION (void apply_substances_cbk, (Widget w, XtPointer client_data,
					     XtPointer call_data) );
STATIC_FUNCTION (void slider_num_substances_cbk,
		 (Widget w, XtPointer client_data, XtPointer call_data) );


/*  Private data  */
static Widget substance_control_track_label = NULL;
static Kcolourmap subs_cmap = NULL;
static Widget num_substances_sld = NULL;


/*  Public functions follow  */

Widget create_substance_win (Widget main_shell, Visual *visual)
{
    Kdisplay dpy_handle;
    unsigned int visual_type;
    Colormap cmap;
    unsigned long fg, bg;
    Widget dialog;
    Widget form, cnv, close_btn, filewin, load_btn, save_btn, apply_btn, w;
    Widget popup;
    extern KPixCanvas substance_pixcanvas;
    extern int num_substances;
    static char function_name[] = "create_substance_win";

    popup = XtVaCreatePopupShell ("substancewinpopup",
				  topLevelShellWidgetClass, main_shell,
				  XtNtitle, "Substance Window",
				  NULL);
    form = XtVaCreateManagedWidget ("form", formWidgetClass, popup,
				    XtNborderWidth, 0,
				    NULL);
    close_btn = XtVaCreateManagedWidget ("closeButton", commandWidgetClass,
					 form,
					 XtNlabel, "Close",
					 XtNtop, XtChainTop,
					 XtNbottom, XtChainTop,
					 XtNleft, XtChainLeft,
					 XtNheight, 20,
					 NULL);
    XtAddCallback (close_btn, XtNcallback, xtmisc_popdown_cbk, popup);
    filewin = XtVaCreatePopupShell
	("filewinPopup", filepopupWidgetClass, main_shell,
	 XkwNfilenameTester, XkwFilewinStandardFileTester_3D,
	 XtNtitle, "Xray Substance File Selector",
	 NULL);
    XtAddCallback (filewin, XkwNfileSelectCallback, subsfileselect_cbk, NULL);
    load_btn = XtVaCreateManagedWidget ("button", commandWidgetClass, form,
					XtNlabel, "Load",
					XtNfromHoriz, close_btn,
					XtNtop, XtChainTop,
					XtNbottom, XtChainTop,
					XtNheight, 20,
					NULL);
    XtAddCallback (load_btn, XtNcallback, xtmisc_popup_cbk, filewin);
    save_btn = XtVaCreateManagedWidget ("button", commandWidgetClass, form,
					XtNlabel, "Save",
					XtNfromHoriz, load_btn,
					XtNtop, XtChainTop,
					XtNbottom, XtChainTop,
					XtNheight, 20,
					NULL);
    apply_btn = XtVaCreateManagedWidget ("button", commandWidgetClass, form,
					 XtNlabel, "Apply",
					 XtNfromHoriz, save_btn,
					 XtNtop, XtChainTop,
					 XtNbottom, XtChainTop,
					 XtNheight, 20,
					 NULL);
    XtAddCallback (apply_btn, XtNcallback, apply_substances_cbk, NULL);
    w = XtVaCreateManagedWidget ("track", labelWidgetClass, form,
				 XtNlabel, "Offset:",
				 XtNborderWidth, 0,
				 XtNfromVert, apply_btn,
				 XtNtop, XtChainTop,
				 XtNbottom, XtChainTop,
				 XtNheight, 20,
				 XtNresizable, True,
				 NULL);
    substance_control_track_label = w;
    dialog = XtVaCreateManagedWidget ("filename", dialogWidgetClass, form,
				      XtNlabel,
				      "Output Substance Arrayfile name:",
				      XtNborderWidth, 0,
				      XtNfromVert, w,
				      XtNvalue, "newsubs",
				      XtNresizable, TRUE,
				      NULL);
    XtAddCallback (save_btn, XtNcallback, save_substances_cbk, dialog);
    w = XtVaCreateManagedWidget ("slider", simpleSliderWidgetClass, form,
				 XtNlabel, "Number of Substances",
				 XtNfromVert, dialog,
				 XtNwidth, 300,
				 XkwNminimum, 2,
				 XkwNmaximum, MAX_SUBSTANCES,
				 XkwNmodifier, 1,
				 XtNvalue, num_substances,
				 XkwNvaluePtr, &num_substances,
				 NULL);
    num_substances_sld = w;
    XtAddCallback (w, XkwNvalueChangeCallback, slider_num_substances_cbk,NULL);
    cnv = XtVaCreateManagedWidget ("canvas", canvasWidgetClass,
				   form,
				   XtNfromVert, w,
				   XtNleft, XtChainLeft,
				   XtNright, XtChainRight,
				   XtNwidth, 500,
				   XtNheight, 400,
				   XtNvisual, visual,
				   XkwNretainFgBg, True,
				   XkwNsilenceUnconsumed, True,
				   XtNresizable, True,
				   XtNtop, XtChainTop,
				   XtNbottom, XtChainBottom,
				   NULL);
    XtRealizeWidget (popup);
    XtVaGetValues (cnv,
		   XkwNmonoPixCanvas, &substance_pixcanvas,
		   XtNcolormap, &cmap,
		   XtNforeground, &fg,
		   XtNbackground, &bg,
		   NULL);
    kwin_get_attributes (substance_pixcanvas,
			 KWIN_ATT_VISUAL, &visual_type,
			 KWIN_ATT_END);
    if (visual_type == KWIN_VISUAL_PSEUDOCOLOUR)
    {
	if ( ( dpy_handle = xc_get_dpy_handle (XtDisplay (cnv), cmap) )
	     == NULL )
	{
	    fprintf (stderr, "Error getting display handle\n");
	    a_prog_bug (function_name);
	}
    }
    else dpy_handle = NULL;
    setup_substance_canvases (substance_pixcanvas, dpy_handle,
			      substance_control_track_label, fg, bg,
			      &subs_cmap);
    return (popup);
}   /*  End Function create_substance_win  */

static void subsfileselect_cbk (Widget w, XtPointer client_data,
				XtPointer call_data)
{
    char *arrayfile_name;
    multi_array *multi_desc;
    extern KVolumeRenderContext main_vrend_ctx;
    extern int num_substances;
    extern substance substances[MAX_SUBSTANCES];
    /*static char function_name[] = "load_substances";*/

    arrayfile_name = (char *) call_data;
    if ( ( multi_desc = dsxfr_get_multi (arrayfile_name, FALSE, K_CH_MAP_LOCAL,
					 FALSE) )
	 == NULL ) return;
    if ( ( num_substances =
	   scan_substance_structure (substances, multi_desc->headers[0],
				     multi_desc->data[0]) ) < 1 )
    {
	ds_dealloc_multi (multi_desc);
	return;
    }
    ds_dealloc_multi (multi_desc);
    XtPopdown (w);
    XSync (XtDisplay (w), False);
    process_substances (main_vrend_ctx);
    XtVaSetValues (num_substances_sld,
		   XtNvalue, num_substances,
		   NULL);
}   /*  End Function subsfileselect_cbk  */

static void save_substances_cbk (Widget w, XtPointer client_data,
				 XtPointer call_data)
{
    char *arrayfile_name;
    Widget dialog = (Widget) client_data;
    multi_array *multi_desc;
    extern int num_substances;
    extern substance substances[MAX_SUBSTANCES];
    static char function_name[] = "save_substances_cbk";

    arrayfile_name = XawDialogGetValueString (dialog);
    if (arrayfile_name == NULL) arrayfile_name = "newsubs";
    if (strlen (arrayfile_name) < 1) arrayfile_name = "newsubs";
    if ( ( multi_desc = ds_alloc_multi (1) ) == NULL )
    {
	m_abort (function_name, "multi_array descriptor");
    }
    if ( !make_substance_structure (substances, num_substances,
				    multi_desc->headers, multi_desc->data) )
    {
	m_abort (function_name, "substance structure");
    }
    if (dsxfr_put_multi (arrayfile_name, multi_desc) != TRUE)
    {
	fprintf (stderr, "Error writing substance information\n");
    }
    ds_dealloc_multi (multi_desc);
}   /*  End Function save_substances_cbk  */

static void apply_substances_cbk (Widget w, XtPointer client_data,
				  XtPointer call_data)
{
    unsigned int count;
    extern KVolumeRenderContext main_vrend_ctx;
    extern int num_substances;
    extern substance substances[MAX_SUBSTANCES];
    extern substance temp_substances[MAX_SUBSTANCES];

    for (count = 0; count < num_substances; ++count)
    {
	substances[count].max_voxel = temp_substances[count].max_voxel;
	substances[count].red = temp_substances[count].red;
	substances[count].green = temp_substances[count].green;
	substances[count].blue = temp_substances[count].blue;
	substances[count].opacity = temp_substances[count].opacity;
    }
    process_substances (main_vrend_ctx);
}   /*  End Function apply_substances_cbk  */

static void slider_num_substances_cbk (Widget w, XtPointer client_data,
				       XtPointer call_data)
/*  This is the slider number of substances callback.
*/
{
    extern KPixCanvas substance_pixcanvas;
    extern int num_substances;

    if (subs_cmap != NULL) kcmap_change (subs_cmap, NULL, num_substances,
					 FALSE);
    kwin_resize (substance_pixcanvas, TRUE, 0, 0, 0, 0);
}   /*  End Function slider_threecolour_hg_cbk   */
