/*  threecolour_hg.c

    Source file for  xray  (Karma volume rendering tool).

    Copyright (C) 1996-1998  Richard Gooch

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Richard Gooch may be reached by email at  karma-request@atnf.csiro.au
    The postal address is:
      Richard Gooch, c/o ATNF, P. O. Box 76, Epping, N.S.W., 2121, Australia.
*/

/*  This file contains the source for the three colour Hot Gas colourmap
  display.


    Written by      Richard Gooch   6-MAY-1996

    Updated by      Richard Gooch   26-MAY-1996: Cleaned code to keep
  gcc -Wall -pedantic-errors happy.

    Updated by      Richard Gooch   22-NOV-1996: Supported peak opacities.

    Last updated by Richard Gooch   25-MAY-1997: Made use of
  <transmit_fullcolour_info> function.


*/
#include <stdio.h>
#include <math.h>
#include <karma.h>
#include <k_event_codes.h>
#include <karma_overlay.h>
#include <karma_canvas.h>
#include <karma_a.h>
#include "xraycast_3d.h"


/*  Private functions  */


/*  Private data  */


/*  Public functions follow  */

void compute_threecolour_hg_info ()
/*  This routine will compute the three colour Hot Gas information.
    The alpha value must be given by  alpha  .
    The routine returns nothing.
*/
{
    int i_count, index;
    float section_start, section_end;
    float Ialpha, Oalpha, peak_opacity;
    float f_count, intensity, opacity;
    float red, green, blue;
    extern flag byte_cube_valid;
    extern KVolumeRenderContext main_vrend_ctx;
    extern int threecolour_hg_section1_red_intensity;
    extern int threecolour_hg_section1_green_intensity;
    extern int threecolour_hg_section1_blue_intensity;
    extern int threecolour_hg_section1_intensity_alpha_tenths;
    extern int threecolour_hg_section1_opacity_alpha_tenths;
    extern int threecolour_hg_section1_opacity_peak;
    extern int threecolour_hg_section1_end;
    extern int threecolour_hg_section2_red_intensity;
    extern int threecolour_hg_section2_green_intensity;
    extern int threecolour_hg_section2_blue_intensity;
    extern int threecolour_hg_section2_intensity_alpha_tenths;
    extern int threecolour_hg_section2_opacity_alpha_tenths;
    extern int threecolour_hg_section2_opacity_peak;
    extern int threecolour_hg_section2_end;
    extern int threecolour_hg_section3_red_intensity;
    extern int threecolour_hg_section3_green_intensity;
    extern int threecolour_hg_section3_blue_intensity;
    extern int threecolour_hg_section3_intensity_alpha_tenths;
    extern int threecolour_hg_section3_opacity_alpha_tenths;
    extern int threecolour_hg_section3_opacity_peak;
    extern int threecolour_hg_blank_min, threecolour_hg_blank_max;
    extern float fullcolour_hg_reds[FULLCOLOUR_HOT_GAS_CMAP_SIZE];
    extern float fullcolour_hg_greens[FULLCOLOUR_HOT_GAS_CMAP_SIZE];
    extern float fullcolour_hg_blues[FULLCOLOUR_HOT_GAS_CMAP_SIZE];
    extern float fullcolour_hg_opacities[FULLCOLOUR_HOT_GAS_CMAP_SIZE];
    /*static char function_name[] = "compute_threecolour_hg_info";*/

    if (!byte_cube_valid) return;
    for (i_count = -127; i_count < 128; ++i_count)
    {
	f_count = i_count;
	index = i_count + 127;
	if ( (i_count >= threecolour_hg_blank_min) &&
	     (i_count < threecolour_hg_blank_max) )
	{
	    /*  Kill it entirely  */
	    fullcolour_hg_reds[index] = 0.0;
	    fullcolour_hg_greens[index] = 0.0;
	    fullcolour_hg_blues[index] = 0.0;
	    fullcolour_hg_opacities[index] = 0.0;
	    continue;
	}
	/*  Generate values depending on colour  */
	Ialpha = 0.0;
	Oalpha = 0.0;
	red = 0.0;
	green = 0.0;
	blue = 0.0;
	section_start = -127.0;
	section_end = 127.0;
	if (i_count <= threecolour_hg_section1_end)
	{
	    /*  Section 1  */
	    Ialpha = 0.1*(float)threecolour_hg_section1_intensity_alpha_tenths;
	    Oalpha = 0.1 * (float)threecolour_hg_section1_opacity_alpha_tenths;
	    peak_opacity = 0.01 * (float) threecolour_hg_section1_opacity_peak;
	    red = (float) threecolour_hg_section1_red_intensity;
	    green = (float) threecolour_hg_section1_green_intensity;
	    blue = (float) threecolour_hg_section1_blue_intensity;
	    section_start = -127.0;
	    section_end = threecolour_hg_section1_end;
	}
	else if (i_count <= threecolour_hg_section2_end)
	{
	    /*  Section 2  */
	    Ialpha = 0.1*(float)threecolour_hg_section2_intensity_alpha_tenths;
	    Oalpha = 0.1 * (float)threecolour_hg_section2_opacity_alpha_tenths;
	    peak_opacity = 0.01 * (float) threecolour_hg_section2_opacity_peak;
	    red = (float) threecolour_hg_section2_red_intensity;
	    green = (float) threecolour_hg_section2_green_intensity;
	    blue = (float) threecolour_hg_section2_blue_intensity;
	    section_start = threecolour_hg_section1_end;
	    section_end = threecolour_hg_section2_end;
	}
	else
	{
	    /*  Section 3  */
	    Ialpha = 0.1*(float)threecolour_hg_section3_intensity_alpha_tenths;
	    Oalpha = 0.1 * (float)threecolour_hg_section3_opacity_alpha_tenths;
	    peak_opacity = 0.01 * (float) threecolour_hg_section3_opacity_peak;
	    red = (float) threecolour_hg_section3_red_intensity;
	    green = (float) threecolour_hg_section3_green_intensity;
	    blue = (float) threecolour_hg_section3_blue_intensity;
	    section_start = threecolour_hg_section2_end;
	    section_end = 127;
	}
	red /= (float) INTENSITY_GRADATIONS;
	green /= (float) INTENSITY_GRADATIONS;
	blue /= (float) INTENSITY_GRADATIONS;
	/*  Compute current intensity  */
	if (Ialpha >= 0.0)
	{
	    /*  Intensity increases for increasing voxel value  */
	    intensity = (f_count -section_start) /(section_end -section_start);
	    intensity = pow (intensity, Ialpha);
	}
	else
	{
	    /*  Intensity decreases for increasing voxel value  */
	    intensity = (section_end - f_count) / (section_end -section_start);
	    intensity = pow (intensity, -Ialpha);
	}
	/*  Compute current opacity  */
	if (Oalpha >= 0.0)
	{
	    /*  Opacity increases for increasing voxel value  */
	    opacity = (f_count - section_start) / (section_end -section_start);
	    opacity = peak_opacity * pow (opacity, Oalpha);
	}
	else
	{
	    /*  Opacity decreases for increasing voxel value  */
	    opacity = (section_end - f_count) / (section_end - section_start);
	    opacity = peak_opacity * pow (opacity, -Oalpha);
	}
	fullcolour_hg_reds[index] = red * intensity;
	fullcolour_hg_greens[index] = green * intensity;
	fullcolour_hg_blues[index] = blue * intensity;
	fullcolour_hg_opacities[index] = opacity;
    }
    transmit_fullcolour_info (main_vrend_ctx);
}   /*  End Function compute_threecolour_hg_info  */
