/*
$Id: wavelet.c,v 1.1 1996/04/23 03:16:27 toosterl Exp toosterl $
$Log: wavelet.c,v $
Revision 1.1  1996/04/23 03:16:27  toosterl
Initial revision

*/
#define   WAVELET   1

#include <stdio.h>
#include <karma_iarray.h>
/*#include "util.h"*/

#include "wavelet.h"
#include "memutil.h"

static void sort ();

/*
  This file contains the source for the wavelet routines on KARMA iarrays


  Updates:
    19-Apr-95:  started writing
    14-May-95:  added filter_mode flag
    17-May-95:  added routine mask_filter
    23-Jul-95:  rewrote iarray_to_wavelet, renamed to iarray_wavelet_trans
                now use iarray_wavelet_2D/3D, added 3 dim transform
                retained iarray_to_wavelet for compatibility
*/    



int iarray_to_wavelet(iarray inarray, int wavelet_type, int level, 
                      iarray outarray, int channel, int filter_mode)
{
  iarray    array1, array2, array3, in, out, tmp;
  int       xlen, ylen, dimen, outxlen, outylen, extralen;
  int       oxlen, oylen, ozlen;
  int       data_type, k, x, y, step, step2;
  double    zero = 0.0;
  
  
  if (inarray == NULL) {
    return(FALSE);
  }
  if (outarray == NULL) {
    return(FALSE);
  }

  if (level < 0) {
    return(FALSE);
  }

  if ((data_type = iarray_type(inarray)) != K_FLOAT) {
    return(FALSE);
  }
  if ((data_type = iarray_type(outarray)) != K_FLOAT) {
    return(FALSE);
  }
  
  dimen = iarray_num_dim(inarray);
  if ((dimen != 2) && (dimen != 3)) {
    return(FALSE);
  }
  
  k = iarray_num_dim(outarray);
  if (k != 3) {
    return(FALSE);
  }
  
  xlen = iarray_dim_length(inarray, dimen-1);
  ylen = iarray_dim_length(inarray, dimen-2);

  oxlen = iarray_dim_length(outarray, 2);
  oylen = iarray_dim_length(outarray, 1);
  ozlen = iarray_dim_length(outarray, 0); 

  if ( (oxlen != xlen) || (oylen != ylen) || (ozlen != (level+1))) {
    return(FALSE);
  }

  switch (filter_mode) {
    default         :
    case WAVELET_XY :
    case WAVELET_XV :
    case WAVELET_YV : {
      switch (wavelet_type) {
        case  SPLINE_ATROUS : {
          if (level == 0) {                   /* just a copy of the data */
            extralen = 0;
    
            if (dimen == 2) {
              for (x = 0; x < xlen; x++) {
                for (y = 0; y < ylen; y++) {
                  F3(outarray, 0, y, x) = F2(inarray, y, x);
                }
              }
            } else {
              for (x = 0; x < xlen; x++) {
                for (y = 0; y < ylen; y++) {
                  F3(outarray, 0, y, x) = F3(inarray, channel, y, x);
                }
              }
            }
            return(TRUE);
          }
    
          extralen = 2  << (level-1);
    
    
          outxlen   = xlen + 2 * extralen;
          outylen   = ylen + 2 * extralen;
    
          array1 = iarray_create_2D(outylen, outxlen, K_FLOAT);
          array2 = iarray_create_2D(outylen, outxlen, K_FLOAT);
          array3 = iarray_create_2D(outylen, outxlen, K_FLOAT);
    
          zero = 0.0;
          iarray_fill(array1, &zero);
          iarray_fill(array2, &zero);
          iarray_fill(array3, &zero);
    
          if (dimen == 2) {
            for (x = 0; x < xlen; x++) {
              for (y = 0; y < ylen; y++) {
                F2(array1, y+extralen, x+extralen) = F2(inarray, y, x);
              }
            }
          } else {
            for (x = 0; x < xlen; x++) {
              for (y = 0; y < ylen; y++) {
                F2(array1, y+extralen, x+extralen) = 
                                                    F3(inarray, channel, y, x);
              }
            }
          }

          in = array1;
          out = array3;
          
          step = 1;
          step2 = 2*step;
          for (k = 1; k <= level; k++) {
    
            for (y = extralen; y < ylen+extralen; y++) {
              for (x = extralen; x < xlen+extralen; x++) {
                F2(array2, y, x) = 0.0625* (F2(in, y, x - step2) + 
                                            F2(in, y, x + step2))    +
                                   0.25  * (F2(in, y, x - step)  + 
                                            F2(in, y, x + step))     +
                                   0.375 *  F2(in, y, x);
              }
            }
    
            for (x = extralen; x < xlen+extralen; x++) {
              for (y = extralen; y < ylen+extralen; y++) {
                F2(out, y, x) = 0.0625* (F2(array2, y - step2, x) + 
                                         F2(array2, y + step2, x))       +
                                0.25  * (F2(array2, y - step,  x) + 
                                         F2(array2, y + step,  x))       +
                                0.375 *  F2(array2, y, x);
              }
            }
    
            for (x = 0; x < xlen; x++) {
              for (y = 0; y < ylen; y++) {
                F3(outarray, k-1, y, x) = F2(in,  y+extralen, x+extralen) - 
                                          F2(out, y+extralen, x+extralen);
              }
            }
            tmp   = out;
            out   = in;
            in    = tmp;
            step  = 2 * step;
            step2 = 2 * step;
          }

          for (x = 0; x < xlen; x++) {
            for (y = 0; y < ylen; y++) {
              F3(outarray, level, y, x) = F2(in, y+extralen, x+extralen);
            }
          }
          iarray_dealloc(array1);
          iarray_dealloc(array2);
          iarray_dealloc(array3);

          break;
        }
        default             :
          return(FALSE);
     }
     break;
    }
    case WAVELET_X :
    case WAVELET_Y :
    case WAVELET_Z : {
      switch (wavelet_type) {
        case  SPLINE_ATROUS : {
          if (level == 0) {                   /* just a copy of the data */
            extralen = 0;
    
            if (dimen == 2) {
              for (x = 0; x < xlen; x++) {
                for (y = 0; y < ylen; y++) {
                  F3(outarray, 0, y, x) = F2(inarray, y, x);
                }
              }
            } else {
              for (x = 0; x < xlen; x++) {
                for (y = 0; y < ylen; y++) {
                  F3(outarray, 0, y, x) = F3(inarray, channel, y, x);
                }
              }
            }
            return(TRUE);
          }
    
          extralen = 2  << (level-1);
    
          outxlen   = xlen + 2 * extralen;
          outylen   = ylen;
    
          array1 = iarray_create_2D(outylen, outxlen, K_FLOAT);
          array2 = iarray_create_2D(outylen, outxlen, K_FLOAT);
    
          zero = 0.0;
          iarray_fill(array1, &zero);
          iarray_fill(array2, &zero);
    
          if (dimen == 2) {
            for (y = 0; y < ylen; y++) {
              for (x = 0; x < xlen; x++) {
                F2(array1, y, x+extralen) = F2(inarray, y, x);
              }
            }
          } else {
            for (y = 0; y < ylen; y++) {
              for (x = 0; x < xlen; x++) {
                F2(array1, y, x+extralen) =  F3(inarray, channel, y, x);
              }
            }
          }
    
          in = array1;
          out = array2;
          
          step = 1;
          step2 = 2*step;
          for (k = 1; k <= level; k++) {

            for (y = 0; y < ylen; y++) {
              for (x = extralen; x < xlen+extralen; x++) {
                F2(out, y, x) = 0.0625* (F2(in, y, x - step2) + 
                                            F2(in, y, x + step2))    +
                                   0.25  * (F2(in, y, x - step)  + 
                                            F2(in, y, x + step))     +
                                   0.375 *  F2(in, y, x);
              }
            }

            for (y = 0; y < ylen; y++) {
              for (x = 0; x < xlen; x++) {
                F3(outarray, k-1, y, x) = F2(in,  y, x+extralen) - 
                                          F2(out, y, x+extralen);
              }
            }
            tmp   = out;
            out   = in;
            in    = tmp;
            step  = 2 * step;
            step2 = 2 * step;
          }
          for (y = 0; y < ylen; y++) {
            for (x = 0; x < xlen; x++) {
              F3(outarray, level, y, x) = F2(in, y, x+extralen);
            }
          }
          iarray_dealloc(array1);
          iarray_dealloc(array2);

          break;
        }
        default             :
          return(FALSE);
      }
      break;
    }
  }
  
  return(TRUE);
}

int iarray_wavelet_trans(iarray inarray, int wavelet_type, int level, 
                      iarray outarray, int channel, int filter_mode)
{
  iarray    array1, array2, array3, array4, in, out, tmp;
  int       xlen, ylen, zlen, dimen;
  int       oxlen, oylen, ozlen, ollen, outdimen;
  int       data_type, k, x, y, z, step, step2;
  double    zero = 0.0;
  
  
  if (inarray == NULL) {
    return(FALSE);
  }
  if (outarray == NULL) {
    return(FALSE);
  }

  if (level < 0) {
    return(FALSE);
  }

  if ((data_type = iarray_type(inarray)) != K_FLOAT) {
    return(FALSE);
  }
  if ((data_type = iarray_type(outarray)) != K_FLOAT) {
    return(FALSE);
  }
  
  switch (filter_mode) {
    default         :
    case WAVELET_XY :
    case WAVELET_XV :
    case WAVELET_YV : {

      dimen = iarray_num_dim(inarray);
      if ((dimen != 2) && (dimen != 3)) {
        return(FALSE);
      }
  
      outdimen = iarray_num_dim(outarray);
      if (outdimen != 3) {
        return(FALSE);
      }
  
      xlen = iarray_dim_length(inarray, dimen-1);
      ylen = iarray_dim_length(inarray, dimen-2);

      oxlen = iarray_dim_length(outarray, outdimen-1);
      oylen = iarray_dim_length(outarray, outdimen-2);
      ozlen = iarray_dim_length(outarray, outdimen-3); 
  
      if ( (oxlen != xlen) || (oylen != ylen) || (ozlen != (level+1))) {
        return(FALSE);
      }

      switch (wavelet_type) {
        default : 
        case  SPLINE_ATROUS : {
          if (level == 0) {                   /* just a copy of the data */
    
            if (dimen == 2) {
              for (y = 0; y < ylen; y++) {
                for (x = 0; x < xlen; x++) {
                  F3(outarray, 0, y, x) = F2(inarray, y, x);
                }
              }
            } else {
              for (y = 0; y < ylen; y++) {
                for (x = 0; x < xlen; x++) {
                  F3(outarray, 0, y, x) = F3(inarray, channel, y, x);
                }
              }
            }
            return(TRUE);
          }
    
          array1 = iarray_create_2D(oylen, oxlen, K_FLOAT);
          array2 = iarray_create_2D(oylen, oxlen, K_FLOAT);
          array3 = iarray_create_2D(oylen, oxlen, K_FLOAT);
          if ( (array1 == NULL) || (array2 == NULL) || 
                (array3 == NULL) ) {
            return(FALSE);
          }
    
          zero = 0.0;
          iarray_fill(array1, &zero);
          iarray_fill(array2, &zero);
          iarray_fill(array3, &zero);
    
          if (dimen == 2) {
            for (y = 0; y < ylen; y++) {
              for (x = 0; x < xlen; x++) {
                F2(array1, y, x) = F2(inarray, y, x);
              }
            }
          } else {
            for (y = 0; y < ylen; y++) {
              for (x = 0; x < xlen; x++) {
                F2(array1, y, x) = F3(inarray, channel, y, x);
              }
            }
          }
    
          in = array1;
          out = array3;
          
          step = 1;
          step2 = 2*step;
          for (k = 1; k <= level; k++) {

/* first left strip */
            iarray_wavelet_2D(in, array2, 0, step, 0, ylen, 
                                WAVELET_X, step, SPLINE_ATROUS_L1); 
/* second left strip */
            iarray_wavelet_2D(in, array2, step, step2, 0, ylen, 
                                WAVELET_X, step, SPLINE_ATROUS_L2); 
/* first right strip */
            iarray_wavelet_2D(in, array2, xlen-step2, xlen-step, 0, ylen, 
                                WAVELET_X, step, SPLINE_ATROUS_U1); 
/* second right strip */
            iarray_wavelet_2D(in, array2, xlen-step, xlen, 0, ylen, 
                                WAVELET_X, step, SPLINE_ATROUS_U2); 
/* central part in X */
            iarray_wavelet_2D(in, array2, step2, xlen-step2, 0, ylen, 
                                WAVELET_X, step, SPLINE_ATROUS); 
/* first bottom row */
            iarray_wavelet_2D(array2, out, 0, xlen, 0, step, 
                                WAVELET_Y, step, SPLINE_ATROUS_L1); 
/* second bottom row */
            iarray_wavelet_2D(array2, out, 0, xlen, step, step2, 
                                WAVELET_Y, step, SPLINE_ATROUS_L2); 
/* first top row */
            iarray_wavelet_2D(array2, out, 0, xlen, ylen-step2, ylen-step, 
                                WAVELET_Y, step, SPLINE_ATROUS_U1); 
/* second top row */
            iarray_wavelet_2D(array2, out, 0, xlen, ylen-step, ylen, 
                                WAVELET_Y, step, SPLINE_ATROUS_U2); 
/* central part in Y */    
            iarray_wavelet_2D(array2, out, 0, xlen, step2, ylen-step2, 
                                WAVELET_Y, step, SPLINE_ATROUS); 

 /* copy to output array */   
            for (y = 0; y < ylen; y++) {
              for (x = 0; x < xlen; x++) {
                F3(outarray, k-1, y, x) = F2(in,  y, x) - F2(out, y, x);
              }
            }
if (k==1) {
  iarray_write(in, "in");
  iarray_write(out, "out");
}
            tmp   = out;
            out   = in;
            in    = tmp;
            step  = 2 * step;
            step2 = 2 * step;
          }
/*  residual plane */          
          for (y = 0; y < ylen; y++) {
            for (x = 0; x < xlen; x++) {
              F3(outarray, level, y, x) = F2(in, y, x);
            }
          }
          iarray_dealloc(array1);
          iarray_dealloc(array2);
          iarray_dealloc(array3);

          break;
        }
      }
      break;
    }
    case WAVELET_XYZ : {
      dimen = iarray_num_dim(inarray);
      if (dimen != 3) {
        return(FALSE);
      }
  
      outdimen = iarray_num_dim(outarray);
      if (outdimen != 4) {
        return(FALSE);
      }
  
      xlen = iarray_dim_length(inarray, 2);
      ylen = iarray_dim_length(inarray, 1);
      zlen = iarray_dim_length(inarray, 0);

      oxlen = iarray_dim_length(outarray, 3);
      oylen = iarray_dim_length(outarray, 2);
      ozlen = iarray_dim_length(outarray, 1); 
      ollen = iarray_dim_length(outarray, 0); 

      if ( (oxlen != xlen) || (oylen != ylen) || (ozlen != zlen) ||
           (ollen != (level+1))) {
        return(FALSE);
      }

      switch (wavelet_type) {
        default : 
        case  SPLINE_ATROUS : {
          if (level == 0) {                   /* just a copy of the data */
            for (z = 0; z < zlen; z++) {
              for (y = 0; y < ylen; y++) {
                for (x = 0; x < xlen; x++) {
                  F4(outarray, 0, z, y, x) = F3(inarray, z, y, x);
                }
              }
            }
            return(TRUE);
          }
    
          array1 = iarray_create_3D(ozlen, oylen, oxlen, K_FLOAT);
          array2 = iarray_create_3D(ozlen, oylen, oxlen, K_FLOAT);
          array3 = iarray_create_3D(ozlen, oylen, oxlen, K_FLOAT);
          array4 = iarray_create_3D(ozlen, oylen, oxlen, K_FLOAT);
          if ( (array1 == NULL) || (array2 == NULL) || 
                (array3 == NULL) || (array4 == NULL) ) {
            return(FALSE);
          }
          zero = 0.0;
          iarray_fill(array1, &zero);
          iarray_fill(array2, &zero);
          iarray_fill(array3, &zero);
          iarray_fill(array4, &zero);
    

          for (z = 0; z < zlen; z++) {
            for (y = 0; y < ylen; y++) {
              for (x = 0; x < xlen; x++) {
                F3(array1, z, y, x) = F3(inarray, z, y, x);
              }
            }
          }
    
          in = array1;
          out = array4;
          
          step = 1;
          step2 = 2*step;
          for (k = 1; k <= level; k++) {
printf("level %d\n", k);
/* in X */
            iarray_wavelet_3D(in, array2, 0, step, 0, ylen, 0, zlen,  
                                WAVELET_X, step, SPLINE_ATROUS_L1); 
            iarray_wavelet_3D(in, array2, step, step2, 0, ylen, 0, zlen,
                                WAVELET_X, step, SPLINE_ATROUS_L2); 

            iarray_wavelet_3D(in, array2, xlen-step2, xlen-step, 0, ylen,
                                0, zlen,  
                                WAVELET_X, step, SPLINE_ATROUS_U1); 
            iarray_wavelet_3D(in, array2, xlen-step, xlen, 0, ylen, 0, zlen,
                                WAVELET_X, step, SPLINE_ATROUS_U2); 
            iarray_wavelet_3D(in, array2, step2, xlen-step2, 0, ylen, 0, zlen,
                                WAVELET_X, step, SPLINE_ATROUS); 
/* in Y */
            iarray_wavelet_3D(array2, array3, 0, xlen, 0, step, 0, zlen,
                                WAVELET_Y, step, SPLINE_ATROUS_L1); 
            iarray_wavelet_3D(array2, array3, 0, xlen, step, step2, 0, zlen, 
                                WAVELET_Y, step, SPLINE_ATROUS_L2); 
            iarray_wavelet_3D(array2, array3, 0, xlen, ylen-step2, ylen-step,
                                0, zlen,  
                                WAVELET_Y, step, SPLINE_ATROUS_U1); 
            iarray_wavelet_3D(array2, array3, 0, xlen, ylen-step, ylen, 0, zlen,
                                WAVELET_Y, step, SPLINE_ATROUS_U2); 
            iarray_wavelet_3D(array2, array3, 0, xlen, step2, ylen-step2, 
                                0, zlen,
                                WAVELET_Y, step, SPLINE_ATROUS); 
/* in Z */

            iarray_wavelet_3D(array3, out, 0, xlen, 0, ylen, 0, step,
                                WAVELET_Z, step, SPLINE_ATROUS_L1); 
            iarray_wavelet_3D(array3, out, 0, xlen, 0, ylen, step, step2, 
                                WAVELET_Z, step, SPLINE_ATROUS_L2); 
            iarray_wavelet_3D(array3, out, 0, xlen, 0, ylen, 
                                zlen-step2, zlen-step,
                                WAVELET_Z, step, SPLINE_ATROUS_U1); 
            iarray_wavelet_3D(array3, out, 0, xlen, 0, ylen, 
                                zlen-step, zlen, 
                                WAVELET_Z, step, SPLINE_ATROUS_U2); 
            iarray_wavelet_3D(array3, out, 0, xlen, 0, ylen, 
                                step2, zlen-step2, 
                                WAVELET_Z, step, SPLINE_ATROUS); 

 /* copy to output array */   

            for (z = 0; z < zlen; z++) {
              for (y = 0; y < ylen; y++) {
                for (x = 0; x < xlen; x++) {
                  F4(outarray, k-1, z, y, x) = F3(in, z, y, x) - 
                                               F3(out, z, y, x);
                }
              }
            }

            tmp   = out;
            out   = in;
            in    = tmp;
            step  = 2 * step;
            step2 = 2 * step;
          }
/*  residual plane */          
          for (z = 0; z < zlen; z++) {
            for (y = 0; y < ylen; y++) {
              for (x = 0; x < xlen; x++) {
                F4(outarray, level, z, y, x) = F3(in, z, y, x);
              }
            }
          }
          iarray_dealloc(array1);
          iarray_dealloc(array2);
          iarray_dealloc(array3);
          iarray_dealloc(array4);

          break;
        }
        case  MEDIAN_ATROUS : {
          if (level == 0) {                   /* just a copy of the data */
            for (z = 0; z < zlen; z++) {
              for (y = 0; y < ylen; y++) {
                for (x = 0; x < xlen; x++) {
                  F4(outarray, 0, z, y, x) = F3(inarray, z, y, x);
                }
              }
            }
            return(TRUE);
          }
    
          array1 = iarray_create_3D(ozlen, oylen, oxlen, K_FLOAT);
          array2 = iarray_create_3D(ozlen, oylen, oxlen, K_FLOAT);
          array3 = iarray_create_3D(ozlen, oylen, oxlen, K_FLOAT);
          array4 = iarray_create_3D(ozlen, oylen, oxlen, K_FLOAT);
          if ( (array1 == NULL) || (array2 == NULL) || 
                (array3 == NULL) || (array4 == NULL) ) {
            return(FALSE);
          }
          zero = 0.0;
          iarray_fill(array1, &zero);
          iarray_fill(array2, &zero);
          iarray_fill(array3, &zero);
          iarray_fill(array4, &zero);
    

          for (z = 0; z < zlen; z++) {
            for (y = 0; y < ylen; y++) {
              for (x = 0; x < xlen; x++) {
                F3(array1, z, y, x) = F3(inarray, z, y, x);
              }
            }
          }
    
          in = array1;
          out = array4;
          
          step = 1;
          for (k = 1; k <= level; k++) {
            iarray_wavelet_3D(in, out, 0, xlen, 0, ylen, 0, zlen,  
                                WAVELET_X, step, MEDIAN_ATROUS);
            for (z = 0; z < zlen; z++) {
              for (y = 0; y < ylen; y++) {
                for (x = 0; x < xlen; x++) {
                  F4(outarray, k-1, z, y, x) = F3(in, z, y, x) - 
                                               F3(out, z, y, x);
                }
              }
            }

            tmp   = out;
            out   = in;
            in    = tmp;
            step  = 2 * step;
          }
/*  residual plane */          
          for (z = 0; z < zlen; z++) {
            for (y = 0; y < ylen; y++) {
              for (x = 0; x < xlen; x++) {
                F4(outarray, level, z, y, x) = F3(in, z, y, x);
              }
            }
          }
          iarray_dealloc(array1);
          iarray_dealloc(array2);
          iarray_dealloc(array3);
          iarray_dealloc(array4);

          break;
        }
      }
      break;
    }
  }
  
  return(TRUE);
}

int   wavelet_to_iarray(iarray inarray, int wavelet_type, int level,
                        iarray outarray, int channel)
{
  int     indimen, outdimen, xlen, ylen, zlen, levlen;
  int     oxlen, oylen, ozlen;
  int     x, y, z, l;

  if (inarray == NULL) {
    return(FALSE);
  }
  if (outarray == NULL) {
    return(FALSE);
  }
    
  indimen = iarray_num_dim(inarray);
  if ( (indimen != 3) && (indimen != 4) ) {
    return(FALSE);
  }
  outdimen = iarray_num_dim(outarray);
  if ((outdimen != 3) && (outdimen != 2)) {
    return(FALSE);
  }
  if (outdimen == 2) {
    if (indimen != 3) {
      return(FALSE);
    }
  }

  if (indimen == 3) {
    xlen = iarray_dim_length(inarray, 2);
    ylen = iarray_dim_length(inarray, 1);
    zlen = 0;  /*  Keep compiler happy  */
    levlen = iarray_dim_length(inarray, 0);
    if (level >= levlen) {
      return(FALSE);
    }
  } else {
    xlen = iarray_dim_length(inarray, 3);
    ylen = iarray_dim_length(inarray, 2);
    zlen = iarray_dim_length(inarray, 1);
    levlen = iarray_dim_length(inarray, 0);
    if (level >= levlen) {
      return(FALSE);
    }
  }

  if (outdimen == 2) {
    oxlen = iarray_dim_length(outarray, 1);
    oylen = iarray_dim_length(outarray, 0);

    for (y = 0; y < ylen; y++) {
      for (x = 0; x < xlen; x++) {
        F2(outarray, y, x) = F3(inarray, 0, y, x);
      }
    }

    for (z = 1; z <= level; z++) {
      for (y = 0; y < ylen; y++) {
        for (x = 0; x < xlen; x++) {
          F2(outarray, y, x) += F3(inarray, z, y, x);
        }
      }
    }
  } else {
    if (indimen == 3) {
      oxlen = iarray_dim_length(outarray, 2);
      oylen = iarray_dim_length(outarray, 1);
      ozlen = iarray_dim_length(outarray, 0);
      if ((channel < 0) || (channel >= ozlen)) {
        return(FALSE);
      }
      for (y = 0; y < ylen; y++) {
        for (x = 0; x < xlen; x++) {
          F3(outarray, channel, y, x) = F3(inarray, 0, y, x);
        }
      }
      for (z = 1; z <= level; z++) {
        for (y = 0; y < ylen; y++) {
          for (x = 0; x < xlen; x++) {
            F3(outarray, channel, y, x) += F3(inarray, z, y, x);
          }
        }
      }
    } else {
      for (z = 0; z < zlen; z++) {
        for (y = 0; y < ylen; y++) {
          for (x = 0; x < xlen; x++) {
            F3(outarray, z, y, x) = F4(inarray, 0, z, y, x);
          }
        }
      }
      
      for (l = 1; l <= level; l++) {
        for (z = 1; z < zlen; z++) {
          for (y = 0; y < ylen; y++) {
            for (x = 0; x < xlen; x++) {
              F3(outarray, z, y, x) += F4(inarray, l, z, y, x);
            }
          }
        }
      }
    }
  }
  

   
  return(TRUE);


}

static void iarray_wavelet_2D(iarray inarray, iarray outarray, 
                        int xlow, int xhig, int ylow, int yhig, 
                        int mode, int step, int wavelet_type)
{
  int   x, y;
  int   step2;
  
  if ( (inarray == NULL) || (outarray == NULL) ) {
    return;
  }
  if ( (mode != WAVELET_X) && (mode != WAVELET_Y) ) {
    return;
  }

  step2 = step * 2;
  
  switch (wavelet_type) {
    case SPLINE_ATROUS_L1 : { 
      switch (mode) {
        case  WAVELET_X : {
          for (y = ylow; y < yhig; y++) {
            for (x = xlow; x < xhig; x++) {
              F2(outarray, y, x) = 0.0625 *  F2(inarray, y, x + step2) +
                                   0.25   *  F2(inarray, y, x + step)  +
                                   0.375  *  F2(inarray, y, x);
            }
          }
          break;
        }
        case  WAVELET_Y : {
          for (y = ylow; y < yhig; y++) {
            for (x = xlow; x < xhig; x++) {
              F2(outarray, y, x) = 0.0625 *  F2(inarray, y + step2, x) +
                                   0.25   *  F2(inarray, y + step, x)  +
                                   0.375  *  F2(inarray, y, x);
            }
          }
          break;
        }
      }
      break;
    }
    case SPLINE_ATROUS_L2 : { 
      switch (mode) {
        case  WAVELET_X : {
          for (y = ylow; y < yhig; y++) {
            for (x = xlow; x < xhig; x++) {
              F2(outarray, y, x) = 0.0625 *  F2(inarray, y, x + step2)  +
                                   0.25   *  (F2(inarray, y, x + step)  +
                                              F2(inarray, y, x - step)) +
                                   0.375  *  F2(inarray, y, x);
            }
          }
          break;
        }
        case  WAVELET_Y : {
          for (y = ylow; y < yhig; y++) {
            for (x = xlow; x < xhig; x++) {
              F2(outarray, y, x) = 0.0625 *  F2(inarray, y + step2, x)  +
                                   0.25   *  (F2(inarray, y + step, x)  +
                                              F2(inarray, y - step, x)) +
                                   0.375  *  F2(inarray, y, x);
            }
          }
          break;
        }
      }
      break;
    }
    case SPLINE_ATROUS_U1 : {
      switch (mode) {
        case  WAVELET_X : {
          for (y = ylow; y < yhig; y++) {
            for (x = xlow; x < xhig; x++) {
              F2(outarray, y, x) = 0.0625 *  F2(inarray, y, x - step2)  +
                                   0.25   *  (F2(inarray, y, x + step)  +
                                              F2(inarray, y, x - step)) +
                                   0.375  *  F2(inarray, y, x);
            }
          }
          break;
        }
        case  WAVELET_Y : {
          for (y = ylow; y < yhig; y++) {
            for (x = xlow; x < xhig; x++) {
              F2(outarray, y, x) = 0.0625 *  F2(inarray, y - step2, x)  +
                                   0.25   *  (F2(inarray, y + step, x)  +
                                              F2(inarray, y - step, x)) +
                                   0.375  *  F2(inarray, y, x);
            }
          }
          break;
        }
      }
      break;
    }
    case SPLINE_ATROUS_U2 : {
      switch (mode) {
        case  WAVELET_X : {
          for (y = ylow; y < yhig; y++) {
            for (x = xlow; x < xhig; x++) {
              F2(outarray, y, x) = 0.0625 *  F2(inarray, y, x - step2) +
                                   0.25   *  F2(inarray, y, x - step)  +
                                   0.375  *  F2(inarray, y, x);
            }
          }
          break;
        }
        case  WAVELET_Y : {
          for (y = ylow; y < yhig; y++) {
            for (x = xlow; x < xhig; x++) {
              F2(outarray, y, x) = 0.0625 *  F2(inarray, y - step2, x) +
                                   0.25   *  F2(inarray, y - step, x)  +
                                   0.375  *  F2(inarray, y, x);
            }
          }
          break;
        }
      }
      break;
    }
    case SPLINE_ATROUS  : {
      switch (mode) {
        case  WAVELET_X : {
          for (y = ylow; y < yhig; y++) {
            for (x = xlow; x < xhig; x++) {
              F2(outarray, y, x) = 0.0625 * (F2(inarray, y, x - step2)  + 
                                             F2(inarray, y, x + step2)) +
                                   0.25   * (F2(inarray, y, x - step)   + 
                                             F2(inarray, y, x + step))  +
                                   0.375  *  F2(inarray, y, x);
            }
          }
          break;
        }
        case WAVELET_Y : {
          for (y = ylow; y < yhig; y++) {
            for (x = xlow; x < xhig; x++) {
              F2(outarray, y, x) = 0.0625* (F2(inarray, y - step2, x) + 
                                       F2(inarray, y + step2, x))       +
                              0.25  * (F2(inarray, y - step,  x) + 
                                       F2(inarray, y + step,  x))       +
                              0.375 *  F2(inarray, y, x);
            }
          }
          break;
        }
      }
      break;
    }
    default : {
      return;
    }
  }
  
}

static void iarray_wavelet_3D(iarray inarray, iarray outarray, 
                        int xlow, int xhig, int ylow, int yhig, 
                        int zlow, int zhig, 
                        int mode, int step, int wavelet_type)
{
  int       x, y, z, xx, yy, zz, i;
  int       step2;
  double    *s_array;
  
  if ( (inarray == NULL) || (outarray == NULL) ) {
    return;
  }
  if ( (mode != WAVELET_X) && (mode != WAVELET_Y) && (mode != WAVELET_Z) ) {
    return;
  }

  step2 = step * 2;

  switch (wavelet_type) {
    case SPLINE_ATROUS_L1 : { 
      switch (mode) {
        case  WAVELET_X : {
          for (z = zlow; z < zhig; z++) {
            for (y = ylow; y < yhig; y++) {
              for (x = xlow; x < xhig; x++) {
                F3(outarray, z, y, x) = 0.0625 *  F3(inarray, z, y, x + step2) +
                                     0.25   *  F3(inarray, z, y, x + step)  +
                                     0.375  *  F3(inarray, z, y, x);
              }
            }
          }
          break;
        }
        case  WAVELET_Y : {
          for (z = zlow; z < zhig; z++) {
            for (y = ylow; y < yhig; y++) {
              for (x = xlow; x < xhig; x++) {
                F3(outarray, z, y, x) = 0.0625 *  F3(inarray, z, y+step2, x) +
                                     0.25   *  F3(inarray, z, y+step, x)  +
                                     0.375  *  F3(inarray, z, y, x);
              }
            }
          }
          break;
        }
        case  WAVELET_Z : {
          for (z = zlow; z < zhig; z++) {
            for (y = ylow; y < yhig; y++) {
              for (x = xlow; x < xhig; x++) {
                F3(outarray, z, y, x) = 0.0625 *  F3(inarray, z+step2, y, x) +
                                     0.25   *  F3(inarray, z+step, y, x)  +
                                     0.375  *  F3(inarray, z, y, x);
              }
            }
          }
          break;
        }
      }
      break;
    }
    case SPLINE_ATROUS_L2 : { 
      switch (mode) {
        case  WAVELET_X : {
          for (z = zlow; z < zhig; z++) {
            for (y = ylow; y < yhig; y++) {
              for (x = xlow; x < xhig; x++) {
                F3(outarray, z, y, x) = 0.0625 *  F3(inarray, z, y, x + step2) +
                                     0.25   *  (F3(inarray, z, y, x + step)  +
                                                F3(inarray, z, y, x-step) ) +
                                     0.375  *  F3(inarray, z, y, x);
              }
            }
          }
          break;
        }
        case  WAVELET_Y : {
          for (z = zlow; z < zhig; z++) {
            for (y = ylow; y < yhig; y++) {
              for (x = xlow; x < xhig; x++) {
                F3(outarray, z, y, x) = 0.0625 *  F3(inarray, z, y+step2, x) +
                                     0.25   *  (F3(inarray, z, y+step, x)  +
                                                F3(inarray, z, y-step, x) ) +
                                     0.375  *  F3(inarray, z, y, x);
              }
            }
          }
          break;
        }
        case  WAVELET_Z : {
          for (z = zlow; z < zhig; z++) {
            for (y = ylow; y < yhig; y++) {
              for (x = xlow; x < xhig; x++) {
                F3(outarray, z, y, x) = 0.0625 *  F3(inarray, z+step2, y, x) +
                                     0.25   * (F3(inarray, z+step, y, x)  +
                                                F3(inarray, z-step, y, x) ) +
                                     0.375  *  F3(inarray, z, y, x);
              }
            }
          }
          break;
        }
      }
      break;
    }
    case SPLINE_ATROUS_U1 : {
      switch (mode) {
        case  WAVELET_X : {
          for (z = zlow; z < zhig; z++) {
            for (y = ylow; y < yhig; y++) {
              for (x = xlow; x < xhig; x++) {
                F3(outarray, z, y, x) = 0.0625 *  F3(inarray, z, y, x-step2) +
                                     0.25   *  (F3(inarray, z, y, x + step)  +
                                                F3(inarray, z, y, x-step) ) +
                                     0.375  *  F3(inarray, z, y, x);
              }
            }
          }
          break;
        }
        case  WAVELET_Y : {
          for (z = zlow; z < zhig; z++) {
            for (y = ylow; y < yhig; y++) {
              for (x = xlow; x < xhig; x++) {
                F3(outarray, z, y, x) = 0.0625 *  F3(inarray, z, y-step2, x) +
                                     0.25   *  (F3(inarray, z, y+step, x)  +
                                                F3(inarray, z, y-step, x) ) +
                                     0.375  *  F3(inarray, z, y, x);
              }
            }
          }
          break;
        }
        case  WAVELET_Z : {
          for (z = zlow; z < zhig; z++) {
            for (y = ylow; y < yhig; y++) {
              for (x = xlow; x < xhig; x++) {
                F3(outarray, z, y, x) = 0.0625 *  F3(inarray, z-step2, y, x) +
                                     0.25   * (F3(inarray, z+step, y, x)  +
                                                F3(inarray, z-step, y, x) ) +
                                     0.375  *  F3(inarray, z, y, x);
              }
            }
          }
          break;
        }
      }
      break;
    }
    case SPLINE_ATROUS_U2 : {
      switch (mode) {
        case  WAVELET_X : {
          for (z = zlow; z < zhig; z++) {
            for (y = ylow; y < yhig; y++) {
              for (x = xlow; x < xhig; x++) {
                F3(outarray, z, y, x) = 0.0625 *  F3(inarray, z, y, x-step2) +
                                     0.25   *  F3(inarray, z, y, x-step) +
                                     0.375  *  F3(inarray, z, y, x);
              }
            }
          }
          break;
        }
        case  WAVELET_Y : {
          for (z = zlow; z < zhig; z++) {
            for (y = ylow; y < yhig; y++) {
              for (x = xlow; x < xhig; x++) {
                F3(outarray, z, y, x) = 0.0625 *  F3(inarray, z, y-step2, x) +
                                     0.25   *  F3(inarray, z, y-step, x)  +
                                     0.375  *  F3(inarray, z, y, x);
              }
            }
          }
          break;
        }
        case  WAVELET_Z : {
          for (z = zlow; z < zhig; z++) {
            for (y = ylow; y < yhig; y++) {
              for (x = xlow; x < xhig; x++) {
                F3(outarray, z, y, x) = 0.0625 *  F3(inarray, z-step2, y, x) +
                                     0.25   *  F3(inarray, z-step, y, x)  +
                                     0.375  *  F3(inarray, z, y, x);
              }
            }
          }
          break;
        }
      }
      break;
    }
    case SPLINE_ATROUS  : {
      switch (mode) {
        case  WAVELET_X : {
          for (z = zlow; z < zhig; z++) {
            for (y = ylow; y < yhig; y++) {
              for (x = xlow; x < xhig; x++) {
                F3(outarray, z, y, x) = 0.0625 *  (F3(inarray, z, y, x-step2) +
                                                   F3(inarray, z, y, x+step2)) +
                                     0.25   *  (F3(inarray, z, y, x + step)  +
                                                F3(inarray, z, y, x-step) ) +
                                     0.375  *  F3(inarray, z, y, x);
              }
            }
          }
          break;
        }
        case  WAVELET_Y : {
          for (z = zlow; z < zhig; z++) {
            for (y = ylow; y < yhig; y++) {
              for (x = xlow; x < xhig; x++) {
                F3(outarray, z, y, x) = 0.0625 *  (F3(inarray, z, y-step2, x) +
                                                   F3(inarray, z, y+step2, x)) +
                                     0.25   *  (F3(inarray, z, y+step, x)  +
                                                F3(inarray, z, y-step, x) ) +
                                     0.375  *  F3(inarray, z, y, x);
              }
            }
          }
          break;
        }
        case  WAVELET_Z : {
          for (z = zlow; z < zhig; z++) {
            for (y = ylow; y < yhig; y++) {
              for (x = xlow; x < xhig; x++) {
                F3(outarray, z, y, x) = 0.0625 *  (F3(inarray, z-step2, y, x) +
                                                   F3(inarray, z+step2, y, x)) +
                                     0.25   *  (F3(inarray, z+step, y, x)  +
                                                F3(inarray, z-step, y, x) ) +
                                     0.375  *  F3(inarray, z, y, x);
              }
            }
          }
          break;
        }
      }
      break;
    }
    case MEDIAN_ATROUS : {
      s_array = mal_dvector(1, 27);
      for (z = zlow+step; z < zhig-step; z++) {
        for (y = ylow+step; y < yhig-step; y++) {
          for (x = xlow+step; x < xhig-step; x++) {
            i = 1;
            for (zz = -step; zz <= step; zz += step) {
              for (yy = -step; yy <= step; yy += step) {
                for (xx = -step; xx <= step; xx += step) {
                  s_array[i++] = F3(inarray, z+zz, y+yy, x+xx);
                }
              }
            }
            sort(27, s_array);
            F3(outarray, z, y, x) = s_array[14];
          }
        }
      }
      free_dvector(s_array, 1, 27);
      break;
    }
    default : {
      return;
    }
  }
  
}
   
    


int  mask_filter(iarray *mask)
/*
  Performs median filter on a multi-resolution mask. Returns FALSE if *mask is
  not of type K_INT, or if dimension is not 2.
  
  Known bugs: does not handle boundary yet.
  
  Updates:  
    17-May-95:  Created
    18-May-95:  changed from normal median to median in every plane. The
                algorithm is not optinal for this, but who cares.....

*/
{
  iarray  newmask, klad;
  int     x, y, xx, yy;
  int     xlen, ylen, dimen, data_type;
  int     comp, c, level;
  
  if ((data_type = iarray_type(*mask)) != K_INT) {
    return(FALSE);
  }
  
  dimen = iarray_num_dim(*mask);
  if (dimen != 2) {
    return(FALSE);
  }

  xlen = iarray_dim_length(*mask, 1);
  ylen = iarray_dim_length(*mask, 0);
  
  newmask = iarray_create_2D(ylen, xlen, K_INT);
  for (y = 0; y < ylen; y++) {
    for (x = 0; x < xlen; x++) {
      I2(newmask, y, x) = 0;
    }
  }

    c = 1;
    
    for (y = 1; y < ylen-1; y++) {
      for (x = 1; x < xlen-1; x++) {
        comp = 0;
        for (yy = -1; yy < 2; yy++) {
          for (xx = -1; xx < 2; xx++) {
            comp += I2(*mask, y+yy, x+xx) & c;
          }
        }
        if (comp > 4) {
          I2(newmask, y, x) += c;
        }
      }
    }
  for (level = 1; level < 8 ; level++) {
    c = 1 << level;
    
    for (y = 1; y < ylen-1; y++) {
      for (x = 1; x < xlen-1; x++) {
        comp = 0;
        for (yy = -1; yy < 2; yy++) {
          for (xx = -1; xx < 2; xx++) {
            comp += (I2(*mask, y+yy, x+xx) & c) >> level;
          }
        }
        if (comp > 4) {
          I2(newmask, y, x) += c;
        }
      }
    }
  }
  klad = *mask;
  *mask = newmask;
  newmask = klad;
  iarray_dealloc(newmask);
  
  return(TRUE); 

}

#undef  WAVELET

static void sort (n, ra)
int       n;
double *ra;
{
  int l,j,ir,i;
  double rra;

  l=(n >> 1)+1;
  ir=n;
  for (;;) {
    if (l > 1) {
      rra=ra[--l];
    } else {
      rra=ra[ir];
      ra[ir]=ra[1];
      if (--ir == 1) {
        ra[1]=rra;
        return;
      }
    }
    i=l;
    j=l << 1;
    while (j <= ir) {
      if (j < ir && ra[j] < ra[j+1])
        ++j;
      if (rra < ra[j]) {
        ra[i]=ra[j];
        j += (i=j);
      }	else
        j=ir+1;
    }
    ra[i]=rra;
  }
}
