#include <karma_vrender_def.h>
#include <karma_overlay.h>
#include <karma_iarray_def.h>
#include <karma_canvas.h>
#include <karma_kcmap.h>
#include "render.h"


#define DEFAULT_COLOURMAP_NAME "Greyscale1"
#define DEFAULT_DISTANCE 50  /*  centimeters from screen (cube centre)  */
#define INTENSITY_GRADATIONS 100

#define ANGLE_RANGE 360

#define NUM_PROJECTIONS 2

#ifdef NEED_PROJECTION_ALTERNATIVES
static char *projection_alternatives[NUM_PROJECTIONS] =
{
    "orthogonal",
    "perspective"
};
#endif

#define MODE_NORMAL         0
#define MODE_SMOOTH_CACHE   1
#define MODE_SMOOTH_ALWAYS  2
#define MODE_SMOOTH_DELAYED 3
#define NUM_MODES 4

#ifdef NEED_MODE_ALTERNATIVES
static char *mode_alternatives[NUM_MODES] =
{
    "normal",
    "smooth cache",
    "always smooth",
    "smooth delayed",
};
#endif

#define CURSOR_3D_NONE                  0
#define CURSOR_3D_OVERLAY_2HAIR_IN_2D   1
#define CURSOR_3D_OVERLAY_3HAIR         2
#define CURSOR_3D_SHADED_2HAIR_IN_2D    3
#define CURSOR_3D_SHADED_3HAIR          4
#define CURSOR_3D_SHADED_2HAIR_IN_3D    5
#define CURSOR_3D_SHADED_SPHERE         6
#define NUM_CURSOR_3D_ALTERNATIVES      7

#ifdef NEED_CURSOR_3D_ALTERNATIVES
static char *cursor_3d_alternatives[NUM_CURSOR_3D_ALTERNATIVES] =
{
    "none",
    "overlay 2-hair in 2-D",
    "overlay 3-hair",
    "shaded 2-hair in 2-D",
    "shaded 3-hair",
    "shaded 2-hair in 3-D",
    "shaded sphere",
};
#endif

#define FILTER_SIMPLE             0
#define FILTER_MEDMASK_FAST       1
#define FILTER_MEDMASK_SLOW       2
#define FILTER_3D                 3
#define FILTER_SUBTRACT_CONTINUUM 4
#define NUM_FILTER_ALTERNATIVES 5

#ifdef NEED_FILTER_ALTERNATIVES
static char *filter_alternatives[NUM_FILTER_ALTERNATIVES] =
{
    "simple",
    "slow median mask",
    "fast median mask",
    "3D",
    "subtract continuum",
};
#endif

#define CURSOR_2D_DUMB   0
#define CURSOR_2D_ACTIVE 1
#define NUM_CURSOR_2D_ALTERNATIVES 2
#ifdef NEED_CURSOR_2D_ALTERNATIVES
static char *cursor_2d_alternatives[] =
{
    "dumb",
    "active",
};
#endif

typedef struct
{
    float roll;   /*  Radians                            */
    float pitch;  /*  Radians                            */
    float yaw;    /*  Radians                            */
    float rho;    /*  Distance to screen in centimeters  */
} rotation_specification;

typedef struct
{
    int x;
    int y;
    int z;
} i_coord_3d;

typedef struct
{
    float red;
    float green;
    float blue;
} RGBcolour;

typedef struct
{
    flag did_stereo;
} post_render_info;

typedef struct segment_list_type
{
    void *popup_widget;
    unsigned int num_segments;
    Kcoord_3d *p1;
    Kcoord_3d *p2;
    char colourname[STRING_LENGTH];
    struct segment_list_type *prev;
    struct segment_list_type *next;
} segment_list;

typedef struct
{
    double bottom;
    double top;
} histogram_scale;


/*  External functions  */
/*  File: initialise.c  */
EXTERN_FUNCTION (void initialise, () );
EXTERN_FUNCTION (void register_voxel_offset, (signed char offset) );

/*  File: compute.c  */
EXTERN_FUNCTION (void initialise, () );
EXTERN_FUNCTION (flag process_cube, (multi_array *multi_desc) );
EXTERN_FUNCTION (void compute, () );
EXTERN_FUNCTION (void compute_histogram, (iarray cube) );

/*  File: orient.c  */
EXTERN_FUNCTION (void initialise_orientation_canvas, (KPixCanvas pixcanvas) );
EXTERN_FUNCTION (flag orientation_post_render_callback,
		 (void *object, void *client1_data,
		  void *call_data, void *client2_data) );

/*  File: spaceball.c  */
#ifdef HAVE_SPACEBALL
EXTERN_FUNCTION (flag initialise_spaceball, () );
EXTERN_FUNCTION (void spin_sb, (float roll, float pitch, float yaw) );
#endif

/*  File: cont_hg.c  */
EXTERN_FUNCTION (void setup_cont_hg_canvas,
		 (KPixCanvas hist_pixcanvas, KPixCanvas cmap_pixcanvas,
		  Kcolourmap cmap) );
EXTERN_FUNCTION (void compute_cont_hg_info, () );

/*  File: threecolour_hg.c  */
EXTERN_FUNCTION (void compute_threecolour_hg_info, () );

/*  File: mono_hg.c  */
EXTERN_FUNCTION (void setup_mono_hg_canvas, (KPixCanvas pixcanvas) );
EXTERN_FUNCTION (void compute_mono_hg_info, () );
EXTERN_FUNCTION (void transmit_mono_hg_info, (KVolumeRenderContext context) );


/*  File: main_xt.c  */
EXTERN_FUNCTION (void freeze_controls, (flag freeze) );
EXTERN_FUNCTION (void reset_view, () );
EXTERN_FUNCTION (flag offset_position_func,
		 (KWorldCanvas canvas, double x, double y,
		  unsigned int event_code, void *e_info,
		  void **f_info, double x_lin, double y_lin) );
EXTERN_FUNCTION (void load_and_setup,
		 (CONST char *filename, flag reset_the_view) );
EXTERN_FUNCTION (void select_visual, (KWorldCanvas canvas) );
EXTERN_FUNCTION (void show_animate_control, (flag display) );
EXTERN_FUNCTION (void refresh_display_window, () );
EXTERN_FUNCTION (void do_rotate, (int roll_inc, int yaw_inc, int pitch_inc) );

/*  File: display.c  */
EXTERN_FUNCTION (flag render, () );
EXTERN_FUNCTION (void destroy_all_vimages, () );

/*  File: geom.c  */
EXTERN_FUNCTION (void geom_rotate_3d_r, (Kcoord_3d *vector, float roll) );
EXTERN_FUNCTION (void geom_rotate_3d_p, (Kcoord_3d *vector, float pitch) );
EXTERN_FUNCTION (void geom_rotate_3d_y, (Kcoord_3d *vector, float yaw) );
EXTERN_FUNCTION (void geom_rotate_3d_rpy,
		 (Kcoord_3d *vector, float roll, float pitch, float yaw) );
EXTERN_FUNCTION (void geom_project_3d_ortho_2d,
		 (Kcoord_3d eye_pos, Kcoord_3d vision,
		  Kcoord_3d vertical, Kcoord_3d point_3d,
		  Kcoord_2d *point_2d) );
EXTERN_FUNCTION (void geom_rect_to_polar,
		 (float x, float y, float z,
		  float *rho, float *theta, float *phi) );
EXTERN_FUNCTION (void geom_polar_to_rect,
		 (float rho, float theta, float phi,
		  float *x, float *y, float *z) );
EXTERN_FUNCTION (void geom_rotate_in_plane,
		 (Kcoord_3d *vector, float angle, Kcoord_3d normal) );
EXTERN_FUNCTION (void geom_vector_multiply,
		 (Kcoord_3d *out, Kcoord_3d vec1, Kcoord_3d vec2) );
EXTERN_FUNCTION (float geom_vector_dot_product,
		 (Kcoord_3d vec1, Kcoord_3d vec2) );


/*  File: active_2d_cursor.c  */
EXTERN_FUNCTION (CONST signed char *get_ray_from_image_plane,
		 (Kcoord_3d *coords, Kcoord_2d pos, unsigned int *ray_length,
		  CONST char *shader_name, unsigned int *depth,
		  float *opacities) );
EXTERN_FUNCTION (unsigned int get_feature_depth,
		 (CONST signed char *ray,unsigned int ray_length,
		  void *pixel_ptr, CONST char *shader_name,
		  float *opacities) );

/*  File: save_movie.c  */
EXTERN_FUNCTION (flag save_movie,
		 (CONST char *file, array_desc *im_arr_desc, CONST char *movie,
		  unsigned int num_frames) );

/*  File: subs_hg.c  */
EXTERN_FUNCTION (void setup_substance_canvases,
		 (KPixCanvas parent, Kdisplay dpy_handle,
		  void *offset_widget, unsigned long foreground_pixel,
		  unsigned long background_pixel, Kcolourmap *subs_cmap) );
EXTERN_FUNCTION (void process_substances, (KVolumeRenderContext context) );
EXTERN_FUNCTION (void transmit_substances, (KVolumeRenderContext context) );
EXTERN_FUNCTION (unsigned int scan_substance_structure,
		 (substance *substances,
		  packet_desc *pack_desc, char *packet) );
EXTERN_FUNCTION (flag make_substance_structure,
		 (substance *substances, unsigned int num_substances,
		  packet_desc **pack_desc, char **packet) );


/*  File: offset.c  */
#ifdef X11
EXTERN_FUNCTION (KWorldCanvas setup_offset_control_canvas,
		 (KPixCanvas pixcanvas, Display *dpy, void *offset_widget,
		 void (*x_func) (signed char x), int *x_ptr) );
#endif

/*  File: ray_display.c  */
#ifdef X11
EXTERN_FUNCTION (void setup_ray_display_canvas,
		 (KPixCanvas pixcanvas, Display *display) );
#endif
EXTERN_FUNCTION (void register_ray,
		 (CONST signed char *ray, unsigned int length,
		  unsigned int depth, CONST float *opacities) );

/*  File: cursor_3d.c  */
EXTERN_FUNCTION (void register_cursor_func,
		 (KWorldCanvas canvas, KVolumeRenderContext context,
		  unsigned int eye_view, flag has_image) );
EXTERN_FUNCTION (flag segments_post_render_callback,
		 (void *object, void *client1_data,
		  void *call_data, void *client2_data) );
EXTERN_FUNCTION (flag change_3d_cursor, (Kcoord_3d new_cursor, flag move) );
EXTERN_FUNCTION (void init_depth_shaders, () );
EXTERN_FUNCTION (void draw_cut_plane, (float cut_plane_depth) );
EXTERN_FUNCTION (flag cursor_3d_position_func,
		 (KWorldCanvas canvas, double x, double y,
		  unsigned int event_code, void *e_info,
		  void **f_info, double x_lin, double y_lin) );
EXTERN_FUNCTION (void compute_cursor_3d_strings,
		 (Kcoord_3d cursor, char pixel_string[STRING_LENGTH],
		  char world_string[STRING_LENGTH]) );
EXTERN_FUNCTION (flag get_colour,
		 (KWorldCanvas canvas, CONST char *name,
		  unsigned long *pixel_value, RGBcolour *colour) );
EXTERN_FUNCTION (flag subvolume_prerender, (KVolumeRenderContext vrend_ctx) );

/*  File: fullcolour_hg.c  */
EXTERN_FUNCTION (KWorldCanvas create_fullcolour_hg_canvas,
		 (KPixCanvas hist_pixcanvas) );
EXTERN_FUNCTION (void transmit_fullcolour_info,
		 (KVolumeRenderContext context) );

/*  File: load_disp.c  */
EXTERN_FUNCTION (flag display_file,
		 (CONST char *inp_filename, KWorldCanvas pseudo_canvas,
		  KWorldCanvas rgb_canvas) );

/*  File: router.c  */
EXTERN_FUNCTION (void register_new_byte_cube,
		 (iarray cube, double scale, double offset) );
EXTERN_FUNCTION (flag generic_new_cube_callback,
		 (void *object, void *client1_data,
		  void *call_data, void *client2_data) );
EXTERN_FUNCTION (void register_subcube_size, () );
EXTERN_FUNCTION (flag generic_subcube_callback,
		 (void *object, void *client1_data,
		  void *call_data, void *client2_data) );
EXTERN_FUNCTION (void register_rotation_change, (rotation_specification rot) );
EXTERN_FUNCTION (flag generic_rotation_callback,
		 (void *object, void *client1_data,
		  void *call_data, void *client2_data) );
EXTERN_FUNCTION (void register_shader, (CONST char *shader) );
EXTERN_FUNCTION (flag generic_shader_callback,
		 (void *object, void *client1_data,
		  void *call_data, void *client2_data) );
EXTERN_FUNCTION (void register_projection, (unsigned int projection) );
EXTERN_FUNCTION (flag generic_projection_callback,
		 (void *object, void *client1_data,
		  void *call_data, void *client2_data) );
EXTERN_FUNCTION (void register_eye_separation, (double eye_separation) );
EXTERN_FUNCTION (flag generic_eye_separation_callback,
		 (void *object, void *client1_data,
		  void *call_data, void *client2_data) );
EXTERN_FUNCTION (void register_histogram_change, (double bottom, double top) );
EXTERN_FUNCTION (flag generic_histogram_change_callback,
		 (void *object, void *client1_data,
		  void *call_data, void *client2_data) );
EXTERN_FUNCTION (void register_3d_cursor_change, (Kcoord_3d new_cursor) );

/*  File: movie_xt.c  */
EXTERN_FUNCTION (void freeze_save_movie_button, (flag freeze) );
