/*  kfindmemleak.c

    Source file for  kfindmemleak  (memory leak finding tool for Karma).

    Copyright (C) 2001  Richard Gooch

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Richard Gooch may be reached by email at  karma-request@atnf.csiro.au
    The postal address is:
      Richard Gooch, c/o ATNF, P. O. Box 76, Epping, N.S.W., 2121, Australia.
*/

/*
    This programme will analyse Karma memory debugging logs and will report
    memory blocks which were not deallocated.


    Written by      Richard Gooch   20-MAR-2001

    Last updated by Richard Gooch   20-MAR-2001


*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <karma.h>
#include <errno.h>


struct memblock_struct
{
    void *ptr;
    char line[STRING_LENGTH];
    struct memblock_struct *prev;
    struct memblock_struct *next;
};


int main (int argc, char **argv)
{
    FILE *in, *out;
    struct memblock_struct *first_memblock = NULL;
    struct memblock_struct *last_memblock = NULL;
    struct memblock_struct *block;
    char line[STRING_LENGTH];

    if (argc > 3)
    {
	fputs ("Usage:\tkfindmemleak [infile] [outfile]\n", stderr);
	exit (1);
    }
    if (argc > 1)
    {
	if ( ( in = fopen (argv[1], "r") ) == NULL )
	{
	    fprintf (stderr, "Error opening: \"%s\"\t%s\n", argv[1],ERRSTRING);
	    exit (1);
	}
    }
    else in = stdin;
    if (argc > 2)
    {
	if ( ( out = fopen (argv[2], "w") ) == NULL )
	{
	    fprintf (stderr, "Error opening: \"%s\"\t%s\n", argv[2],ERRSTRING);
	    exit (1);
	}
    }
    else out = stdout;
    while (fgets (line, STRING_LENGTH, in) != NULL)
    {
	char *char_ptr;
	void *ptr;

	block = NULL;
	if ( (strncmp (line, "Freed", 5) == 0) &&
	     ( ( char_ptr = strstr (line, "ptr: ") ) != NULL ) )
	{
	    ptr = (void *) strtol (char_ptr + 5, NULL, 0);
	    /*  Search  */
	    for (block = first_memblock; block != NULL; block = block->next)
	    {
		if (block->ptr == ptr) break;
	    }
	}
	if (block != NULL)
	{
	    /*  Found a block to destroy  */
	    if (block->prev == NULL) first_memblock = block->next;
	    else block->prev->next = block->next;
	    if (block->next == NULL) last_memblock = block->prev;
	    else block->next->prev = block->prev;
	    free (block);
	    continue;
	}
	if ( ( block = malloc (sizeof *block) ) == NULL )
	{
	    fputs ("Error allocating block\n", stderr);
	    exit (1);
	}
	if ( (strncmp (line, "Allocated", 9) == 0) &&
	     ( ( char_ptr = strstr (line, "ptr: ") ) != NULL ) )
	{
	    block->ptr = (void *) strtol (char_ptr + 5, NULL, 0);
	}
	else block->ptr = NULL;
	strcpy (block->line, line);
	block->prev = last_memblock;
	block->next = NULL;
	if (first_memblock == NULL) first_memblock = block;
	if (last_memblock != NULL) last_memblock->next = block;
	last_memblock = block;
    }
    /*  Now write what's left  */
    for (block = first_memblock; block != NULL; block = block->next)
	fputs (block->line, out);
    return (0);
}   /*  End Function main  */
